!
! Module: exposure
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This module contains routines used in exposure calculations.
! 
! Comments:
!	TODO:***** write in the headers of all appropriate subroutines/functions (i) what cal variables they need to have set; (ii) for the calling program to deallocate any returned pointers. **** ALSO!! when it is necessary to call CAL_getState, the previous state should be stored and reloaded at the end of the routine.
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module exposure

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the '*' from the lib*.a name
! in <package>/src/Makefile.

  use cal		! contains 'use caloaldefs' and 'use caltypes'
  use dal
  use errorHandling
  use dss
!  use eventAttributes	! in saslib/sasdal. For checking which flags have been applied.

  use array_utils	! for smallestTrueIndex, largestTrueIndex,&
    ! trapezoidSum, getGridDiffs, interpolateGridDiff.
  use constants
!  use coordinate	! for instToTan
!  use dal_aux
  use dss_aux
  use dss_ranges_aux
  use geometric_types
!  use intervals_aux	! for isWithinInterval() & andIntervals().
!  use splines
use interpolations
  use ssc_misc
!  use wcs_aux
use dump

  implicit none 

  private 

  public convertPiToEeV
  public convertPhaToEeV
  public sampleEnergyRange
  public getEffectiveArea
  public getFilterTransmission
  public getQe
  public constructQeArray
  public calcQeMap
  public constructPsfCubeTelcoord

  public constructVignettingCube
  public calcVigAtSourcePos
  public multiplyByVigFunction
  public calcVigMap
  public interpolateVigMap
  public interpolateVigMapToArray
  public calcPsfArray
  public checkDss
  public calcTotalGoodTime
!  public constrainSpectrum
!  public correctForGtiAndAlias

  interface convertPiToEeV
    module procedure convertPiRangeToEeV,  convertPiBoundToEeV
    module procedure convertPiRangesToEeV, convertPiBoundsToEeV
    module procedure convertPiToEeVScalarSingle, convertPiToEeVScalarDouble
    module procedure convertPiToEeVVectorSingle, convertPiToEeVVectorDouble
  end interface

  interface convertPhaToEeV
    module procedure convertPhaRangeToEeV,  convertPhaBoundToEeV
    module procedure convertPhaRangesToEeV, convertPhaBoundsToEeV
    module procedure convertPhaToEeVScalar, convertPhaToEeVVector
  end interface

  interface sampleEnergyRange
    module procedure sampleEnergyRangeNormal
    module procedure sampleEnergyRangeVector, sampleEnergyRangeScalar
  end interface

  interface calcVigAtSourcePos
    module procedure calcVigAtSourcePosWithCube, calcVigAtSourcePosNoCube
  end interface

!  type, public :: frameType
!  type :: frameType
!    real(single)    :: fracExp, deadTimeFraction
!    real(double)    :: time
!    type(IntervalT) :: interval
!    logical(bool)   :: mask
!  end type frameType

!type :: bufferedFrameType
!    type(frameType), pointer :: frames(:) => null()
!end type bufferedFrameType

! These don't seem to be used by any package so I'll comment them out (IMS 2006-02-21):
!  real(double), parameter, public ::&
!    minEeV = 50.0,&
!    maxEeV = 20000.0,&
!    binSizeInLnEeV = 0.3,&
!    coarseVigThetaArcsecBinSize = 60.0,&
!    coarseVigPhiBinSize = Pi / 18.0

contains 



! The subroutines up to and including constructPsfCubeTelcoord might be better located in cal_aux..?


  !---------------------------------------------------------------------
  function convertPiRangeToEeV(piRange) result(EeVrange)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    type(RangeT), intent(in) :: piRange
    type(RangeT)             :: EeVrange

    !-------------------------------------------------------------------
    ! Start

    EeVrange%lower = convertPiBoundToEeV(piRange%lower)
    EeVrange%upper = convertPiBoundToEeV(piRange%upper)

  end function convertPiRangeToEeV

  !---------------------------------------------------------------------
  function convertPiRangesToEeV(piRanges) result(EeVranges)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    type(RangeT), intent(in) :: piRanges(:)
    type(RangeT)             :: EeVranges(size(piRanges))

    !-------------------------------------------------------------------
    ! Start

    EeVranges%lower = convertPiBoundsToEeV(piRanges%lower)
    EeVranges%upper = convertPiBoundsToEeV(piRanges%upper)

  end function convertPiRangesToEeV

  !---------------------------------------------------------------------
  function convertPiBoundToEeV(piBound) result(EeVBound)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    type(BoundT), intent(in) :: piBound
    type(BoundT)             :: EeVBound

    !-------------------------------------------------------------------
    ! Start

    EeVBound%type = piBound%type
    EeVBound%value = convertPiToEeVScalarDouble(piBound%value)

  end function convertPiBoundToEeV

  !---------------------------------------------------------------------
  function convertPiBoundsToEeV(piBounds) result(EeVBounds)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    type(BoundT), intent(in) :: piBounds(:)
    type(BoundT)             :: EeVBounds(size(piBounds))

    !-------------------------------------------------------------------
    ! Start

    EeVBounds%type = piBounds%type
    EeVBounds%value = convertPiToEeVVectorDouble(piBounds%value)

  end function convertPiBoundsToEeV

  !---------------------------------------------------------------------
  function convertPiToEeVScalarSingle(piValue) result(EeVValue)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    real(single), intent(in) :: piValue
    real(single)             :: EeVValue

    real(single) :: EeVValues(1)

    !-------------------------------------------------------------------
    ! Start

    EeVValues = convertPiToEeVVectorSingle((/piValue/))
    EeVValue = EeVValues(1)

  end function convertPiToEeVScalarSingle

  !---------------------------------------------------------------------
  function convertPiToEeVScalarDouble(piValue) result(EeVValue)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    real(double), intent(in) :: piValue
    real(double)             :: EeVValue

    real(double) :: EeVValues(1)

    !-------------------------------------------------------------------
    ! Start

    EeVValues = convertPiToEeVVectorDouble((/piValue/))
    EeVValue = EeVValues(1)

  end function convertPiToEeVScalarDouble


  !---------------------------------------------------------------------
  function convertPiToEeVVectorSingle(piValues) result(EeVValues)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    real(single), intent(in) :: piValues(:)
    real(single)             :: EeVValues(size(piValues))

    real(single), pointer :: ePointer(:) => null()

    !-------------------------------------------------------------------
    ! Start

    call CAL_piToeV(piValues, ePointer)	! SOC-SPR-2424 was closed (and MAS has confirmed the behaviour.)
    !EeVValues = real(ePointer, double)
    EeVValues = ePointer
    call CAL_releaseMemory(ePointer)
    nullify(ePointer)	! unnecessary?

!EeVValues = piValues
!call warning('piToEvWorkaround',"ssclib:convertPiToEeVVector: The CAL call &
!&that is supposed to do this doesn't work (see soc SPR 2424). As a temporary &
!&workaround, it is being assumed that PI values are equal to eV values. Don't &
!&worry, PI does = eV; this is just a software engineering issue.")

  end function convertPiToEeVVectorSingle

  !---------------------------------------------------------------------
  function convertPiToEeVVectorDouble(piValues) result(EeVValues)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    real(double), intent(in) :: piValues(:)
    real(double)             :: EeVValues(size(piValues))

    !-------------------------------------------------------------------
    ! Start

    EeVValues = real(convertPiToEeVVectorSingle(real(piValues, single)), double)

  end function convertPiToEeVVectorDouble


  !---------------------------------------------------------------------
  function convertPhaRangeToEeV(timeStamp, phaRange) result(EeVrange)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument, ccdChipId, ccdNodeId
  ! and temperature before this routine is called.

    implicit none

    real(double), intent(in) :: timeStamp
    type(RangeT), intent(in) :: phaRange
    type(RangeT)             :: EeVrange

    !-------------------------------------------------------------------
    ! Start

    EeVrange%lower = convertPhaBoundToEeV(timeStamp, phaRange%lower)
    EeVrange%upper = convertPhaBoundToEeV(timeStamp, phaRange%upper)

  end function convertPhaRangeToEeV

  !---------------------------------------------------------------------
  function convertPhaRangesToEeV(timeStamp, phaRanges) result(EeVranges)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument, ccdChipId, ccdNodeId
  ! and temperature before this routine is called.

    implicit none

    real(double), intent(in) :: timeStamp
    type(RangeT), intent(in) :: phaRanges(:)
    type(RangeT)             :: EeVranges(size(phaRanges))

    !-------------------------------------------------------------------
    ! Start

    EeVranges%lower = convertPhaBoundsToEeV(timeStamp, phaRanges%lower)
    EeVranges%upper = convertPhaBoundsToEeV(timeStamp, phaRanges%upper)

  end function convertPhaRangesToEeV

  !---------------------------------------------------------------------
  function convertPhaBoundToEeV(timeStamp, phaBound) result(EeVBound)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument, ccdChipId, ccdNodeId
  ! and temperature before this routine is called.

    implicit none

    real(double), intent(in) :: timeStamp
    type(BoundT), intent(in) :: phaBound
    type(BoundT)             :: EeVBound

    !-------------------------------------------------------------------
    ! Start

    EeVBound%type = phaBound%type
    EeVBound%value = convertPhaToEeVScalar(timeStamp, phaBound%value)

  end function convertPhaBoundToEeV

  !---------------------------------------------------------------------
  function convertPhaBoundsToEeV(timeStamp, phaBounds) result(EeVBounds)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument, ccdChipId, ccdNodeId
  ! and temperature before this routine is called.

    implicit none

    real(double), intent(in) :: timeStamp
    type(BoundT), intent(in) :: phaBounds(:)
    type(BoundT)             :: EeVBounds(size(phaBounds))

    !-------------------------------------------------------------------
    ! Start

    EeVBounds%type = phaBounds%type
    EeVBounds%value = convertPhaToEeVVector(timeStamp, phaBounds%value)

  end function convertPhaBoundsToEeV

  !---------------------------------------------------------------------
  function convertPhaToEeVScalar(timeStamp, phaValue) result(EeVValue)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument, ccdChipId, ccdNodeId
  ! and temperature before this routine is called.

    implicit none

    real(double), intent(in) :: timeStamp
    real(double), intent(in) :: phaValue
    real(double)             :: EeVValue

    real(double) :: EeVValues(1)

    !-------------------------------------------------------------------
    ! Start

    EeVValues = convertPhaToEeVVector(timeStamp, (/phaValue/))
    EeVValue = EeVValues(1)

  end function convertPhaToEeVScalar

  !---------------------------------------------------------------------
  function convertPhaToEeVVector(timeStamp, phaValues) result(EeVValues)
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument, ccdChipId, ccdNodeId
  ! and temperature before this routine is called.

    implicit none

    real(double), intent(in) :: timeStamp
    real(double), intent(in) :: phaValues(:)
    real(double)             :: EeVValues(size(phaValues))

    real(single), pointer :: piValues(:) => null()

    !-------------------------------------------------------------------
    ! Start

    call CAL_gainCorrect(spread(timeStamp, 1, size(phaValues))&
      , real(phaValues, single), piValues)
    EeVValues = convertPiToEeVVectorDouble(real(piValues, double))
!    deallocate(piValues)
call CAL_releaseMemory(piValues)

  end function convertPhaToEeVVector


  !---------------------------------------------------------------------
  subroutine sampleEnergyRangeNormal(EeVlo, EeVhi, EeVsamples)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: EeVlo, EeVhi
    real(double), intent(out) :: EeVsamples(:)

    type(RangeT)  :: EeVrange

    !-------------------------------------------------------------------
    ! Start

    EeVrange%lower%type  = INCLUSIVE
    EeVrange%lower%value = EeVlo
    EeVrange%upper%type  = INCLUSIVE
    EeVrange%upper%value = EeVhi

    call sampleEnergyRangeScalar(EeVrange, EeVsamples)

  end subroutine sampleEnergyRangeNormal

  !---------------------------------------------------------------------
  subroutine sampleEnergyRangeScalar(EeVrange, EeVsamples)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in)  :: EeVrange
    real(double), intent(out) :: EeVsamples(:)

    integer(int32) :: i, numEsamples
    real(double) :: frac

    !-------------------------------------------------------------------
    ! Start

    numEsamples = size(EeVsamples)

    ! Construct the set of energy samples. Take samples such that their
    ! logs are distributed evenly across the full energy range.

    if  (EeVrange%lower%type == UNDEFINED&
    .or. EeVrange%upper%type == UNDEFINED)&
call error('undefinedERangeBounds',"ssclib:sampleEnergyRangeScalar: The energy range you asked to &
&sample is undefined.")

    if (EeVrange%lower%value <= 0.0)&
call error('zeroELowerBound',"ssclib:sampleEnergyRangeScalar: The lower bound&
& of the energy range must be greater than zero.")

!    logEmin = log(EeVrange%lower%value)
!    logEmax = log(EeVrange%upper%value)
    do i = 1, numEsamples
      frac = (i - 1) / real(numEsamples - 1)
!      EeVsamples(i) = exp(fracVal(logEmin, logEmax, frac))
EeVsamples(i) = (EeVrange%lower%value**(1.0-frac))*(EeVrange%upper%value**frac)
    end do

  end subroutine sampleEnergyRangeScalar

  !---------------------------------------------------------------------
  subroutine sampleEnergyRangeVector(EeVranges, EeVsamples)
  !---------------------------------------------------------------------
!**** NOTE this can be useful for lccorr.

    implicit none

    type(RangeT), intent(in)  :: EeVranges(:)
    real(double), intent(out) :: EeVsamples(:)

    integer(int32) ::&
      numRanges, numEsamples, sampleNum, rangeNum,&
      maxNumSamplesAllowedThisRange, numSamplesThisRange, i
    type(RangeT), allocatable :: logEeVranges(:)
    real(double) :: totalRemainingLogErange, approxSampleGap

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(EeVranges)
    if (numRanges <= 0)&
call error('zeroSizeRangesVector',"ssclib:sampleEnergyRangeVector: The input &
&vector argument EeVranges is of zero size.")

    allocate(logEeVranges(numRanges))

    if (.not. rangesAreWellFormed(EeVranges))&
call error('badlyFormedRangeVector',"ssclib:sampleEnergyRangeVector: The input &
&vector argument EeVranges are not 'well-formed'.")
    if ( EeVranges(1        )%lower%type == UNDEFINED&
    .or. EeVranges(numRanges)%upper%type == UNDEFINED)&
call error('rangeEndsUnbounded',"ssclib:sampleEnergyRangeVector: The input &
&vector argument EeVranges is unbounded at one or both ends.")

!vvv you may want to make the following more general in order to accommodate other transforms.
    if (any(EeVranges%lower%value <= 0.0))&
call error('zeroELowerBound',"ssclib:sampleEnergyRangeVector: The lower bound&
& of the lowest energy range must be greater than zero.")
    logEeVranges%lower%value = log(EeVranges%lower%value)
    logEeVranges%upper%value = log(EeVranges%upper%value)
!^^^

    numEsamples = size(EeVsamples)
    if (numEsamples < numRanges * 2)&
call error('tooFewSamples',"ssclib:sampleEnergyRangeVector: The number of samples &
&must be at least twice the number of ranges.")

    sampleNum = 0
    do rangeNum = 1, numRanges - 1
      totalRemainingLogErange = sum(logEeVranges(rangeNum:)%upper%value)&
        - sum(logEeVranges(rangeNum:)%lower%value) ! includes the present
      ! range, because of course we have not yet divided it up.

      approxSampleGap = totalRemainingLogErange / (numEsamples - sampleNum)

      maxNumSamplesAllowedThisRange = (numEsamples - sampleNum)&
        - 2 * (numRanges - rangeNum) ! total number available minus the
      ! number already used, minus 2 each (the minimum number possible
      ! per range) for each remaining range.

      numSamplesThisRange = 1 + nint((logEeVranges(rangeNum)%upper%value&
        - logEeVranges(rangeNum)%lower%value) / approxSampleGap)

      numSamplesThisRange = max(2, min(maxNumSamplesAllowedThisRange&
        , numSamplesThisRange))

      sampleNum = sampleNum + 1
      EeVsamples(sampleNum) = EeVranges(rangeNum)%lower%value
      do i = 2, numSamplesThisRange - 1
        sampleNum = sampleNum + 1
        EeVsamples(sampleNum)&
          = exp(fracVal(real(logEeVranges(rangeNum)%lower%value)&
          , real(logEeVranges(rangeNum)%upper%value)&
          , (i - 1) / real(numSamplesThisRange - 1)))
      end do
      sampleNum = sampleNum + 1
      EeVsamples(sampleNum) = EeVranges(rangeNum)%upper%value
    end do

    rangeNum = numRanges
    numSamplesThisRange = numEsamples - sampleNum
    sampleNum = sampleNum + 1
    EeVsamples(sampleNum) = EeVranges(rangeNum)%lower%value
    do i = 2, numSamplesThisRange - 1
      sampleNum = sampleNum + 1
      EeVsamples(sampleNum)&
        = exp(fracVal(real(logEeVranges(rangeNum)%lower%value)&
        , real(logEeVranges(rangeNum)%upper%value)&
        , (i - 1) / real(numSamplesThisRange - 1)))
    end do
    sampleNum = sampleNum + 1
    EeVsamples(sampleNum) = EeVranges(rangeNum)%upper%value

  end subroutine sampleEnergyRangeVector

  !---------------------------------------------------------------------
  subroutine getEffectiveArea(EeVsamples, effArea)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: EeVsamples(:)
    real(single), intent(out) :: effArea(size(EeVsamples))

    integer(int32) :: numEsamples, eI
!integer(int32) :: instrumentId

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Getting effective area curve.")

    numEsamples = size(EeVsamples)
    do eI = 1, numEsamples
      effArea(eI) = CAL_getEffectiveArea(EeVsamples(eI), 0.0d0, 0.0d0)
    end do

!call CAL_getState(instrument=instrumentId)
!print *, 'instrumentId, EMOS1, EMOS2, EPN:', instrumentId, EMOS1, EMOS2, EPN
!print *, 'effArea(50): ', effArea(50)
!print *, ''

  end subroutine getEffectiveArea

  !---------------------------------------------------------------------
  subroutine getFilterTransmission(EeVsamples, filterTrans)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: EeVsamples(:)
    real(single), intent(out) :: filterTrans(size(EeVsamples))

    integer(int32) :: numEsamples, eI

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Getting filter transmission curve.")

    numEsamples = size(EeVsamples)
    do eI = 1, numEsamples
      filterTrans(eI) = CAL_getFilterTransmission(EeVsamples(eI), 0.0d0, 0.0d0)
    end do

  end subroutine getFilterTransmission

  !---------------------------------------------------------------------
  subroutine getQe(EeVsamples, usePattern, qe, rawXiArg, rawYiArg)
  !---------------------------------------------------------------------
  ! NOTE! The cal MUST be set up for the correct instrument and CCD number
  ! before this function is called.
  !
  ! NOTE! This calculates the QE at pixel (rawXiArg, rawYiArg) of the
  ! currently selected CCD chip.
  !.....................................................................

    implicit none

    real(double),  intent(in)  :: EeVsamples(:)
    logical(bool), intent(in)  :: usePattern(:)
    real(single),  intent(out) :: qe(size(EeVsamples))
    integer(int16), intent(in), optional :: rawXiArg, rawYiArg

    real(single), allocatable :: qeArray(:,:)
    integer(int32) :: Ei

    !-------------------------------------------------------------------
    ! Start

    allocate(qeArray(size(usePattern),size(EeVsamples)))
    call constructQeArray(EeVsamples, qeArray, rawXiArg, rawYiArg)
    do Ei = 1, size(EeVsamples)
      qe(Ei) = sum(qeArray(:,Ei), usePattern)
    end do

  end subroutine getQe

  !---------------------------------------------------------------------
  subroutine constructQeArray(EeVsamples, qeArray, rawXiArg, rawYiArg, doDummy)
  !---------------------------------------------------------------------
  ! NOTE! The cal MUST be set up for the correct instrument and CCD number
  ! before this function is called.
  !
  !*** NOTE! This calculates the QE at pixel (50,100) of the currently
  ! selected CCD chip.
  !.....................................................................

    implicit none

    real(single),   intent(out)          :: qeArray(:,:)
    real(double),   intent(in)           :: EeVsamples(size(qeArray, 2))
    integer(int16), intent(in), optional :: rawXiArg, rawYiArg
    logical(bool),  intent(in), optional :: doDummy

    integer(int32) :: numPatterns, numEsamples, patternId, eI, instrumentId
    logical(bool) :: localDoDummy
    type(PatternFractionServerT) :: pFServer
    type(CurveT) :: totalQeCurve, patFracCurve
    integer(int16) :: rawXi, rawYi

    !-------------------------------------------------------------------
    ! Start

    if (present(doDummy)) then
      localDoDummy = doDummy
    else
      localDoDummy = .false.
    end if

    if (present(rawXiArg)) then
      rawXi = rawXiArg
    else
      rawXi = 50
    end if

    if (present(rawYiArg)) then
      rawYi = rawYiArg
    else
      rawYi = 50
    end if

    call message(AppMsg, Verbose, "  Constructing array (with &
      &pattern-energy axes) of samples of the quantum efficiency for &
      &this CCD.")

    numPatterns = size(qeArray, 1)
    numEsamples = size(qeArray, 2)

    if (localDoDummy) then
      call warning('constructQeArray', "ssclib:constructQeArray: This is a dummy routine! It &
        &doesn't return the real QE function.")

      qeArray = 1.0 / real(numPatterns)

    else
      call CAL_getState(instrument=instrumentId)
      if (instrumentId==EPN) then
        pFServer = CAL_getPatternFractionServer()
        call CAL_patternFractionAt(pFServer, rawXi, rawYi)
        totalQeCurve = CAL_getPatternFractionTotalQE(pFServer)
        patFracCurve = CAL_getPatternFractionInEnergy(pFServer, GRADE_SINGLE)
        do eI = 1, numEsamples
          qeArray(1, eI) = CAL_curveValueAt(totalQeCurve, EeVsamples(eI))&
            * CAL_curveValueAt(patFracCurve, EeVsamples(eI))
        end do

        patFracCurve = CAL_getPatternFractionInEnergy(pFServer, GRADE_DOUBLE)
        do eI = 1, numEsamples
          qeArray(2:5, eI) = CAL_curveValueAt(totalQeCurve, EeVsamples(eI))&
            * CAL_curveValueAt(patFracCurve, EeVsamples(eI)) / 4.0
        end do

      else ! instrumentId .ne. EPN
        do patternId = 0, numPatterns - 1
          do eI = 1, numEsamples
            qeArray(patternId + 1, eI)&
              = CAL_getQuantumEfficiency(EeVsamples(eI)&
              , rawXi, rawYi, patternId)
          end do
        end do
      end if
    end if

  end subroutine constructQeArray

  !---------------------------------------------------------------------
  subroutine constructQeArray_20040430(EeVsamples, qeArray, doDummy)
  !---------------------------------------------------------------------
  ! NOTE! The cal MUST be set up for the correct instrument and CCD number
  ! before this function is called.
  !
  !*** NOTE! This calculates the QE at pixel (50,100) of the currently
  ! selected CCD chip.
  !.....................................................................

    implicit none

    real(single),  intent(out)          :: qeArray(:,:)
    real(double),  intent(in)           :: EeVsamples(size(qeArray, 2))
    logical(bool), intent(in), optional :: doDummy

    integer(int32) :: numPatterns, numEsamples, patternId, eI, instrumentId
    logical(bool) :: localDoDummy
    type(PatternFractionServerT) :: pFServer
    type(CurveT) :: totalQeCurve, patFracCurve

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, ''
if (testFlag) print *, 'start constructQeArray'

    if (present(doDummy)) then
      localDoDummy = doDummy
    else
      localDoDummy = .false.
    end if

    call message(AppMsg, Verbose, "  Constructing array (with &
      &pattern-energy axes) of samples of the quantum efficiency for &
      &this CCD.")

    numPatterns = size(qeArray, 1)
    numEsamples = size(qeArray, 2)

    if (localDoDummy) then
      call warning('constructQeArray', "This is a dummy routine! It &
        &doesn't return the real QE function.")

      qeArray = 1.0 / real(numPatterns)
    else
      call CAL_getState(instrument=instrumentId)
      if (instrumentId==EPN) then
if (testFlag) print *, 'numPatterns=', numPatterns
if (numPatterns .ne. 5) call error('badQeArraySize',"")
if (testFlag) print *, 'aaa'

        pFServer = CAL_getPatternFractionServer()
if (testFlag) print *, 'bbb'
        call CAL_patternFractionAt(pFServer, 50_int16, 100_int16)
if (testFlag) print *, 'ccc'
        totalQeCurve = CAL_getPatternFractionTotalQE(pFServer)
if (testFlag) print *, 'ddd'
        patFracCurve = CAL_getPatternFractionInEnergy(pFServer, GRADE_SINGLE)
if (testFlag) print *, 'eee'
        do eI = 1, numEsamples
          qeArray(1, eI) = CAL_curveValueAt(totalQeCurve, EeVsamples(eI))&
            * CAL_curveValueAt(patFracCurve, EeVsamples(eI))
if (testFlag) print *, 'eI, g1pattFrac:', eI, CAL_curveValueAt(patFracCurve, EeVsamples(eI))
        end do
if (testFlag) print *, 'fff'

        patFracCurve = CAL_getPatternFractionInEnergy(pFServer, GRADE_DOUBLE)
if (testFlag) print *, 'ggg'
        do eI = 1, numEsamples
          qeArray(2:5, eI) = CAL_curveValueAt(totalQeCurve, EeVsamples(eI))&
            * CAL_curveValueAt(patFracCurve, EeVsamples(eI)) / 4.0
if (testFlag) print *, 'eI, g2pattFrac:', eI, CAL_curveValueAt(patFracCurve, EeVsamples(eI))
        end do
if (testFlag) print *, 'hhh'

if (testFlag) then
  print *, ''
  do eI = 1, numESamples
    print *, 'eI, EeVsamples(eI), sum(qeArray(:, eI)):', eI, EeVsamples(eI), sum(qeArray(:, eI))
  end do
  print *, ''
end if


      else ! instrumentId .ne. EPN
        do patternId = 0, numPatterns - 1
          do eI = 1, numEsamples
            qeArray(patternId + 1, eI)&
              = CAL_getQuantumEfficiency(EeVsamples(eI)&
              , 50_int16, 100_int16, patternId)
          end do
        end do
      end if
    end if
if (testFlag) print *, 'end constructQeArray'
if (testFlag) print *, ''

  end subroutine constructQeArray_20040430

  !---------------------------------------------------------------------
  subroutine calcQeMap(qeMap, pattern, EeV1, EeV2, mask)
  !---------------------------------------------------------------------
  ! The 1st/2nd indices of the map array are assumed to be equal to rawx/y.


!**** generalize with eSpectrum rather than EeV1 and EeV2.


    implicit none

    real(single),   intent(out)          :: qeMap(:,:)
    integer(int32), intent(in)           :: pattern
    real(double),   intent(in)           :: EeV1
    real(double),   intent(in), optional :: EeV2
    logical(bool),  intent(in), optional :: mask(size(qeMap, 1), size(qeMap, 2))

    real(double) :: EeV, EeVlo, EeVhi
    logical(bool), allocatable :: localMask(:,:)
    integer(int16) :: xi, yi, xSize, ySize

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, '  Calculating Quantum Efficiency map &
      &for this CCD.')

    xSize = size(qeMap, 1)
    ySize = size(qeMap, 2)
    allocate(localMask(xSize,ySize))

    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if

    where(localMask) qeMap = 0.0

    if (present(EeV2)) then
      EeVlo = EeV1
      EeVhi = EeV2
      do xi = 1, xSize
        do yi = 1, ySize
        if (.not. localMask(xi, yi)) cycle
          qeMap(xi, yi) = CAL_getQuantumEfficiency(EeVlo, EeVhi, xi, yi&
          , pattern)
        end do
      end do

    else
      EeV = EeV1
      do xi = 1, xSize
        do yi = 1, ySize
        if (.not. localMask(xi, yi)) cycle
          qeMap(xi, yi) = CAL_getQuantumEfficiency(EeV, xi, yi, pattern)
        end do
      end do
    end if

  end subroutine calcQeMap

  !---------------------------------------------------------------------
  subroutine constructPsfCubeTelcoord(EeVsamples, srcThetaArcsec, srcPhi&
  , desiredEnergyFraction, psfSamplesCubeTelcoord, psfPixelSizeMm&
  , actualEnergyFraction)
  !---------------------------------------------------------------------
!*** NOTE! at present this assumes accuracylevel = medium (else would be able to set pixel size).

    implicit none

    real(double),      intent(in)            :: EeVsamples(:),&
                                                srcThetaArcsec,&
                                                srcPhi,&
                                                desiredEnergyFraction
    real(single),      pointer               :: psfSamplesCubeTelcoord(:,:,:)
    type(PsfBinSizeT), intent(out)           :: psfPixelSizeMm
    real(double),      intent(out), optional :: actualEnergyFraction

    type(PsfValidityRangesT) :: validPsfRanges
    type(PsfT) :: psfHandle
    type(PsfWindowT) :: psfWindowMm
    integer :: numEsamples, eI
    real(single), pointer :: psfImageAtEnergy(:,:) => null()
    real(double) :: localEnergyFraction, effectiveSrcThetaArcsec
    character(5) :: thetaArcsecStr
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa exposure:constructPsfCubeTelcoord'

    call message(AppMsg, Verbose, "Constructing cube (with &
      &energy-detx-dety axes) of samples of the PSF.")

    numEsamples = size(EeVsamples)

    validPsfRanges = CAL_psfValidityRanges()

    if  (any(EeVsamples < validPsfRanges%eMin&
    .or. EeVsamples > validPsfRanges%eMax))&
call error('energyOutOfRange',"ssclib:constructPsfCubeTelcoord: At least one of &
&the energy samples is outside the range returned by CAL_psfValidityRanges()")

    if (srcThetaArcsec < validPsfRanges%thetaMin)&
call error('thetaOutOfRange',"ssclib:constructPsfCubeTelcoord: The source theta value &
&is outside the range returned by CAL_psfValidityRanges()")

    if (srcPhi < validPsfRanges%phiMin&
    .or. srcPhi > validPsfRanges%phiMax)&
call error('phiOutOfRange',"ssclib:constructPsfCubeTelcoord: The source phi value &
&is outside the range returned by CAL_psfValidityRanges()")

    if (srcThetaArcsec > validPsfRanges%thetaMax) then
      thetaArcsecStr = intTostr(ceiling(srcThetaArcsec)) ! in ssc_misc.
!      write(thetaArcMinStr, '(i4)') ceiling(srcThetaArcsec / 60.0)

!print *, 'srcThetaArcsec, ceiling(srcThetaArcsec), thetaArcsecStr'&
!, srcThetaArcsec, ceiling(srcThetaArcsec), thetaArcsecStr, trim(thetaArcsecStr)

      call warning('thetaOutOfRange',"ssclib:constructPsfCubeTelcoord: The angle &
        &between the source position and the optical axis is "//trim(thetaArcsecStr)&
        //" arcsec. However 900 arcsec is currently the largest angle &
        &for which a PSF is available from the calibration files. The &
        &900-arcsec PSF will be used for this source.")

      effectiveSrcThetaArcsec = validPsfRanges%thetaMax&
        * (1.0d0 - epsilon(srcThetaArcsec))
    else
      effectiveSrcThetaArcsec = srcThetaArcsec
    end if

    ! Get size of psf arrays (use the lowest energy for all sizes because the psf is widest there):

    psfHandle = CAL_getPSF(EeVsamples(1), effectiveSrcThetaArcsec, srcPhi)

    psfPixelSizeMm = CAL_psfBinSize(psfHandle)
    psfWindowMm = CAL_psfEnboxedEnergy(psfHandle, desiredEnergyFraction)
    call CAL_psfImage(psfHandle, psfWindowMm, psfImageAtEnergy)
    localEnergyFraction = sum(psfImageAtEnergy) * psfPixelSizeMm%binX&
      * psfPixelSizeMm%binY

    allocate(psfSamplesCubeTelcoord(numEsamples&
      , lbound(psfImageAtEnergy, 1):ubound(psfImageAtEnergy, 1)&
      , lbound(psfImageAtEnergy, 2):ubound(psfImageAtEnergy, 2)))

    psfSamplesCubeTelcoord(1,:,:) = psfImageAtEnergy

    do eI = 2, numEsamples
      psfHandle = CAL_getPSF(EeVsamples(eI), effectiveSrcThetaArcsec, srcPhi)

      call CAL_psfImage(psfHandle, psfWindowMm, psfImageAtEnergy)
      psfSamplesCubeTelcoord(eI,:,:) = psfImageAtEnergy
    end do

    call CAL_releaseMemory(psfImageAtEnergy)
    if (present(actualEnergyFraction))&
      localEnergyFraction = actualEnergyFraction

if (testFlag) print *, '  zzz exposure:constructPsfCubeTelcoord'

  end subroutine constructPsfCubeTelcoord


! Remaining routines can stay here as well as anywhere:


  !---------------------------------------------------------------------
  subroutine constructVignettingCube(EeVsamples, thetaArcsecSamples&
  , phiSamples, vigSamplesCube, doDummy)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: EeVsamples(:),&
                                 thetaArcsecSamples(:),&
                                 phiSamples(:)
    real(single), intent(out) :: vigSamplesCube(&
                                   size(EeVsamples),&
                                   size(thetaArcsecSamples),&
                                   size(phiSamples)&
                                 )
    logical, intent(in), optional :: doDummy

    integer(int32) ::&
      numEsamples, numThetaSamples, numPhiSamples, thetaI, phiI, eI
    real(double) ::&
      thetaArcsec,&
      maxThetaArcsec = 1199 ! should be comfortably above
    ! the FOV radius, which is 900, yet below the vignet limit of 1200.
real :: dummyR

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Constructing cube (with &
      &energy-theta-phi axes) of samples of the vignetting function.")

if (present(doDummy)) then
  call warning('constructVignettingCube', "This is a dummy routine! It doesn't return the real vig function.")
end if

    numEsamples     = size(EeVsamples)
    numThetaSamples = size(thetaArcsecSamples)
    numPhiSamples   = size(phiSamples)

    do thetaI = 1, numThetaSamples
      if (thetaArcsecSamples(thetaI) <= maxThetaArcsec) then
        thetaArcsec = thetaArcsecSamples(thetaI)
      else
        thetaArcsec = maxThetaArcsec
        call warning('thetaOverLimit',"")
      end if

      do phiI = 1, numPhiSamples
        do eI = 1, numEsamples
if (present(doDummy)) then
  dummyR = thetaArcsecSamples(thetaI) / maxThetaArcsec / (1.0 - 0.667 * (eI / real(numEsamples)))
  vigSamplesCube(eI, thetaI, phiI) = 1.0/(1 + dummyR*dummyR)

else
          vigSamplesCube(eI, thetaI, phiI)&
            = CAL_getVignettingFactor(&
              EeVsamples(eI),&
              thetaArcsec,&
              phiSamples(phiI)&
            )
end if
        end do
      end do
    end do

!*** NOTE! Occasionally the vig function can be more than 1. Apparently this is due to wiggles in the interpolating polynomials in the cal. Lop these protrusions off as a workaround.
    where(vigSamplesCube > 1.0) vigSamplesCube = 1.0

  end subroutine constructVignettingCube

  !---------------------------------------------------------------------
  subroutine calcVigAtSourcePosWithCube(srcThetaArcsec, srcPhi, thetaSamples&
  , phiSamples, vigSamplesCube, vigEfunctionAtSrc)
  !---------------------------------------------------------------------
  ! The fundamental input here is the 3-d array vigSamplesCube(). This contains a set of samples of the vignetting function for the specified instrument. The first index of this cube runs over energy, the second over theta (angle away from the optical axis in the TELCOORD system) and the third over phi, the azimuthal angle.
  !
  ! The subroutine uses cubic splines to interpolate this cube of vignet samples at a single (theta,phi) pair. The result is returned in the vector vigEfunctionAtSrc() which contains these interpolated samples of the vignetting function at the same range of energy samples as is implicit in the first axis of the input cube.
  !
  ! NOTE:
  !	(i) It is required that the cal is set up with the appropriate instrument, ccdChipId, ccdNodeId and ccdModeId beforehand.
  !	(ii) This routine calls CAL_SetState(randomize=.false.).

    implicit none

    real(double), intent(in)  :: srcThetaArcsec, srcPhi
    real(single), intent(in)  :: thetaSamples(:), phiSamples(:),&
                                 vigSamplesCube(:,:,:)
    real(single), intent(out) :: vigEfunctionAtSrc(:)

    integer(int32) :: eI
    integer(int16) ::&
      numEsamples, numThetaSamples, numPhiSamples,&
!      loThetaI, hiThetaI,&
      loPhiI, hiPhiI
    real(single), allocatable :: vig2ndDerivWrtTheta(:)
    real(single) :: vigLoPhi, vigHiPhi, phiFrac

    !-------------------------------------------------------------------
    ! Start

    numEsamples   = size(vigSamplesCube, 1)
    numThetaSamples = size(thetaSamples)
    numPhiSamples   = size(phiSamples)
    ! NOTE these all must be >= 2.

    allocate(vig2ndDerivWrtTheta(numThetaSamples))

    if  (numEsamples   .ne. size(vigEfunctionAtSrc)&
    .or. numThetaSamples .ne. size(vigSamplesCube, 2)&
    .or. numPhiSamples   .ne. size(vigSamplesCube, 3)) then
call error('exposure_bug0',"ssclib:calcVigAtSourcePosWithCube: Bug - please contact the code developer.")
    end if

    ! Find the applicable range of theta and phi indices:

    if  (srcThetaArcsec < thetaSamples(1)&
    .or. srcThetaArcsec > thetaSamples(numThetaSamples)) then
call error('srcThetaOutOfLimits',"ssclib:calcVigAtSourcePosWithCube: The input &
&argument srcThetaArcsec is not within the range of thetaSamples.")
    end if

    if  (srcPhi < phiSamples(1)&
    .or. srcPhi > phiSamples(numPhiSamples)) then
call error('srcPhiOutOfLimits',"ssclib:calcVigAtSourcePosWithCube: The input &
&argument srcPhi is not within the range of phiSamples.")
    end if

!    loThetaI =  largestTrueIndex(thetaSamples <= srcThetaArcsec)
!    hiThetaI = smallestTrueIndex(thetaSamples >= srcThetaArcsec)
!
!    if (loThetaI == hiThetaI) then
!      if (hiThetaI < numThetaSamples) then
!        hiThetaI = hiThetaI + 1
!      else
!        loThetaI = loThetaI - 1
!      end if
!    end if

    loPhiI =  largestTrueIndex(logical(phiSamples <= srcPhi, bool))
    hiPhiI = smallestTrueIndex(logical(phiSamples >= srcPhi, bool))

    if (loPhiI == hiPhiI) then
      if (hiPhiI < numPhiSamples) then
        hiPhiI = hiPhiI + 1
      else
        loPhiI = loPhiI - 1
      end if
    end if

    phiFrac = (srcPhi - phiSamples(loPhiI))&
      / (phiSamples(hiPhiI) - phiSamples(loPhiI))

    do eI = 1, numEsamples
!      call calcCubicSplines(thetaSamples&
!        , vigSamplesCube(eI, :, loPhiI)&
!        , loBoundType='fixed1st', loBoundValue=0.0&
!        , hiBoundType='fixed2nd', hiBoundValue=0.0& ! only an approx,
!                                                    ! don't know the
!                                                    ! correct outer BC.
!        , y2=vig2ndDerivWrtTheta&
!      )
      call calcY2ForCubicSplines(thetaSamples&
        , vigSamplesCube(eI, :, loPhiI)&
        , vig2ndDerivWrtTheta&
        , loBoundType=SPLINE_FIXED_1ST, loBoundValue=0.0&
        , hiBoundType=SPLINE_FIXED_2ND, hiBoundValue=0.0& ! only an approx,
                                                    ! don't know the
                                                    ! correct outer BC.
      )

!      call cubicSplineInterpolate(thetaSamples&
!        , vigSamplesCube(eI, :, loPhiI)&
!        , vig2ndDerivWrtTheta, real(srcThetaArcsec)&
!        , vigLoPhi&
!      )
      vigLoPhi = cubicSplineInterpolate(thetaSamples&
        , vigSamplesCube(eI, :, loPhiI)&
        , vig2ndDerivWrtTheta, real(srcThetaArcsec)&
      )

!      call calcCubicSplines(thetaSamples&
!        , vigSamplesCube(eI, :, hiPhiI)&
!        , loBoundType='fixed1st', loBoundValue=0.0&
!        , hiBoundType='fixed2nd', hiBoundValue=0.0& ! only an approx,
!                                                    ! don't know the
!                                                    ! correct outer BC.
!        , y2=vig2ndDerivWrtTheta&
!      )
      call calcY2ForCubicSplines(thetaSamples&
        , vigSamplesCube(eI, :, hiPhiI)&
        , vig2ndDerivWrtTheta&
        , loBoundType=SPLINE_FIXED_1ST, loBoundValue=0.0&
        , hiBoundType=SPLINE_FIXED_2ND, hiBoundValue=0.0& ! only an approx,
                                                    ! don't know the
                                                    ! correct outer BC.
      )

!      call cubicSplineInterpolate(thetaSamples&
!        , vigSamplesCube(eI, :, hiPhiI)&
!        , vig2ndDerivWrtTheta, real(srcThetaArcsec)&
!        , vigHiPhi&
!      )
      vigHiPhi = cubicSplineInterpolate(thetaSamples&
        , vigSamplesCube(eI, :, hiPhiI)&
        , vig2ndDerivWrtTheta, real(srcThetaArcsec)&
      )

      vigEfunctionAtSrc(eI) = fracVal(vigLoPhi, vigHiPhi, phiFrac)
    end do

  end subroutine calcVigAtSourcePosWithCube

  !---------------------------------------------------------------------
  subroutine calcVigAtSourcePosNoCube(srcThetaArcsec, srcPhi, EeVsamples&
  , vigEfunctionAtSrc)
  !---------------------------------------------------------------------
  ! NOTE:
  !	(i) It is required that the cal is set up with the appropriate instrument, ccdChipId, ccdNodeId and ccdModeId beforehand.
  !	(ii) This routine calls CAL_SetState(randomize=.false.).

    implicit none

    real(double), intent(in)  :: srcThetaArcsec, srcPhi
    real(double), intent(in)  :: EeVSamples(:)
    real(single), intent(out) :: vigEfunctionAtSrc(size(EeVSamples))

    integer(int32) :: eI

!****** add test and workaround for theta > max!!!!

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Sampling the vignetting function.")

    do eI = 1, size(EeVSamples)
      vigEfunctionAtSrc(eI) = CAL_getVignettingFactor(EeVsamples(eI)&
        , srcThetaArcsec, srcPhi)
    end do

!*** NOTE! Occasionally the vig function can be more than 1. Apparently this is due to wiggles in the interpolating polynomials in the cal. Lop these protrusions off as a workaround.
    where(vigEfunctionAtSrc > 1.0) vigEfunctionAtSrc = 1.0

  end subroutine calcVigAtSourcePosNoCube

  !---------------------------------------------------------------------
  subroutine multiplyByVigFunction(ccdMap, EeVsamples, thetaSamples&
  , phiSamples, vigSamplesCube, EeVfunctions, EeVranges)
  !---------------------------------------------------------------------
  ! NOTE:
  !	(i) It is required that the cal is set up with the appropriate instrument, ccdChipId, ccdNodeId and ccdModeId beforehand.
  !	(ii) This routine calls CAL_SetState(randomize=.false.).

    implicit none

    real(single), intent(inout) :: ccdMap(:,:)
    real(double), intent(in)    :: EeVsamples(:),&
                                   thetaSamples(:),&
                                   phiSamples(:)
    real(single), intent(in)    :: vigSamplesCube(&
                                     size(EeVsamples),&
                                     size(thetaSamples),&
                                     size(phiSamples)&
                                   ),&
                                   EeVfunctions(size(EeVsamples))
    type(RangeT), intent(in)    :: EeVranges(:)

    integer(int32) :: numRows, row
    integer(int16) ::&
      numEsamples, numThetaSamples, numPhiSamples, numRanges,&
      xSize, ySize, loPhiI, hiPhiI,&
      phiI, thetaI, i, j
    logical(bool), allocatable :: mask(:,:)
    integer(int16), allocatable ::&
      tempInt16Array(:,:), rawXvector(:), rawYvector(:)
!    integer(int16), pointer :: indices(:) => null()
    real(double), pointer :: cc2X(:) => null(), cc2Y(:) => null(), cc2Z(:) => null()
    real(double), pointer :: thetaVector(:) => null(), phiVector(:) => null()
    real(single), allocatable ::&
      vigSubMap(:,:), vig2ndDerivWrtTheta(:,:), vigVector(:)
    real(single) :: thetaArcsec, phi, vigLeft, vigRight, phiFrac
!    integer(int32) :: specialRow

    !-------------------------------------------------------------------
    ! Start

    numEsamples   = size(EeVsamples)
    numThetaSamples = size(thetaSamples)
    numPhiSamples   = size(phiSamples)
    ! NOTE these all must be >= 2.

!    if  (numEsamples   .ne. size(vigSamplesCube, 1)&
!    .or. numThetaSamples .ne. size(vigSamplesCube, 2)&
!    .or. numPhiSamples   .ne. size(vigSamplesCube, 3)) then
!call error('bug_exp9', "Array sizes don't match up. Shouldn't happen.")
!    end if

    numRanges = size(EeVranges)
    if (numRanges <= 0) then
      call warning('noEnergyRanges',"ssclib:multiplyByVigFunction: The input &
        &vector argument EeVranges has zero size. The CCD map must therefore &
        &be set to zero.")
      ccdMap = 0.0
  return
    end if

  if (count(ccdMap > 0.0) <= 0) return

    xSize = size(ccdMap, 1)
    ySize = size(ccdMap, 2)
    allocate(mask(xSize, ySize))

!call dumpImageToFits(ccdMap, 'mapDump.ds')
    mask = (ccdMap > 0.0)
    numRows = count(mask)

    ! Calc theta, phi values for the centres of all pixels which
    ! have values > 0:

    allocate(tempInt16Array(xSize, ySize))
    allocate(rawXvector(numRows), rawYvector(numRows))

!    indices => indicize(xSize) ! in ssc_misc
!    tempInt16Array = spread(indices, 2, ySize)
tempInt16Array = spread((/(i, i=1,xSize)/), 2, ySize)
!    deallocate(indices)
    rawXvector = pack(tempInt16Array, mask)

!    indices => indicize(ySize) ! in ssc_misc
!    tempInt16Array = spread(indices, 1, xSize)
tempInt16Array = spread((/(i, i=1,ySize)/), 1, xSize)
!    deallocate(indices)
    rawYvector = pack(tempInt16Array, mask)

    deallocate(tempInt16Array)

!i = smallestTrueIndex((rawXvector == 90).and.(rawYvector == 351))
!print *, any((rawXvector == 90).and.(rawYvector == 351))
!print *, any(rawXvector == 90)
!print *, any(rawYvector == 351)
!print *, smallestTrueIndex(rawXvector == 90)
!print *, smallestTrueIndex(rawYvector == 351)
!print *, 'i: ', i
!print *, 'i, rawXvector(i), rawYvector(i): ', i, rawXvector(i), rawYvector(i)


    call CAL_SetState(randomize=.false.)
    call CAL_rawXY2mm(rawXvector, rawYvector, cc2X, cc2Y, cc2Z&
      , coordSysType=CAMCOORD2)

!!!    deallocate(rawXvector, rawYvector)
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, thetaVector, phiVector)
!!!    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

!if (count((rawXvector == 90).and.(rawYvector == 351)) == 1) then
!  specialRow = smallestTrueIndex((rawXvector == 90).and.(rawYvector == 351))
!row = specialRow
!  print *, 'row: ', row
!  print *, 'rawXvector(row), rawYvector(row): ', rawXvector(row), rawYvector(row)
!  print *, 'cc2X(row-1), cc2Y(row-1): ', cc2X(row-1), cc2Y(row-1)
!  print *, 'thetaVector(row-1), phiVector(row-1): ', thetaVector(row-1), phiVector(row-1)
!else
!  print *, "count((rawXvector == 90).and.(rawYvector == 351)) .ne. 1!"
!end if

!print *, ''
!do row = 1, numRows
!  print *, row, rawXvector(row), rawYvector(row), cc2X(row-1), cc2Y(row-1), thetaVector(row-1), phiVector(row-1)
!end do

deallocate(rawXvector, rawYvector)
call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

    ! Find the applicable range of theta and phi indices:


    if  (minval(thetaVector) < thetaSamples(1)) then
print *, 'minval(thetaVector), thetaSamples(1): ', minval(thetaVector), thetaSamples(1)
call error('exposure_bug1',"ssclib:multiplyByVigFunction: Should not happen. Please contact the code developer.")
    end if

    if  (minval(phiVector) < phiSamples(1)&
    .or. maxval(phiVector) > phiSamples(numPhiSamples)) then
print *, 'minval(phiVector), phiSamples(1): ', minval(phiVector), phiSamples(1)
print *, 'maxval(phiVector), phiSamples(n): ', maxval(phiVector), phiSamples(numThetaSamples)
call error('exposure_bug2',"ssclib:multiplyByVigFunction: Should not happen. Please contact the code developer.")
    end if


    loPhiI =  largestTrueIndex(logical(phiSamples <= minval(phiVector), bool))
    hiPhiI = smallestTrueIndex(logical(phiSamples >= maxval(phiVector), bool))

    if (loPhiI == hiPhiI) then ! unlikely as it implies that
      ! there is only 1 value of phi among the ccd pixels, and
      ! also that that value lies exactly on one of the samples
      ! in phiSamples. Nevertheless, could happen I suppose.

      if (hiPhiI < numPhiSamples) then
        hiPhiI = hiPhiI + 1
      else
        loPhiI = loPhiI - 1
      end if
    end if

    ! Construct the vig function sub-map by integrating the vig
    ! function cube over the specified energy ranges, and also preload
    ! the 2nd derivatives for the cubic spline:


!print *, 'integral at (26,7): ', trapezoidSum(EeVsamples, vigSamplesCube(:,26,7)*EeVfunctions)
    allocate(vigSubMap(numThetaSamples, loPhiI:hiPhiI))
    allocate(vig2ndDerivWrtTheta(numThetaSamples, loPhiI:hiPhiI))
    do phiI = loPhiI, hiPhiI
      do thetaI = 1, numThetaSamples
        vigSubMap(thetaI, phiI) = integrateRanges(real(EeVsamples, single)&
          , vigSamplesCube(:,thetaI,phiI) * EeVfunctions, EeVranges)
!print *, 'thetaI, phiI, thetaSamples(thetaI), phiSamples(phiI), vigSubMap(thetaI, phiI): '&
!, thetaI, phiI, thetaSamples(thetaI), phiSamples(phiI), vigSubMap(thetaI, phiI)
      end do

!      call calcCubicSplines(real(thetaSamples, single), vigSubMap(:, phiI)&
!        , loBoundType='fixed1st', loBoundValue=0.0&
!        , hiBoundType='fixed2nd', hiBoundValue=0.0& ! only an approx,
!                                                    ! don't know the
!                                                    ! correct outer BC.
!        , y2=vig2ndDerivWrtTheta(:, phiI))
      call calcY2ForCubicSplines(real(thetaSamples, single), vigSubMap(:, phiI)&
        , vig2ndDerivWrtTheta(:, phiI)&
        , loBoundType=SPLINE_FIXED_1ST, loBoundValue=0.0&
        , hiBoundType=SPLINE_FIXED_2ND, hiBoundValue=0.0& ! only an approx,
                                                    ! don't know the
                                                    ! correct outer BC.
        )
    end do
!call dumpImageToFits(vigSubMap, 'vigSubMapDump.ds')

    ! Now interpolate the vignetting map for all pixels of ccdMap
    ! which have a value >0:

    allocate(vigVector(numRows))
    vigVector = 0.0
    do row = 1, numRows
      thetaArcsec = thetaVector(row - 1)! type conversion from
    if (thetaArcsec > thetaSamples(numThetaSamples)) cycle

      phi = phiVector(row - 1)		! double to single.

      ! Get the indices of the lower LH corner of the theta-phi
      ! rectangle in theta-phi space that contains both thetaArcsec
      ! and phi:

      i = largestTrueIndex(logical(thetaSamples <= thetaArcsec, bool))
      if (i == numThetaSamples) i = numThetaSamples - 1
      j = largestTrueIndex(logical(phiSamples <= phi, bool))
      if (j == hiPhiI) j = hiPhiI - 1

      ! Get the cubic-spline interpolation at each phi:

!      call cubicSplineInterpolate(real(thetaSamples(i:i+1), single), vigSubMap(i:i+1, j)&
!        , vig2ndDerivWrtTheta(i:i+1, j),   thetaArcsec, vigLeft)
!      call cubicSplineInterpolate(real(thetaSamples(i:i+1), single), vigSubMap(i:i+1, j+1)&
!        , vig2ndDerivWrtTheta(i:i+1, j+1), thetaArcsec, vigRight)
      vigLeft = cubicSplineInterpolate(real(thetaSamples(i:i+1), single), vigSubMap(i:i+1, j)&
        , vig2ndDerivWrtTheta(i:i+1, j),   thetaArcsec&
      )
      vigRight = cubicSplineInterpolate(real(thetaSamples(i:i+1), single), vigSubMap(i:i+1, j+1)&
        , vig2ndDerivWrtTheta(i:i+1, j+1), thetaArcsec&
      )

      ! Finally do a linear interpolation between these two values:

      phiFrac = (phi - phiSamples(j)) / (phiSamples(j+1) - phiSamples(j))
      vigVector(row) = fracVal(vigLeft, vigRight, phiFrac)

!if (row == specialRow) then
!  print *, 'row, thetaArcsec, phi, thetaIlo, phiLo, vigAtThetaIlo&
!    &, vigAtThetaIloPlus1, phiFrac, vigVector(row): ', row, thetaArcsec&
!    , phi, i, j, vigLeft, vigRight, phiFrac, vigVector(row)
!end if
    end do

    deallocate(vigSubMap, vig2ndDerivWrtTheta)

    if (any(thetaVector > thetaSamples(numThetaSamples)))&
      call warning('thetaOutsideRange',"ssclib:multiplyByVigFunction:")

    call CAL_releaseMemory(thetaVector, phiVector)

    ! Finally, reshape vigVector to an array and fold it into ccdMap:

    ccdMap = ccdMap * unpack(vigVector, mask, ccdMap)
    deallocate(vigVector)

  end subroutine multiplyByVigFunction

  !---------------------------------------------------------------------
  subroutine calcVigMap(EeVsamples, EeVfunction, EeVranges, vigMap)
  !---------------------------------------------------------------------
  ! NOTE! This requires that the cal be correctly set up first for instrument, ccdModeId, ccdNum and ccdNodeId.

    implicit none

    real(double), intent(in)  :: EeVsamples(:)
    real(single), intent(in)  :: EeVfunction(size(EeVsamples))
    type(RangeT), intent(in)  :: EeVranges(:)
    real(single), intent(out) :: vigMap(:,:)

    integer(int16) ::&
      numEsamples, numRanges, xSize, ySize, arrayShape(2), eI
    integer(int16) :: i
    integer(int32) :: numRows, vectorShape(1), row
    integer(int16), allocatable ::&
      tempInt16Array(:,:), rawXvector(:), rawYvector(:)
    real(double), pointer ::&
       cc2X(:) => null(), cc2Y(:) => null(), cc2Z(:) => null()&
     , thetaVector(:) => null(), phiVector(:) => null()
    real(double) :: thetaArcsec, phi
    real(single), allocatable :: vigEfunction(:), vigVector(:)

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, '  Calculating full vignetting map &
      &for this CCD.')

    xSize = size(vigMap, 1)
    ySize = size(vigMap, 2)

    numEsamples = size(EeVsamples)
    if (numEsamples < 2)&
call error('tooFewEnergySamples',"ssclib:calcVigMap: There must be at least &
&2 energy samples for interpolation to be possible.")

    allocate(vigEfunction(numEsamples))

    numRanges = size(EeVranges)
    if (numRanges <= 0) then
      call warning('noEnergyRanges',"ssclib:calcVigMap:  The input &
        &vector argument EeVranges has zero size. The vignetting map must therefore &
        &be set to zero.")
      vigMap = 0.0
  return
    end if

    numRows = xSize * ySize
    vectorShape(1) = numRows
    arrayShape(1) = xSize
    arrayShape(2) = ySize

    ! Calc theta, phi values for the centres of all pixels which
    ! have values > 0:

    allocate(tempInt16Array(xSize, ySize))
    allocate(rawXvector(numRows), rawYvector(numRows))

    tempInt16Array = spread((/(i, i=1,xSize)/), 2, ySize)
    rawXvector = reshape(tempInt16Array, vectorShape)

    tempInt16Array = spread((/(i, i=1,ySize)/), 1, xSize)
    rawYvector = reshape(tempInt16Array, vectorShape)

    deallocate(tempInt16Array)

    call CAL_SetState(randomize=.false.)
    call CAL_rawXY2mm(rawXvector, rawYvector, cc2X, cc2Y, cc2Z&
      , coordSysType=CAMCOORD2)

    deallocate(rawXvector, rawYvector)
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, thetaVector, phiVector)
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

    allocate(vigVector(numRows))
    do row = 1, numRows
!      call printEvery(row, numRows, fraction=0.01)
      errStr = sprintEvery(row, numRows, fraction=0.01)
      ! in ssclib/ssc_misc.
      if (errStr .ne. '') call message(appMsg, verbose)

      thetaArcsec = thetaVector(row - 1)! type conversion from
      phi = phiVector(row - 1)		! double to single.

      do eI = 1, numEsamples
        vigEfunction(eI) = CAL_getVignettingFactor(EeVsamples(eI)&
          , thetaArcsec, phi)
      end do

      vigVector(row) = integrateRanges(real(EeVsamples, single), vigEfunction * EeVfunction&
        , EeVranges)
    end do

    call CAL_releaseMemory(thetaVector, phiVector)

    ! Finally, reshape vigVector to an array:

    vigMap = reshape(vigVector, arrayShape)
    deallocate(vigVector)

  end subroutine calcVigMap

  !---------------------------------------------------------------------
  subroutine interpolateVigMap(phi, tanThetaArcsec, EeVSamples, thetaSamples&
  , phiSamples, vigSamplesCube, EeVRanges, vigMap)
  !---------------------------------------------------------------------
  ! The fundamental input here is the 3-d array vigSamplesCube(). This contains a set of samples of the vignetting function for the specified instrument. The first index of this cube runs over energy, the second over theta (angle away from the optical axis in the TELCOORD system) and the third over phi, the azimuthal angle.
  !
  ! The subroutine uses cubic splines to interpolate this cube of vignet samples at an array of theta/phi values. At each of these points the cube is numerically integrated in the energy direction. The result is returned in the array vigMap().
!*** must alter the description somewhat. IMS 20040903
  !
  ! NOTE:
  !	(i) It is required that the cal is set up with the appropriate instrument, ccdChipId, ccdNodeId and ccdModeId beforehand.
  !	(ii) This routine calls CAL_SetState(randomize=.false.).

    implicit none

    real(double), intent(in)  :: phi(:,:),&
                                 tanThetaArcsec(size(phi,1),size(phi,2)),&
                                 EeVSamples(:),&
                                 thetaSamples(:),&
                                 phiSamples(:)
    real(single), intent(in)  :: vigSamplesCube(&
                                   size(EeVSamples),&
                                   size(thetaSamples),&
                                   size(phiSamples))
    type(RangeT), intent(in)  :: EeVRanges(:)
    real(single), intent(out) :: vigMap(size(phi,1),size(phi,2))

    integer(int16) ::&
      numThetaSamples, numPhiSamples, ti, pi
    real(single), allocatable :: vigSamples2d(:,:)

    !-------------------------------------------------------------------
    ! Start:

    numThetaSamples = size(thetaSamples)
    numPhiSamples   = size(phiSamples)

    allocate(vigSamples2d(numThetaSamples, numPhiSamples))

    do ti = 1, numThetaSamples
      do pi = 1, numPhiSamples
        vigSamples2d(ti, pi) = integrateRanges(real(EeVSamples, single)&
          , vigSamplesCube(:,ti,pi), EeVRanges)
      end do
    end do

    call interpolateVigMapToArray(phi, tanThetaArcsec, thetaSamples, phiSamples&
      , vigSamples2d, vigMap)

  end subroutine interpolateVigMap

  !---------------------------------------------------------------------
  subroutine interpolateVigMapToArray(phi, tanThetaArcsec, thetaSamples&
  , phiSamples, vigSamples2d, vigMap)
  !---------------------------------------------------------------------
  ! The fundamental input here is the 2-d array vigSamples2d(). This contains a set of samples of the vignetting function for the specified instrument. The first index of this array runs over theta (angle away from the optical axis in the TELCOORD system) and the second over phi, the azimuthal angle.
  !
  ! The subroutine uses cubic splines to interpolate this polar array of vignet samples at an array of theta/phi values. The result is returned in the array vigMap().
  !
  ! NOTE:
  !	(i) It is required that the cal is set up with the appropriate instrument, ccdChipId, ccdNodeId and ccdModeId beforehand.
  !	(ii) This routine calls CAL_SetState(randomize=.false.).

    implicit none

    real(double), intent(in)  :: phi(:,:),&
                                 tanThetaArcsec(size(phi,1),size(phi,2)),&
                                 thetaSamples(:),&
                                 phiSamples(:)
    real(single), intent(in)  :: vigSamples2d(&
                                   size(thetaSamples),&
                                   size(phiSamples))
    real(single), intent(out) :: vigMap(size(phi,1),size(phi,2))

    integer(int16) ::&
      numThetaSamples, numPhiSamples, xSize, ySize,&
      xi, yi, loPhiI, hiPhiI
    real(single), allocatable :: vig2ndDerivWrtTheta(:)
    real(single) :: thetaVal, phiVal, vigLoPhi, vigHiPhi, phiFrac

    !-------------------------------------------------------------------
    ! Start:

    numThetaSamples = size(thetaSamples)
    numPhiSamples   = size(phiSamples)
    ! NOTE these all must be >= 2.

    xSize = size(phi,1)
    ySize = size(phi,2)

!print *, 'phi corner vals:', phi(1,1), phi(1,ySize), phi(xSize,1)&
!, phi(xSize,ySize)
!print *, 'the corner vals:', tanThetaArcsec(1,1), tanThetaArcsec(1,ySize)&
!, tanThetaArcsec(xSize,1), tanThetaArcsec(xSize,ySize)
!print *, 'min/max phi samples:', minval(phiSamples), maxval(phiSamples)
!print *, 'min/max the samples:', minval(thetaSamples), maxval(thetaSamples)
!call error('',"")

    allocate(vig2ndDerivWrtTheta(numThetaSamples))

    ! Find the applicable range of theta and phi indices:

    vigMap = 0.0
    do xi = 1, xSize
      do yi = 1, ySize
        thetaVal = tanThetaArcsec(xi,yi)
        phiVal = phi(xi,yi)

        if  (thetaVal < thetaSamples(1)&
        .or. thetaVal > thetaSamples(numThetaSamples))&
      cycle

        if  (phiVal < phiSamples(1)&
        .or. phiVal > phiSamples(numPhiSamples))&
      cycle

        loPhiI =  largestTrueIndex(phiSamples <= phiVal)
        hiPhiI = smallestTrueIndex(phiSamples >= phiVal)

        if (loPhiI == hiPhiI) then
          if (hiPhiI < numPhiSamples) then
            hiPhiI = hiPhiI + 1
          else
            loPhiI = loPhiI - 1
          end if
        end if

        phiFrac = (phiVal - phiSamples(loPhiI))&
          / (phiSamples(hiPhiI) - phiSamples(loPhiI))

        call calcY2ForCubicSplines(real(thetaSamples, single)&
          , vigSamples2d(:, loPhiI)&
          , vig2ndDerivWrtTheta&
          , loBoundType=SPLINE_FIXED_1ST, loBoundValue=0.0&
          , hiBoundType=SPLINE_FIXED_2ND, hiBoundValue=0.0& ! only an approx,
                                                      ! don't know the
                                                      ! correct outer BC.
          )

!        call cubicSplineInterpolate(real(thetaSamples, single)&
!          , vigSamples2d(:, loPhiI)&
!          , vig2ndDerivWrtTheta, thetaVal&
!          , vigLoPhi&
!        )
        vigLoPhi = cubicSplineInterpolate(real(thetaSamples, single)&
          , vigSamples2d(:, loPhiI)&
          , vig2ndDerivWrtTheta, thetaVal&
        )

!        call calcCubicSplines(real(thetaSamples, single)&
!          , vigSamples2d(:, hiPhiI)&
!          , loBoundType='fixed1st', loBoundValue=0.0&
!          , hiBoundType='fixed2nd', hiBoundValue=0.0& ! only an approx,
!                                                      ! don't know the
!                                                      ! correct outer BC.
!          , y2=vig2ndDerivWrtTheta&
!        )
        call calcY2ForCubicSplines(real(thetaSamples, single)&
          , vigSamples2d(:, hiPhiI)&
          , vig2ndDerivWrtTheta&
          , loBoundType=SPLINE_FIXED_1ST, loBoundValue=0.0&
          , hiBoundType=SPLINE_FIXED_2ND, hiBoundValue=0.0& ! only an approx,
                                                      ! don't know the
                                                      ! correct outer BC.
          )

!        call cubicSplineInterpolate(real(thetaSamples, single)&
!          , vigSamples2d(:, hiPhiI)&
!          , vig2ndDerivWrtTheta, thetaVal&
!          , vigHiPhi&
!        )
        vigHiPhi = cubicSplineInterpolate(real(thetaSamples, single)&
          , vigSamples2d(:, hiPhiI)&
          , vig2ndDerivWrtTheta, thetaVal&
        )

        vigMap(xi,yi) = fracVal(vigLoPhi, vigHiPhi, phiFrac)
      end do
    end do

  end subroutine interpolateVigMapToArray

  !---------------------------------------------------------------------
  subroutine calcPsfArray(EeVsamples, EeVspectrum, EeVranges&
  , thetaArcsec, phi, desiredEnergyFraction, psfArrayTelcoord, psfPixelSizeMm)
  !---------------------------------------------------------------------

    implicit none 

    real(double),      intent(in)  :: EeVsamples(:),&
                                      thetaArcsec,&
                                      phi,&
                                      desiredEnergyFraction
    real(single),      intent(in)  :: EeVspectrum(size(EeVsamples))
    type(RangeT),      intent(in)  :: EeVranges(:)
    real(single),      pointer     :: psfArrayTelcoord(:,:)
    type(PsfBinSizeT), intent(out) :: psfPixelSizeMm

    real(single), pointer :: psfSamplesCubeTelcoord(:,:,:) => null()
    integer(int16) :: startXi, finisXi, startYi, finisYi, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    call constructPsfCubeTelcoord(EeVsamples, thetaArcsec, phi&
      , desiredEnergyFraction, psfSamplesCubeTelcoord, psfPixelSizeMm)

    startXi = lbound(psfSamplesCubeTelcoord, 2)
    finisXi = ubound(psfSamplesCubeTelcoord, 2)
    startYi = lbound(psfSamplesCubeTelcoord, 3)
    finisYi = ubound(psfSamplesCubeTelcoord, 3)
    allocate(psfArrayTelcoord(startXi:finisXi, startYi:finisYi))

    do xi = startXi, finisXi
      do yi = startYi, finisYi
        psfArrayTelcoord(xi, yi) = integrateRanges(real(EeVsamples, single)&
          , psfSamplesCubeTelcoord(:,xi,yi) * EeVspectrum, EeVranges)
      end do
    end do
    
    deallocate(psfSamplesCubeTelcoord)

  end subroutine calcPsfArray

  !---------------------------------------------------------------------
  function checkDss(inBlock, checkCcd) result(status)
  !---------------------------------------------------------------------
  ! This is called by a task which wishes to exposure-correct a product. The product data set should be supplied in the variable 'set'. The first thing to check is that there is not more than 1 component for each of the specified ccd numbers.

    implicit none

    type(BlockT),  intent(in) :: inBlock
    logical(bool), intent(in) :: checkCcd(:)
    integer :: status

    integer :: numCcds, ccdNum!!!, numComponentsFoundThisCcd
    type(DssT)      :: dssPointer

    !-------------------------------------------------------------------
    ! Start

    status = 0

    dssPointer = dataSubspace(inBlock)

    numCcds = size(checkCcd)
    do ccdNum = 1, numCcds
    if (.not. checkCcd(ccdNum)) cycle
      if (numDssCompsWithValue(dssPointer, 'CCDNR', real(ccdNum, single))&
      .gt. 1) then
        status = -1
  return
      end if
    end do

    call dssRelease(dssPointer)

  end function checkDss

  !---------------------------------------------------------------------
  function calcTotalGoodTime(set, inBlock, ccdNum) result(expDuration)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    type(BlockT),   intent(in) :: inBlock
    integer,        intent(in) :: ccdNum
    real(single)               :: expDuration

    type(IntervalT), pointer :: gti(:) => null()
    integer(int32) :: gtiNum, numGtis
!    integer :: status

    !-------------------------------------------------------------------
    ! Start

!    call readGtis_20030703(set, inBlock, ccdNum, gti, status)
call readGtisFromDss(set, inBlock, ccdNum, gti)
!if (status .lt. 0) call error('dssNoGtis','')
!    if (btest(status, 0)) call warning('dssUnknownTimeFilter','')

    numGtis = size(gti)
    expDuration = 0.0
    do gtiNum = 1, numGtis
      expDuration = expDuration + gti(gtiNum)%upper - gti(gtiNum)%lower
    end do
    deallocate(gti)

  end function calcTotalGoodTime

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module exposure
