!
! Module: fftw2_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains a header for fftw (version 2) and one or two other helpful routines. NOTE that a program should NOT use both this module and fftw3_aux - they are incompatible.
! 
! Comments: TODO:
! 
! $Id: fftw2_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module fftw2_aux

  use types

  use constants

  implicit none 

  public

!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
!     This file contains PARAMETER statements for various constants
!     that can be passed to FFTW routines.  You should include
!     this file in any FORTRAN program that calls the fftw_f77
!     routines (either directly or with an #include statement
!     if you use the C preprocessor).

      integer, parameter, public ::&
        FFTW_FORWARD          =  -1,&
        FFTW_BACKWARD         =   1,&
        FFTW_REAL_TO_COMPLEX  =  -1,&
        FFTW_COMPLEX_TO_REAL  =   1,&
        FFTW_ESTIMATE         =   0,&
        FFTW_MEASURE          =   1,&
        FFTW_OUT_OF_PLACE     =   0,&
        FFTW_IN_PLACE         =   8,&
        FFTW_USE_WISDOM       =  16,&
        FFTW_THREADSAFE       = 128

!     Constants for the MPI wrappers:
      integer, parameter, public ::&
        FFTW_TRANSPOSED_ORDER =     1,&
        FFTW_NORMAL_ORDER     =     0,&
        FFTW_SCRAMBLED_INPUT  =  8192,&
        FFTW_SCRAMBLED_OUTPUT = 16384

!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  ! Arrangements so that fftw calls work on all platforms:
  type FftwPlanType
    integer, pointer :: cPointer
  end type

  private template
  private calcPowerSpectrum
  public findNextHighest2357multiple
  public symmetriseAryForFFT

contains 

  !---------------------------------------------------------------------
  function findNextHighest2357multiple(i) result(result)
  !---------------------------------------------------------------------
  ! This is designed to look for the smallest integer that satisfies the
  ! following conditions: (i) it is greater than or equal to the argument
  ! i; (ii) it is a product only of the numbers 2, 3, 5 and 7. The library
  ! fftw works most efficiently on arrays that have the second property,
  ! so images are padded in size (i) to prevent any 'wrapping' of the
  ! convolver and (ii) so they have dimensions divisable by 2, 3, 5 and/or 7.

    implicit none

    integer, intent(in) ::&
      i
    integer ::&
      result

    integer ::&
      maxa,&
      maxb,&
      maxc,&
      maxd,&
      a,&
      b,&
      c,&
      d,&
      besta,&
      bestb,&
      bestc,&
      bestd
    real(single) ::&
      lni,&
      term,&
      bestTerm
    logical ::&
      firstOver

    !-------------------------------------------------------------------
    ! Start:

    lni = log(real(i))

    maxa = ceiling(lni / lnTwo)		! These are defined
    maxb = ceiling(lni / lnThree)	! in ssclib/constants.
    maxc = ceiling(lni / lnFive)	!
    maxd = ceiling(lni / lnSeven)	!

    firstOver = .true.
    bestTerm = 0.0 ! just to define it, so the inner 'if' generates no
    ! complaints from the compiler.
    do a = 0, maxa
      do b = 0, maxb
        do c = 0, maxc
          DLOOP: do d = 0, maxd
            term = a * lnTwo + b * lnThree + c * lnFive + d * lnSeven
            if (term >= lni .and. (term < bestTerm .or. firstOver)) then
              bestTerm = term
              besta = a
              bestb = b
              bestc = c
              bestd = d

              if (firstOver) firstOver = .false.

          exit DLOOP
            end if
          end do DLOOP
        end do
      end do
    end do

    result = (2**besta) * (3**bestb) * (5**bestc) * (7**bestd)

  end function findNextHighest2357multiple

  !---------------------------------------------------------------------
  subroutine calcPowerSpectrum(fftArray, psArray)
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in) :: fftArray(:, :)
    real(single), intent(out) :: psArray(:, :)

    !-------------------------------------------------------------------
    ! Start

    psArray = fftArray * conjg(fftArray)

  end subroutine calcPowerSpectrum

  !---------------------------------------------------------------------
  subroutine symmetriseAryForFFT(array)
  !---------------------------------------------------------------------
  ! The purpose of this subroutine is manipulate a real array (of size (I, J)) which has the property that
  !
  !	array(i,j) = array(i,J+2-j) = array(I+2-i,j) = array(I+2-i,J+2-j)
  !
  ! for i and j > 1. In other words, it should be symmetrical about reflection in either the i = I/2+1 or j = J/2+1 axes, and therefore give rise to no imaginary terms upon back-fft.
  !
  ! The object is accomplished by averaging the contributions from the four corners of the array.

    implicit none

    real(single), intent(inout) :: array(:, :)

    integer(int16) :: xSize, ySize, xi, yi

    !-------------------------------------------------------------------
    ! Start

    xSize = size(array, 1)
    ySize = size(array, 2)

    do xi = 2, xSize/2 + 1
      do yi = 2, ySize/2 + 1
        array(xi, yi) = (array(xi, yi) + array(xi, ySize + 2 - yi)&
        + array(xSize + 2 - xi, yi) + array(xSize + 2 - xi, ySize + 2 - yi))&
        / 4.0
        array(xi,             ySize + 2 - yi) = array(xi, yi)
        array(xSize + 2 - xi, yi)             = array(xi, yi)
        array(xSize + 2 - xi, ySize + 2 - yi) = array(xi, yi)
      end do
    end do

  end subroutine symmetriseAryForFFT

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module fftw2_aux
