!
! Module: linear
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines to do with least-squares estimation and solution of linear equations.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module linear

  use types
  use errorHandling

  implicit none 

logical, parameter :: testFlag = .false.

  private 

  public stdDev
  public robustMean
  public fitLine
  public fitPolynomial
  private calcPolyChi2
  public solveLinearTriDiag
  public solveLinearEquations
  public invertPosDefMatrix
!  private decomposeCholesky
  private invertCholeskyDouble

  interface robustMean
    module procedure robustMean1d
    module procedure robustMean2d
  end interface

  interface solveLinearTriDiag
    module procedure solveLinearTriDiagSingle, solveLinearTriDiagDouble
  end interface

  interface solveLinearEquations
    module procedure solveLinearEquationsSingle, solveLinearEquationsDouble
  end interface

  interface invertPosDefMatrix
    module procedure invertPosDefMatrixSingle, invertPosDefMatrixDouble
  end interface

contains 

  !---------------------------------------------------------------------
  function stdDev(vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: vector(:)
    real(single)             :: stdDev

    real(single)    :: average, ssr, residual
    integer :: nsize, n

    !-------------------------------------------------------------------
    ! Start

    nsize = size(vector)

    if (nsize > 1) then
      average = sum(vector) / real(nsize)
      ssr = 0.0
      do n = 1, nsize
        residual = vector(n) - average
        ssr = ssr + residual * residual
      end do
      stdDev = sqrt(ssr / real(nsize - 1))

    else
      stdDev = 0.0
    end if

  end function stdDev

  !---------------------------------------------------------------------
  subroutine robustMean2d(array, uncertsArray, largestDevianceRatio&
  , maskedMean, mask, uncertInMean, reducedChiSquared, status)
  !---------------------------------------------------------------------

    implicit none

    real(single),  intent(in)              :: array(:,:),&
                                              uncertsArray(&
                                                size(array,1)&
                                               ,size(array,2)),&
                                              largestDevianceRatio
    real(single),  intent(out)             :: maskedMean
    real(single),  intent(out),   optional :: uncertInMean,&
                                              reducedChiSquared
    logical(bool), intent(inout), optional :: mask(&
                                                size(array,1)&
                                               ,size(array,2))
    integer,       intent(out),   optional :: status

    integer ::&
      foreignStatus, vectorShape(1), arrayShape(2),&
      xSize, ySize, vSize
    logical(bool), allocatable :: localMask(:)
    real(single), allocatable :: vector(:), uncertsVector(:)
    real(single) :: localUncertInMean, localReducedChiSquared

    !-------------------------------------------------------------------
    ! Start

    xSize = size(array, 1)
    ySize = size(array, 2)
    vSize = xSize * ySize
    allocate(localMask(vSize), vector(vSize), uncertsVector(vSize))

    maskedMean = 0.0

    vectorShape(1) = size(array, 1) * size(array, 2)
    arrayShape(1) = size(array, 1)
    arrayShape(2) = size(array, 2)
    vector = reshape(array, vectorShape)
    uncertsVector = reshape(uncertsArray, vectorShape)
    if (present(mask)) then
      localMask = reshape(mask, vectorShape)
      call robustMean1d(vector, uncertsVector, largestDevianceRatio&
        , maskedMean, localMask, localUncertInMean, localReducedChiSquared&
        , foreignStatus)
      mask = reshape(localMask, arrayShape)

    else
      call robustMean1d(vector, uncertsVector, largestDevianceRatio&
        , maskedMean, uncertInMean=localUncertInMean&
        , reducedChiSquared=localReducedChiSquared, status=foreignStatus)
    end if

    if(present(uncertInMean))      uncertInMean      = localUncertInMean
    if(present(uncertInMean))      uncertInMean      = localUncertInMean
    if(present(reducedChiSquared)) reducedChiSquared = localReducedChiSquared
    if(present(status))            status            = foreignStatus

  end subroutine robustMean2d

  !---------------------------------------------------------------------
  subroutine robustMean1d(y, yUncert, largestDevianceRatio&
  , maskedMean, mask, uncertInMean, reducedChiSquared, status)
  !---------------------------------------------------------------------
  ! Usually when one has a sequence of values which is known to include some
  ! 'outliers', the quickest way to find a sort of average which doesn't
  ! include the outliers is to take the median of the sequence. However if
  ! most of the sequence values are either 0 or 1, there is not much use in
  ! taking a median since it too will be either 0 or 1. It is the purpose
  ! of this routine is to return a mean value of a masked subset of such a
  ! sequence, the mask being chosen such that no masked value departs from
  ! the mean of the masked values by more than largestDevianceRatio times the
  ! standard deviation of the masked values. In other words, the explicit
  ! variance of any data point i, calculated as (y_i - mean_y)^2, should not
  ! exceed the implicit variance, which is the square of the uncertainty in
  ! y_i, by more than largestVarianceRatio = largestDevianceRatio^2.
  !
  ! Note that the y values in the calculation of the mean are weighted
  ! by 1 / yUncert^2, which weighting is known to give the lowest uncertainty
  ! in the result.

    implicit none

    real(single),  intent(in)              :: y(:),&
                                              yUncert(size(y)),&
                                              largestDevianceRatio
    real(single),  intent(out)             :: maskedMean
    real(single),  intent(out),   optional :: uncertInMean,&
                                              reducedChiSquared
    logical(bool), intent(inout), optional :: mask(size(y))
    integer,       intent(out),   optional :: status

    integer :: vSize, localStatus, degFree
    logical(bool), allocatable :: localMask(:)
    real(single) :: largestVarianceRatio, maskedMeanvariance
    real(single), allocatable :: deviance(:), variance(:), scatterVariance(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    maskedMean = 0.0
    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if
    if (present(uncertInMean)) uncertInMean = 0.0
    if (present(reducedChiSquared)) reducedChiSquared = 0.0

    vSize = size(y)
    allocate(localMask(vSize))
    allocate(deviance(vSize), variance(vSize), scatterVariance(vSize))

    if (vSize <= 0) then
      if (present(status)) status = -1
  return
    end if
    if (largestDevianceRatio <= 1.0) then
      if (present(status)) status = -2
  return
    end if
    if (any(yUncert <= 0)) then
      localStatus = ibset(localStatus, 0)
      localMask = (localMask .and. (yUncert > 0))
    end if

    variance = yUncert * yUncert

    largestVarianceRatio = largestDevianceRatio * largestDevianceRatio
    do
      if (count(localMask) <= 0) then
        if (present(status)) status = -3
  return
      end if

      maskedMean = sum(y / variance, localMask) / sum(1.0 / variance, localMask)
      deviance = y - maskedMean
      scatterVariance = deviance * deviance

      if (.not. any(scatterVariance / variance  > largestVarianceRatio&
      .and. localMask))&
    exit ! because no new pixels have been masked in this iteration:
         ! hence the algorithm has converged.

      where(scatterVariance / variance > largestVarianceRatio)&
        localMask = .false.
    end do

    maskedMeanvariance = 1.0 / sum(1.0 / variance, localMask)

    if (present(mask)) mask = localMask
    if (present(uncertInMean)) uncertInMean = sqrt(maskedMeanvariance)

    if (present(reducedChiSquared)) then
      degFree = count(localMask) - 1
      if (degFree <= 0) then
        if (present(status)) status = -5
  return
      end if

      reducedChiSquared = sum(scatterVariance / variance, localMask)&
        / real(degFree)
    end if

    if(present(status)) status = localStatus

  end subroutine robustMean1d

  !---------------------------------------------------------------------
  subroutine fitLine(x, y, intercept, slope, variance, covar, status)
  !---------------------------------------------------------------------
  ! Fits a straight line to the set of points defined by x and y. Standard solution which assumes uncertainty in y values only and solves the normal equations to arrive at a solution which minimizes the sum of the squares of the vertical separation between the resulting line and each point.

    implicit none

    real(single), intent(in)            :: x(:), y(size(x))
    real(single), intent(out)           :: intercept, slope
    real(single), intent(out), optional :: variance, covar(2, 2)
    integer,      intent(out), optional :: status

    integer :: numPoints, degFreedom, localStatus
    real(single)    :: averageX, averageY, determinant, localVariance,&
      localCovar(2, 2)
    real(double) :: sumX, sumY, dSumXY, dSumXX, dSumYY,&
      ssr
    real(single), allocatable :: deltaX(:), deltaY(:), residuals(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    slope            = 0.0
    intercept        = 0.0
    localVariance    = 0.0
    localCovar(1, 1) = 0.0

    numPoints = size(x)

    if (numPoints .ne. size(y)) then
      if (present(status)) status = -1
  return
    end if
    if (numPoints .lt. 2) then
      if (present(status)) status = -2
  return
    end if

    allocate(deltaX(numPoints), deltaY(numPoints))

    sumX = sum(x)
    sumY = sum(y)
    averageX = sumX / real(numPoints)
    averageY = sumY / real(numPoints)
    deltaX = x - averageX
    deltaY = y - averageY

    dSumXY = sum(deltaX * deltaY)
    dSumXX = sum(deltaX * deltaX)
    dSumYY = sum(deltaY * deltaY)

    deallocate(deltaX, deltaY)

    determinant = dSumXX * numPoints

    if (determinant <= 0) then
      if (present(status)) status = -3
  return
    end if

    slope = numPoints * dSumXY / determinant
    intercept = averageY - averageX * slope

    allocate(residuals(numPoints))
    residuals = y - intercept - slope * x
    ssr = sum(residuals * residuals)
    deallocate(residuals)

    degFreedom = numPoints - 2
    localVariance = ssr / real(degFreedom)

    localCovar(1, 1) = localVariance * (sumX * averageX + dSumXX)  / determinant
    ! equals square of uncert. in intercept.

    localCovar(1, 2) = -localVariance * sumX / determinant
    localCovar(2, 1) = localCovar(1, 2)

    localCovar(2, 2) = localVariance * numPoints / determinant
    ! equals square of uncert. in slope.

    if (present(variance)) variance = localVariance
    if (present(covar))       covar = localCovar
    if (present(status))     status = localStatus

  end subroutine fitLine

  !---------------------------------------------------------------------
  subroutine fitPolynomial(x, y, coeffs, yVar, pinMask, chi2, errMatrix, status)
  !---------------------------------------------------------------------
  ! This subroutine fits a polynomial to the set of points defined by
  ! x and y. The order of the polynomial is given by the size of the
  ! vector coeffs. As per usual, only the y values are assumed to have
  ! significant uncertainties. The subroutine solves the normal equations
  ! matrix() * coeffs() = b() to arrive at a solution which minimizes chi
  ! squared (chi2) between the data and the fitted polynomial.
  !
  ! Some or all of the coefficients can be 'pinned' or not fitted. These
  ! values should be supplied in the vector coeff() (note that this is of
  ! intent 'inout' as required). The appropriate members of pinMask()
  ! should be set to T; all other members of pinMask() should of course
  ! be F. For example, suppose it was desired to fit to the data a
  ! function of the form y(x) = a + c * x^2. This is equivalent to fitting
  ! a full quadratic function to the data, but with the linear coefficient
  ! pinned at 0. To achieve this result, coeff() and pinMask() should be
  ! of size 3, with coeff(2) = 0.0 and pinMask = /(F T F)/.
  !
  ! The general form of the normal equations in the event of pinning is
  ! perhaps best illustrated by using the above example. In this case the
  ! equations are
  !
  !	[sum(x^0/yVar) sum(x^2/yVar)]   [coeffs(1)]   [sum(d*x^0/yVar)]
  !	[                           ] * [         ] = [               ] -
  !	[sum(x^2/yVar) sum(x^4/yVar)]   [coeffs(3)]   [sum(d*x^2/yVar)]
  !
  !	  [sum(x^1/yVar)]
  !	- [             ] * [coeffs(2)].
  !	  [sum(x^3/yVar)]
  !
  ! Translated into the variable names used in this subroutine, this becomes:
  !
  !	freeMatrix * freeCoeffs = freeVector - pinnedMatrix * pinnedCoeffs.
  !
  ! The pinnedMatrix should have numFreeCoeffs rows and numPinnedCoeffs
  ! columns.
  !
  ! Note that the matrix of uncertainties in the fitted coefficients as
  ! well as the chi squared value at the optimum are also returned.
  !---------------------------------------------------------------------

    implicit none

    real(single),  intent(in)            :: x(:),&
                                            y(size(x))
    real(single),  intent(inout)         :: coeffs(:)
    real(single),  intent(in),  optional :: yVar(size(x))
    logical(bool), intent(in),  optional :: pinMask(size(coeffs))
    real(single),  intent(out), optional :: chi2,&
                                            errMatrix(size(coeffs),&
                                            size(coeffs))
    integer,       intent(out), optional :: status

    integer ::&
      numPoints, order, localStatus, foreignStatus, power,&
      i, j, numPinnedCoeffs, numFreeCoeffs, freeRow, pinnedRow,&
      freeCol, pinnedCol,&
      numLargestFreeCoeff, largestPowerOfX
    integer, allocatable ::&
      freeMatOrd(:,:), freeVecOrd(:),&
      pinnedMatOrd(:,:)
    real(single), allocatable ::&
      freeMatrix(:,:), freeVector(:), pinnedMatrix(:,:),&
      freeCoeffs(:), pinnedCoeffs(:), freeErrMatrix(:,:)
    real(single), allocatable :: xToPowers(:), localYvar(:)
    logical(bool), allocatable :: coeffIsPinned(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    allocate(xToPowers(size(x)), localYvar(size(x)))
    allocate(coeffIsPinned(size(coeffs)))

    if (present(yVar)) then
      localYvar = yVar
    else
      localYvar = 1.0
    end if
    if (present(pinMask)) then
      where(.not. pinMask) coeffs = 0.0
      coeffIsPinned = pinMask
    else
      coeffIsPinned = .false.
    end if
    if (present(chi2)) chi2 = 0.0
    if (present(errMatrix)) errMatrix = 0.0

    numPoints = size(x)
    order = size(coeffs) - 1
    if (numPoints .lt. order + 1) then
      if (present(status)) status = -2
  return
    end if
    if (order .lt. 0) then
      if (present(status)) status = -3
  return
    end if
    if (any(localYvar <= 0)) then
      if (present(status)) status = -4
  return
    end if
!    do i = 1, order
!      do j = i + 1, order + 1
!        if (x(i) .eq. x(j)) then
!          if (present(status)) status = -4
!  return
!        end if
!      end do
!    end do

    numPinnedCoeffs = count(coeffIsPinned)
    numFreeCoeffs   = order + 1 - numPinnedCoeffs
    if (numFreeCoeffs <= 0) then
      if (present(chi2)) chi2 = calcPolyChi2(x, y, localYvar, coeffs)
  return
    end if

    allocate(freeMatrix(numFreeCoeffs, numFreeCoeffs))
    allocate(freeMatOrd(numFreeCoeffs, numFreeCoeffs))
    allocate(freeVector(numFreeCoeffs))
    allocate(freeVecOrd(numFreeCoeffs))
    allocate(pinnedMatrix(numPinnedCoeffs, numFreeCoeffs))
    allocate(pinnedMatOrd(numPinnedCoeffs, numFreeCoeffs))

    allocate(freeCoeffs(numFreeCoeffs))
    allocate(pinnedCoeffs(numPinnedCoeffs))

    ! The following algorithm is designed to be quick. The step which probably takes the most time is the calculation of sum(x^n): hence this is done as few times as possible.

    ! Set up the x power indices (also load the pinned-coeff auxiliary vector):
    freeRow = 0
    pinnedRow = 0
    do i = 1, order + 1
      if (coeffIsPinned(i)) then
        pinnedRow = pinnedRow + 1
        pinnedCoeffs(pinnedRow) = coeffs(i)
      else
        freeRow = freeRow + 1
        freeVecOrd(freeRow) = i - 1

        freeCol = 0
        pinnedCol = 0
        do j = 1, order + 1
          if (coeffIsPinned(j)) then
            pinnedCol = pinnedCol + 1
            pinnedMatOrd(freeRow, pinnedCol) = (i - 1) + (j - 1)
          else
            freeCol = freeCol + 1
            freeMatOrd(freeRow, freeCol) = (i - 1) + (j - 1)
          end if
        end do
      end if
    end do
 
    ! Now load up the actual matrices. First find how large a power of x
    ! is needed:
    numLargestFreeCoeff = 0
    do i = order + 1, 1, -1
      if (.not. coeffIsPinned(i)) then
        numLargestFreeCoeff = i
    exit
      end if
    end do
    if (coeffIsPinned(order + 1)) then ! last coefficient is pinned.
      largestPowerOfX = numLargestFreeCoeff - 1 + order
    else ! last coefficient is free.
      largestPowerOfX = 2 * order
    end if

if (largestPowerOfX .ne. maxval(pinnedMatOrd)&
.and. largestPowerOfX .ne. maxval(freeMatOrd)) then
if (present(status)) status = -100 !BUG!!
return
end if

    ! Now do the calculations:
    xToPowers = 1.0
    do power = 0, largestPowerOfX
      where(freeMatOrd   .eq. power)&
        freeMatrix   = sum(xToPowers / localYvar)
      where(pinnedMatOrd .eq. power)&
        pinnedMatrix = sum(xToPowers / localYvar)
      where(freeVecOrd   .eq. power)&
        freeVector   = sum(xToPowers / localYvar * y)
    end do

!    ! The final step is to multiply the pinned-coeff auxiliaries and add
!    ! the result to freeVector:
!    do i = 1, numFreeCoeffs
!      freeVector(i) = freeVector(i) - matmul(pinnedMatrix, pinnedCoeffs)
!    end do

    call solveLinearEquations(freeMatrix, freeCoeffs&
    , freeVector - matmul(pinnedCoeffs, pinnedMatrix)&
    , status = foreignStatus)
!*** this doesn't exactly _crash_ if numPinnedCoeffs == 0, but I should check to see that in this case freeVector - matmul(pinnedCoeffs, pinnedMatrix) is equal to freeVector as expected.

!**** amalgamate localStatus and foreignStatus.

    freeRow = 0
    do i = 1, order + 1
    if (coeffIsPinned(i)) cycle
      freeRow = freeRow + 1
      coeffs(i) = freeCoeffs(freeRow)
    end do

    if (present(chi2)) chi2 = calcPolyChi2(x, y, localYvar, coeffs)
    if (present(errMatrix)) then
      allocate(freeErrMatrix(numFreeCoeffs, numFreeCoeffs))
      call invertPosDefMatrixSingle(freeMatrix, freeErrMatrix)
      ! The error matrix (matrix of variances in the fitted coefficients) is the inverse of the curvature matrix. And the latter? Conveniently, it happens to be identical to the already-calculated freeMatrix. 

      freeRow = 0
      do i = 1, order + 1
      if (coeffIsPinned(i)) cycle
        freeCol = 0
        do j = 1, order + 1
        if (coeffIsPinned(j)) cycle
          freeCol = freeCol + 1
          errMatrix(i, j) = freeErrMatrix(freeRow, freeCol)
        end do
      end do
    end if

    if (present(status)) status = localStatus

  end subroutine fitPolynomial

  !---------------------------------------------------------------------
  function calcPolyChi2(x, y, yVar, coeffs) result(chi2)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: x(:), y(size(x)), yVar(size(x)), coeffs(:)
    real(single)             :: chi2

    integer :: i
    real(single), allocatable :: xToPowers(:), modelY(:), diffTerms(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(xToPowers(size(x)), modelY(size(x)), diffTerms(size(x)))

    modelY = 0.0
    xToPowers = 1.0
    do i = 1, size(coeffs) - 1
      modelY = modelY + coeffs(i) * xToPowers
      xToPowers = xToPowers * x
    end do
    modelY = modelY + coeffs(size(coeffs)) * xToPowers

    diffTerms = modelY - y
    chi2 = sum(diffTerms * diffTerms / yVar)

  end function calcPolyChi2

  !---------------------------------------------------------------------
  subroutine invertPosDefMatrixSingle(matrix, inverse, exitStatusI)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: matrix(:,:)
    real(single), intent(out)           :: inverse(&
                                             size(matrix,1)&
                                            ,size(matrix,1))
    integer,      intent(out), optional :: exitStatusI

    real(double), allocatable :: inverseDble(:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(inverseDble(size(matrix,1),size(matrix,1)))
    call invertPosDefMatrixDouble(real(matrix, double), inverseDble, exitStatusI)
    inverse = real(inverseDble, single)

  end subroutine invertPosDefMatrixSingle

  !---------------------------------------------------------------------
  subroutine invertPosDefMatrixDouble(matrix, inverse, exitStatusI)
  !---------------------------------------------------------------------
  ! This has been checked and seems to be ok.

    implicit none

    real(double), intent(in)            :: matrix(:,:)
    real(double), intent(out)           :: inverse(&
                                             size(matrix,1)&
                                            ,size(matrix,1))
    integer,      intent(out), optional :: exitStatusI

    integer :: mSize
    real(double), allocatable :: cholesky(:,:), invertedCholesky(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (size(matrix, 1) .ne. size(matrix, 2))&
call error('nonSquareMatrix',"")

    if (size(matrix, 1) <= 0)&
  return

    mSize = size(matrix, 1)
    allocate(cholesky(mSize,mSize))
    allocate(invertedCholesky(mSize,mSize))

    call decomposeCholeskyDouble(matrix, cholesky, exitStatusI)
    call invertCholeskyDouble(cholesky, invertedCholesky)

    inverse = matmul(transpose(invertedCholesky), invertedCholesky)

  end subroutine invertPosDefMatrixDouble

  !---------------------------------------------------------------------
  subroutine decomposeCholeskySingle(matrix, cholesky, exitStatusI)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: matrix(:,:)
    real(single), intent(out)           :: cholesky(&
                                             size(matrix,1)&
                                            ,size(matrix,1))
    integer,      intent(out), optional :: exitStatusI

    real(double), allocatable :: choleskyDouble(:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(choleskyDouble(size(matrix,1),size(matrix,1)))
    call decomposeCholeskyDouble(real(matrix, double), choleskyDouble, exitStatusI)
    cholesky = real(choleskyDouble, single)

  end subroutine decomposeCholeskySingle

  !---------------------------------------------------------------------
  subroutine decomposeCholeskyDouble(matrix, cholesky, exitStatusI)
  !---------------------------------------------------------------------
  ! Performs a cholesky decomposition (with a failure if the matrix is
  ! not positive definite) and leaves the lower triangular half L in the
  ! matrix cholesky(). The rest of cholesky() is filled with zeros.
  !
  ! NOTE that the usual matrix convention is assumed, namely that the
  ! first index is the row number and the second is the column number.

    implicit none

    real(double), intent(in)            :: matrix(:,:)
    real(double), intent(out)           :: cholesky(&
                                             size(matrix,1)&
                                            ,size(matrix,1))
    integer,      intent(out), optional :: exitStatusI

    integer :: mSize, i, j, k
    real(double) :: summ

    !-------------------------------------------------------------------
    ! Start

    if (present(exitStatusI)) exitStatusI = 0

    if (size(matrix, 1) .ne. size(matrix, 2))&
call error('nonSquareMatrix',"")

    if (size(matrix, 1) <= 0)&
  return

    mSize = size(matrix, 1)

    cholesky = 0.0d0

    do i = 1, mSize
      summ = matrix(i, i)
      do k = 1, i - 1
        summ = summ - cholesky(i, k) * cholesky(i, k)
      end do

      if (summ <= 0.0d0) then
        if (present(exitStatusI)) exitStatusI = -1
!call error('badMatrixRowSum',"")
        call warning('badMatrixRowSum',"")
    exit
      end if

      cholesky(i, i) = sqrt(summ)

      do j = i + 1, mSize
        summ = matrix(j, i)
        do k = 1, i - 1
          summ = summ - cholesky(j, k) * cholesky(i, k)
        end do
        cholesky(j, i) = summ / cholesky(i, i)
      end do
    end do

  end subroutine decomposeCholeskyDouble

  !---------------------------------------------------------------------
  subroutine decomposeCholesky_old(matrix, cholesky, status)
  !---------------------------------------------------------------------
  ! Performs a cholesky decomposition (with a failure if the matrix is not positive definite) and leaves the lower triangular half L in the matrix cholesky(). The rest of cholesky() is filled with zeros.
  !
  ! This routine was copied from Numerical Recipes but I want to modify it a bit to suit it better to the present separated input and output matrices.

    implicit none

    real(single), intent(in)            :: matrix(:,:)
    real(single), intent(out)           :: cholesky(&
                                             size(matrix,1)&
                                            ,size(matrix,1))
    integer,      intent(out), optional :: status

    integer :: localStatus, mSize, row, col, i
    real(single) :: summ
real(single), allocatable :: diag(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    allocate(diag(size(matrix, 1)))

    if (size(matrix, 1) <= 0) then
      if (present(status)) status = -1
  return
    end if
    if (size(matrix, 1) .ne. size(matrix, 2)) then
      if (present(status)) status = -2
  return
    end if

    mSize = size(matrix, 1)

    cholesky = matrix
    do row = 1, mSize
      do col = row, mSize
        summ = cholesky(row, col)
        do i = row - 1, 1, -1
          summ = summ - cholesky(row, i) * cholesky(col, i)
        end do

        if (row .eq. col) then
          if (summ <= 0.0) then
            if (present(status)) status = -1
            cholesky = 0.0
  return
          end if
          diag(row) = sqrt(summ)

        else
          cholesky(col, row) = summ / diag(row)
        end if
      end do
    end do

    ! Insert zeros into upper triangle of cholesky:
    do row = 1, mSize - 1
      do col = row + 1, mSize
        cholesky(row, col) = 0.0
      end do
    end do

    if (present(status)) status = localStatus

  end subroutine decomposeCholesky_old

  !---------------------------------------------------------------------
  subroutine invertCholeskyDouble(cholesky, invertedCholesky)
  !---------------------------------------------------------------------
  ! This routine takes a cholesky-factorised matrix (note only the lower
  ! triangle is used) and finds the inverse (which is also lower-
  ! triangular). The upper triangle of the inverse contains zeros.
  !
  ! Note that the routine as written has no error checking. Thus it
  ! shouldn't be public but should just be used by other subroutines that
  ! themselves check such things as that the matrix is square and not of
  ! zero size.

    implicit none

    real(double), intent(in)  :: cholesky(:,:)
    real(double), intent(out) :: invertedCholesky(&
                                   size(cholesky,1)&
                                  ,size(cholesky,1))

    integer :: mSize, i, j, k
    real(double) :: summ

    !-------------------------------------------------------------------
    ! Start

    mSize = size(cholesky, 1)

    invertedCholesky = 0.0d0

    do i = 1, mSize
      do j = 1, i - 1
        summ = 0.0d0
        do k = j, i - 1
          summ = summ - cholesky(i,k) * invertedCholesky(k,j)
        end do
        invertedCholesky(i,j) = summ / cholesky(i,i)
      end do

      invertedCholesky(i,i) = 1.0d0 / cholesky(i,i)
    end do

  end subroutine invertCholeskyDouble

  !---------------------------------------------------------------------
  subroutine invertCholesky_old(cholesky, invertedCholesky)
  !---------------------------------------------------------------------
  ! This routine takes a cholesky-factorised matrix (note only the lower triangle is used) and finds the inverse (which is also lower-triangular). The upper triangle of the inverse contains zeros.
  !
  ! Note that the routine as written has no error checking. Thus it shouldn't be public but should just be used by other subroutines that themselves check such things as that the matrix is square and not of zero size.
  !
  ! The routine was copied from Numerical Recipes but I want to modify it a bit to suit it better to the present separated input and output matrices.

    implicit none

    real(single), intent(in)  :: cholesky(:,:)
    real(single), intent(out) :: invertedCholesky(:,:)

    integer :: mSize, row, col, i
    real(single) :: summ
real(single), allocatable :: diag(:)

    !-------------------------------------------------------------------
    ! Start

    mSize = size(cholesky, 1)
    allocate(diag(mSize))

    invertedCholesky = cholesky
    do row = 1, mSize
      diag(row) = cholesky(row, row)
    end do
    do col = 1, mSize
      invertedCholesky(col, col) = 1.0 / diag(col)
      do row = col + 1, mSize
        summ = 0.0
        do i = col, row - 1
          summ = summ - invertedCholesky(row, i) * invertedCholesky(i, col)
        end do
        invertedCholesky(row, col) = summ / diag(row)
      end do
    end do

    ! Insert zeros into upper triangle of cholesky:
    do row = 1, mSize - 1
      do col = row + 1, mSize
        invertedCholesky(row, col) = 0.0
      end do
    end do

  end subroutine invertCholesky_old

!  !---------------------------------------------------------------------
!  subroutine fitBilinear(dat, coeffs, status)
!  !---------------------------------------------------------------------
!  ! This subroutine fits a bilinear function to a 2-D array of values in dat(). The subroutine solves the normal equations matrix() * coeffs() = b() to arrive at a solution which minimizes the sum of the squares of the vertical separation between each value and the resulting bilinear function.
!
!!*** it is totally stuffed and I will have to rethink it.
!
!    implicit none
!
!    real,    intent(in)            :: dat(:,:)
!    real,    intent(out)           :: coeffs(4)
!    integer, intent(out), optional :: status
!
!    integer :: xsize, ysize, localStatus, foreignStatus, xi, yi
!    real, allocatable :: xToPowers(:), yToPowers(:), matrix(:,:), b(:)
!    real :: matrix(4, 4), b(4)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    localStatus = 0
!
!    coeffs = 0.0
!
!    xsize = size(y, 1)
!    ysize = size(y, 2)
!
!    if (xsize .lt. 2) then
!      if (present(status)) status = -1
!  return
!    end if
!    if (ysize .lt. 2) then
!      if (present(status)) status = -2
!  return
!    end if
!
!    allocate(xToPowers(xsize), xToPowers(ysize))
!    xToPowers = 1.0
!    ! Fill b() and ****the top RH corner of matrix():
!    do row = 1, 4
!      do col = row
!        matrix(i, 2 + power - i) = sum(xToPowers)
!      end do
!      b(power + 1) = sum(xToPowers * y)
!      xToPowers = xToPowers * x
!    end do
!    ! Now fill the bottom RH corner of matrix():
!    do power = order + 1, 2 * order - 1
!      do i = power + 1 - order, order + 1
!        matrix(i, 2 + power - i) = sum(xToPowers)
!      end do
!      xToPowers = xToPowers * x
!    end do
!    matrix(order + 1, order + 1) = sum(xToPowers)
!
!!print *, ''
!!print *, 'matrix:'
!!do power = 0, order
!!  print *, matrix(:, power + 1)
!!end do
!!print *, ''
!!print *, 'b:'
!!print *, b
!
!    call solveLinearEquations(matrix, coeffs, b, status = foreignStatus)
!!**** amalgamate localStatus and foreignStatus.
!!print *, 'solveLinearEquations status =', status
!!print *, ''
!
!    deallocate(xToPowers, matrix, b)
!
!    if (present(status)) status = localStatus
!
!  end subroutine fitBilinear

  !---------------------------------------------------------------------
  subroutine solveLinearTriDiagSingle(triDiag, b, x)
  !---------------------------------------------------------------------
  ! Adapted from Numerical Recipes for Fortran.

    implicit none

    real(single), intent(in)  :: b(:), triDiag(-1:1, size(b))
    real(single), intent(out) :: x(size(b))

    real(double), allocatable :: xDble(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(xDble(size(b)))

    call solveLinearTriDiagDouble(real(triDiag, double), real(b, double)&
      , xDble)

    x = real(xDble, single)

  end subroutine solveLinearTriDiagSingle

  !---------------------------------------------------------------------
  subroutine solveLinearTriDiagDouble(triDiag, b, x)
  !---------------------------------------------------------------------
  ! Adapted from Numerical Recipes for Fortran.

    implicit none

    real(double), intent(in)  :: b(:), triDiag(-1:1, size(b))
    real(double), intent(out) :: x(size(b))

    integer :: n, j
    real(double) :: bet, gam(size(b))

    !-------------------------------------------------------------------
    ! Start

    n = size(b)
  if (n <= 0) return ! trivial.

    x = 0.0d0

    bet = triDiag(0, 1)
    x(1) = b(1) / bet
    do j = 2, n
      gam(j) = triDiag(1, j-1) / bet
      bet = triDiag(0, j) - triDiag(-1, j) * gam(j)

      if (bet == 0.0d0)&
call error('',"")

      x(j) = (b(j) - triDiag(-1, j) * x(j-1)) / bet
    end do
    do j = n - 1, 1, -1
      x(j) = x(j) - gam(j+1) * x(j+1)
    end do

  end subroutine solveLinearTriDiagDouble

  !---------------------------------------------------------------------
  subroutine solveLinearTriDiagFisher(triDiag, b, x, status)
  !---------------------------------------------------------------------
  ! This version was written from the algorithm in M E Fisher. It's the
  ! same as the other, just more explicit. We'll keep it just for reference.

    implicit none

    real(single), intent(in)            :: b(:), triDiag(-1:1, size(b))
    real(single), intent(out)           :: x(size(b))
    integer,      intent(out), optional :: status

    integer :: localStatus, n, j
    real(single) :: decomp(-1:1, size(b)), tempB(size(b))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    n = size(b)
  if (n <= 0) return ! trivial.

    x = 0.0

    ! LU decompose:
    decomp = 0.0
    decomp(1,:n-1) = triDiag(1,:n-1)
    decomp(0,1) = triDiag(0,1)
    do j = 2, n
      decomp(-1,j) = triDiag(-1,j) / decomp(0,j-1)
      decomp(0,j) = triDiag(0,j) - decomp(-1,j) * decomp(1,j-1)
    end do

    ! Forward substitution:
    tempB(1) = b(1)
    do j = 2, n
      tempB(j) = b(j) - decomp(-1,j) * tempB(j-1)
    end do

    ! Backward substitution:
    x(n) = tempB(n) / decomp(0,n)
    do j = n - 1, 1, -1
      x(j) = (tempB(j) - decomp(1,j) * x(j+1)) / decomp(0,j)
    end do

    if (present(status)) status = localStatus

  end subroutine solveLinearTriDiagFisher

  !---------------------------------------------------------------------
  subroutine solveLinearEquationsSingle(m, x, y, pivotPlease, status)
  !---------------------------------------------------------------------
  ! This sub solves the matrix equation m * x = y for matrix m and vector
  ! x and y. It makes use of the fact that a (non-singular?) matrix can be
  ! decomposed into the product of two matrices L and U, L being a lower-
  ! triangular matrix with 1s on the diagonal, U being an upper-triangular
  ! matrix. The fact that L has 1s on its diagonal permits both matrices to
  ! be stored in the same array, as in done here. Partial pivoting is done
  ! if the pivotPlease flag is set .true..
  !
  ! This sub has been tested pretty well.

    implicit none

    real(single), intent(in)            :: y(:), m(size(y), size(y))
    logical,      intent(in),  optional :: pivotPlease
    real(single), intent(out)           :: x(size(y))
    integer,      intent(out), optional :: status

    real(double), allocatable :: mDble(:,:), yDble(:)
    real(double), allocatable :: xDble(:)
    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    allocate(mDble(size(y), size(y)), yDble(size(y)), xDble(size(y)))

    mDble = m
    yDble = y

!    if (present(pivotPlease)) then
      call solveLinearEquationsDouble(mDble, xDble, yDble, pivotPlease&
        , status=foreignStatus)
!    else
!      call solveLinearEquationsDouble(mDble, xDble, yDble&
!        , status=foreignStatus)
!    end if
!print *, 'in solveLinearEquationsSingle - foreignStatus =',foreignStatus

    x = xDble

    if (present(status)) status = foreignStatus

  end subroutine solveLinearEquationsSingle

  !---------------------------------------------------------------------
  subroutine solveLinearEquationsDouble(m, x, y, pivotPlease, status)
  !---------------------------------------------------------------------
  ! This sub solves the matrix equation m * x = y for matrix m and vector
  ! x and y. It makes use of the fact that a (non-singular?) matrix can be
  ! decomposed into the product of two matrices L and U, L being a lower-
  ! triangular matrix with 1s on the diagonal, U being an upper-triangular
  ! matrix. The fact that L has 1s on its diagonal permits both matrices to
  ! be stored in the same array, as in done here. Partial pivoting is done
  ! if the pivotPlease flag is set .true..
  !
  ! This sub has been tested pretty well.

    implicit none

    real(double), intent(in)            :: y(:), m(size(y), size(y))
    logical,      intent(in),  optional :: pivotPlease
    real(double), intent(out)           :: x(size(y))
    integer,      intent(out), optional :: status

    real(double), allocatable :: lu(:,:), tempY(:)
    integer :: numElements, row, col, r, c, maxrow, cc, localStatus
    real(double) :: max, temp
    logical(bool) :: localPivotPlease

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    x = 0.0

    numElements = size(y)
    if (numElements .lt. 1) then
      if (present(status)) status = -1
  return
    end if

    if (present(pivotPlease)) then
      localPivotPlease = pivotPlease
    else
      localPivotPlease = .false.
    end if

    allocate(lu(numElements, numElements), tempY(numElements))
    lu = transpose(m)
    tempY = y

    ! LU factorisation:
    do col = 1, numElements - 1
      if (localPivotPlease) then ! partial pivot.
        max = abs(lu(col, col))
        maxrow = col
        do r = col + 1, numElements
          if (abs(lu(r, col)) .gt. max) then
            max = abs(lu(r, col))
            maxrow = r
          end if
        end do !r
!*** can do this more efficiently with f95 inbuilt functions.

        do c = 1, numElements
          temp = lu(col, c)
          lu(col, c) = lu(maxrow, c)
          lu(maxrow, c) = temp
        end do !c
        temp = tempY(col)
        tempY(col) = tempY(maxrow)
        tempY(maxrow) = temp
      end if ! pivotPlease

      do row = col + 1, numElements
        if (lu(col, col) .eq. 0.0) then
          if (present(status)) status = -2
          deallocate(lu, tempY)
  return
        end if

        lu(row, col) = lu(row, col) / lu(col, col)
        do cc = col + 1, numElements
          lu(row, cc) = lu(row, cc) - lu(row, col) * lu(col, cc)
        end do !cc
      end do !row
    end do !col

    ! Forward substitution to solve L(Ux) = tempY (leaving Ux in array x):
    x(1) = tempY(1)
    do row = 2, numElements
      x(row) = tempY(row)
      do col = 1, row - 1
        x(row) = x(row) - lu(row, col) * x(col)
      end do !col
    end do !row

    ! Back substitution to solve Ux = tempY:
    if (lu(numElements, numElements) .eq. 0.0) then
      if (present(status)) status = -3
      deallocate(lu, tempY)
  return
    end if
    x(numElements) = x(numElements) / lu(numElements, numElements)

    do row = numElements - 1, 1, -1
      do col = row + 1, numElements
        x(row) = x(row) - lu(row, col) * x(col)
      end do !col

      if (lu(row, row) .eq. 0.0) then
        if (present(status)) status = -4
        deallocate(lu, tempY)
  return
      end if
      x(row) = x(row) / lu(row, row)
    end do !row

    deallocate(lu, tempY)

    if (present(status)) then
!print *, 'in solveLinearEquationsSingle - foreignStatus =',foreignStatus
status = localStatus
    end if

  end subroutine solveLinearEquationsDouble

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module linear
