!
! Module: math_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines, mainly concerned with transcendental functions and non-linear transforms.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module math_utils

  use types

  use constants
!  use array_utils	! for getBinNumber
!  use ssc_misc		! for fracVal.

  implicit none 

  private 

  public incompleteGammaP
  public lnIncompleteGammaP
  public incompleteGammaQ
  public lnIncompleteGammaQ
  private incompleteGammaPbySeries
  private incompleteGammaQbyFraction
public incompleteGammaQByAsymtotic
  public lnGamma
  public factorial
  public erf
  private erfApprox0
  private erfApprox1
  public erfc
  private erfcApprox0
  public besselJ0
  public besselJ0Zero
  public besselJ1
  public besselJ1Zero
  public sinc

contains 

  !---------------------------------------------------------------------
  subroutine incompleteGammaP(a, x, p, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! This routine was cribbed from gammp() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: p
    integer, intent(out), optional :: status

    integer :: foreignStatus
    real(single) :: q

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0
    p = 0.0

    if (x < 0.0 .or. a < 0.0 .or. (x == 0.0 .and. a == 0.0)) then
      if (present(status)) status = -1
  return
    end if

    if (x == 0.0) then
      p = 0.0
  return
    else if (a == 0.0) then
      p = 1.0
  return
    end if

    if (x < a + 1) then ! use the series representation:
      call incompleteGammaPbySeries(a, x, p, foreignStatus)
    else ! use the continued-fraction representation:
      call incompleteGammaQbyFraction(a, x, q, foreignStatus)
      p = 1.0 - q
    end if

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 1
    else if (foreignStatus > 0) then
      if (present(status)) status = foreignStatus
    end if

  end subroutine incompleteGammaP

  !---------------------------------------------------------------------
  subroutine incompleteGammaQ(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec function gamic(a,x).
  !
  ! This routine was cribbed from gammq() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: q
    integer, intent(out), optional :: status

    integer :: foreignStatus
    real(single) :: p

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0
    q = 0.0

    if (x < 0.0 .or. a < 0.0 .or. (x == 0.0 .and. a == 0.0)) then
      if (present(status)) status = -1
  return
    end if

    if (x == 0.0) then
      p = 1.0
  return
    else if (a == 0.0) then
      p = 0.0
  return
    end if

    if (x < a + 1) then ! use the series representation:
      call incompleteGammaPbySeries(a, x, p, foreignStatus)
      q = 1.0 - p
    else ! use the continued-fraction representation:
      call incompleteGammaQbyFraction(a, x, q, foreignStatus)
    end if

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 1
    else if (foreignStatus > 0) then
      if (present(status)) status = foreignStatus
    end if

  end subroutine incompleteGammaQ

  !---------------------------------------------------------------------
  subroutine incompleteGammaPbySeries(a, x, p, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! The function is here evaluated from a series.
  !
  ! This routine was cribbed from gser() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: p
    integer,      intent(out) :: status

    integer, parameter :: maxNumIterations = 150
    real(single) :: convergeCriterion, ap, summ, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))

    p = 0.0

    if (x .lt. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0 / a
    del = summ
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap + 1.0
      del = del * x / ap
      summ = summ + del

      if (abs(del) .lt. abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    p = summ * exp(-x + a * log(x) - lnGammaVal)

  end subroutine incompleteGammaPbySeries

  !---------------------------------------------------------------------
  subroutine lnIncompleteGammaP(a, x, lnP)!!!, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! The function is here evaluated from a series.
  !
  ! This routine was cribbed from gser() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)            :: a, x
    real(single), intent(out)           :: lnP
!!!    integer,      intent(out), optional :: status
integer :: status

    integer, parameter :: maxNumIterations = 100
    real(single) :: convergeCriterion, ap, summ, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))

    lnP = 0.0

    if (x .le. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0 / a
    del = summ
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap + 1.0
      del = del * x / ap
      summ = summ + del

      if (abs(del) .lt. abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

!    p = summ * exp(-x + a * log(x) - lnGammaVal)
    lnP = log(summ) - x + a * log(x) - lnGammaVal

  end subroutine lnIncompleteGammaP

  !---------------------------------------------------------------------
  subroutine incompleteGammaQbyFraction(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec routine gamic(a.x).
  !
  ! The function is evaluated by a modified Lentz's continued fraction.
  !
  ! This routine was cribbed from gcf() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: q
    integer,      intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real(single) ::&
      convergeCriterion, verySmallReal, b, c, d, h, an, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
    verySmallReal = epsilon(x)

    b = x + 1.0 - a
    c = 1.0 / verySmallReal
    d = 1.0 / b
    h = d
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      an = -i * (i - a)
      b = b + 2.0
      d = an * d + b
      if (abs(d) .lt. verySmallReal) d = verySmallReal
      c = b + an / c
      if (abs(c) .lt. verySmallReal) c = verySmallReal
      d = 1.0 / d
      del = d * c
      h = h * del
      if(abs(del - 1.0) .lt. convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    q = h * exp(-x + a * log(x) - lnGammaVal)

  end subroutine incompleteGammaQbyFraction

  !---------------------------------------------------------------------
  subroutine lnIncompleteGammaQ(a, x, lnQ)!!!, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec routine gamic(a.x).
  !
  ! The function is evaluated by a modified Lentz's continued fraction.
  !
  ! This routine was cribbed from gcf() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: lnQ
!!!    integer,      intent(out), optional :: status
integer :: status

    integer, parameter :: maxNumIterations = 100
    real(single) :: convergeCriterion, verySmallReal, b, c, d, h, an, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
    verySmallReal = epsilon(x)

    b = x + 1.0 - a
    c = 1.0 / verySmallReal
    d = 1.0 / b
    h = d
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      an = -i * (i - a)
      b = b + 2.0
      d = an * d + b
      if (abs(d) .lt. verySmallReal) d = verySmallReal
      c = b + an / c
      if (abs(c) .lt. verySmallReal) c = verySmallReal
      d = 1.0 / d
      del = d * c
      h = h * del
      if(abs(del - 1.0) .lt. convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

!    q = h * exp(-x + a * log(x) - lnGammaVal)
    lnQ = log(h) - x + a * log(x) - lnGammaVal

  end subroutine lnIncompleteGammaQ

  !---------------------------------------------------------------------
  subroutine incompleteGammaQByAsymtotic(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec function gamic(a,x).
  !
  ! This routine was cribbed from gammq() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: q
    integer,      intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real(single) :: convergeCriterion, ap, summ, del, oldDel, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
!print *, 'convergeCriterion:', convergeCriterion

    q = 0.0

    if (x .lt. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0
    del = summ
    oldDel = del
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap - 1.0
      del = del * ap / x
      summ = summ + del
!print *, 'del, summ:', del, summ

      if (abs(del) < abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if

    if (abs(del) > abs(oldDel)) exit

      oldDel = del
    end do

    if (notConverged) then
      status = -2
!!!  return
!*** for the time being calculate a prob anyway.
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    q = summ * exp(-x + (a - 1) * log(x) - lnGammaVal)
!print *, 'q:', q

  end subroutine incompleteGammaQByAsymtotic

  !---------------------------------------------------------------------
  real function lnGamma(a)
  !---------------------------------------------------------------------
  ! Returns the natural log of the absolute value of the complete gamma
  ! function T(a), which is defined by
  !
  !		       inf
  !		       /
  !		T(a) = | dt exp(-t) t^(a-1).
  !		       /
  !		      0
  !
  ! The function is evaluated by a Lanczos approximation.
  !
  ! The ln|T(a)| function can also be evaluated using the slatec function
  ! alngam(a).
  !
  ! This routine was cribbed from gammln() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in) :: a

    integer :: j
    real(double) :: ser, stp, tmp, x, y, cof(6)

    save cof, stp
    data cof, stp /&
      76.18009172947146d0,&
     -86.50532032941677d0,&
      24.01409824083091d0,&
      -1.231739572450155d0,&
       0.1208650973866179d-2,&
      -0.5395239384953d-5,&
       2.5066282746310005d0/

    !-------------------------------------------------------------------
    ! Start

    x = a
    y = x
    tmp = x + 5.5d0
    tmp = (x + 0.5d0) * log(tmp) - tmp
    ser = 1.000000000190015d0
    do j = 1, 6
      y = y + 1.0d0
      ser = ser + cof(j) / y
    end do
    lnGamma = tmp + log(stp * ser / x)

  end function lnGamma

  !---------------------------------------------------------------------
  integer function factorial(i)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: i

    integer :: j

    !-------------------------------------------------------------------
    ! Start

    factorial = 1
    do j = 2, i
      factorial = factorial * j
    end do

  end function factorial

  !---------------------------------------------------------------------
  real function erf(x, style)
  !---------------------------------------------------------------------
  ! The error function
  !
  !	                     x
  !	             2      /
  !	erf(x) = ---------- | dt exp(-t^2).
  !	          sqrt(pi)  /
  !	                   0
  !

    implicit none

    real(single), intent(in) :: x
    integer, intent(in), optional :: style

    integer :: localStyle

    !-------------------------------------------------------------------
    ! Start

    if (present(style)) then
      localStyle = style
    else
      localStyle = 0
    end if

    select case(localStyle)
    case(0)
      call incompleteGammaP(0.5, x*x, erf)
    case(1)
      erf = erfApprox0(x)
    case(2)
      erf = erfApprox1(x)
    case default
    end select

  end function erf

  !---------------------------------------------------------------------
  real function erfc(x, style)
  !---------------------------------------------------------------------
  ! The complementary error function
  !
  !	                                   inf
  !	                           2      /
  !	erfc(x) = 1 - erf(x) = ---------- | dt exp(-t^2).
  !	                        sqrt(pi)  /
  !	                                 x
  !
  ! Note that, for values obeying a gaussian distribution of mean <y> and
  ! standard deviation s, the probability of a given y value (or greater)
  ! occurring by chance is 0.5*erfc(abs(y - <y>)/s/sqrt(2))

    implicit none

    real(single), intent(in) :: x
    integer, intent(in), optional :: style

    integer :: localStyle

    !-------------------------------------------------------------------
    ! Start

    if (present(style)) then
      localStyle = style
    else
      localStyle = 0
    end if

    select case(localStyle)
    case(0)
      call incompleteGammaQ(0.5, x*x, erfc)
    case(1)
      erfc = 1.0 - erfApprox0(x)
    case(2)
      erfc = 1.0 - erfApprox1(x)
    case default
    end select

  end function erfc

  !---------------------------------------------------------------------
  function erfApprox0(x) result(erf)
  !---------------------------------------------------------------------
  ! Taken from Numerical Methods for Fortran.

    implicit none

    real(single), intent(in) :: x

    real(single) :: t, z, erf
    real(single), parameter ::&
      p0 = -1.26551223,&
      p1 =  1.00002368,&
      p2 =  0.37409196,&
      p3 =  0.09678418,&
      p4 = -0.18628806,&
      p5 =  0.27886807,&
      p6 = -1.13520398,&
      p7 =  1.48851587,&
      p8 = -0.82215223,&
      p9 =  0.17087277

    !-------------------------------------------------------------------
    ! Start

    z = abs(x)
    t = 1.0 / (1.0 + 0.5 * z)
    erf = t * exp(-z * z + p0 + t * (p1 + t * (p2&
        + t * (p3 + t * (p4 + t * (p5 + t * (p6&
        + t * (p7 + t * (p8 + t * p9)))))))))

    if (x .lt. 0.0) erf = 2.0 - erf

  end function erfApprox0

  !---------------------------------------------------------------------
  function erfApprox1(x) result(erf)
  !---------------------------------------------------------------------
  ! This approx uses the formula of Busen K M, Am J Phys 49, 886 (1981).

    implicit none

    real(single), intent(in) :: x
    real(single) :: erf, sign

    !-------------------------------------------------------------------
    ! Start

    sign = 1.0
    if (x < 0.0) sign = -1.0

    if (x > 1.0) then
      erf = sqrt(1.0 - exp(-4.0 * x**1.88 / Pi))
    else
      erf = sqrt(1.0 - exp(-4.0 * x * x / Pi))
    end if

    erf = erf * sign

  end function erfApprox1

  !---------------------------------------------------------------------
  function erfcApprox0(x) result(erfc)
  !---------------------------------------------------------------------
  !x    input	Deviate, range -infinity to +infinity.
  !erfc output	Area under Gaussian from X to +infinity.
  !-
  ! Author	Clive Page	1983 July 5
  ! Standard Fortran-77 version 1984 Sept 25, CGP.
  ! From algorithm AS66 by I.D.Hill, Appl Stats  22,424 (1973).
  ! Adapted to f90 by Ian Stewart 8 Aug 2000

    implicit none 

    real(single), intent(in) :: x
    real(single) :: erfc, y, z

    !-------------------------------------------------------------------
    ! Start

    z = abs(x)
    if(z > 13.0) then
      erfc = 0.0

    else
      y = 0.5 * z * z
      if(z <= 1.28) then
        erfc = 0.5 - z * (0.398942280444 - 0.399903438504 * y /&
        (y + 5.75885480458 - 29.8213557808 /&
        (y + 2.62433121679 + 48.6959930692 /&
        (y + 5.92885724438))))

      else
        erfc = 0.398942280385 * exp(-y) /&
        (z - 3.8052E-8      + 1.00000615302 /&
        (z + 3.98064794E-4  + 1.98615381364 /&
        (z - 0.151689116635 + 5.29330324926 /&
        (z + 4.8385912808   - 15.1508972451 /&
        (z + 0.742380924027 + 30.789933034  /&
        (z + 3.99019417011))))))
      end if
    end if

    if(x < 0.0) erfc = 1.0 - erfc

  end function erfcApprox0

  !---------------------------------------------------------------------
  function besselJ0(arg) result(J0)
  !---------------------------------------------------------------------
  ! Calculates Bessel function J0(x).

    implicit none

    real(double), intent(in) :: arg
    real(double)             :: J0

    integer, parameter :: k = 45
    real(double), parameter :: alpha = 1.0d-100
    real(double) :: convergeCriterion, term, mult, jd, P0, Q0, trial_J1, trial_J0
    integer :: n

!      parameter (precsn = 1.0d-14)

    !-------------------------------------------------------------------
    ! Start

!      include '~/includes/general0.inc' !Definition of pi etc.

    convergeCriterion = 3.0 ** (-precision(arg))

    if (arg < 1.0d0) then ! power series is most convenient.
      mult = arg * arg / 4.0d0
      term = 1.0d0
      jd = 1.0d0
      n = 1

      do
      if (ABS(term) < convergeCriterion) exit
        term = -(term * mult / real(n, double)) / real(n, double)
        jd = jd + term
        n = n + 1
      end do

      J0 = jd

    else if (arg > 16.0d0) then ! asymptotic series is better.
      call PQ0_calc(arg, P0, Q0)
      J0 = (P0 * cos(arg - Pi / 4.0d0) - Q0 * sin(arg - Pi / 4.0d0)) * sqrt(2.0d0 / Pi / arg)
    else ! value from recurrence relation is best.
!      J1 = trial_J1(arg, k, alpha, trial_J0)
      call calcTrialJ0J1(arg, alpha, k, trial_J0, trial_J1)
      J0 = trial_J0
    end if

  end function besselJ0

  !---------------------------------------------------------------------
  function besselJ0Zero(k) result(J0zero)
  !---------------------------------------------------------------------
  ! Calculates zero k of the Bessel function J0(x).

    implicit none

    integer,     intent(in) :: k
    real(double)            :: J0zero

    real(double) :: convergeCriterion, zero, inc
!, J0, J1

!      parameter (precsn = 1.0d-14)

    !-------------------------------------------------------------------
    ! Start

!      include '~/includes/general0.inc' !Definition of pi etc.

    convergeCriterion = 3.0 ** (-precision(J0zero))

    zero = k * Pi - 0.25d0
    inc = besselJ0(zero) / besselJ1(zero)
    zero = zero + inc
    do
    if (ABS(inc) < convergeCriterion) exit
      inc = besselJ0(zero) / besselJ1(zero)
      zero = zero + inc
    end do
    J0zero = zero

  end function besselJ0Zero

  !---------------------------------------------------------------------
  function besselJ1(arg) result(J1)
  !---------------------------------------------------------------------
  ! Calculates Bessel function J1(x).

    implicit none

    real(double), intent(in) :: arg
    real(double)             :: J1

    integer, parameter :: k = 45
    real(double), parameter :: alpha = 1.0d-100
    real(double) :: convergeCriterion, term, mult, jd, P1, Q1, trial_J1, trial_J0
    integer :: n

!      parameter (precsn = 1.0d-14)

    !-------------------------------------------------------------------
    ! Start

!      include '~/includes/general0.inc' !Definition of pi etc.

    convergeCriterion = 3.0 ** (-precision(arg))

    if (arg < 1.0d0) then ! power series is most convenient.
      mult = arg * arg / 4.0d0
      term = arg / 2.0d0
      jd = term
      n = 1

      do
      if (ABS(term) <= convergeCriterion) exit
        term = -(term * mult / real(n, double)) / real(n + 1, double)
        jd = jd + term
        n = n + 1
      end do

      J1 = jd

    else if (arg > 16.0d0) then ! asymptotic series is better.
      call PQ1_calc(arg, P1, Q1)
      J1 = (P1 * cos(arg - 0.75d0 * Pi) - Q1 * sin(arg - 0.75d0 * Pi)) * sqrt(2.0d0 / Pi / arg)
    else ! value from recurrence relation is best.
!      J1 = trial_J1(arg, k, alpha, trial_J0)
      call calcTrialJ0J1(arg, alpha, k, trial_J0, trial_J1)
      J1 = trial_J1
    end if

  end function besselJ1

  !---------------------------------------------------------------------
  function besselJ1Zero(k) result(J1zero)
  !---------------------------------------------------------------------
  ! Calculates zero k of the Bessel function J1(x). k should be >= 0.

    implicit none

    integer,     intent(in) :: k
    real(double)            :: J1zero

    real(double) :: convergeCriterion, zero, inc
!, J0, J1

!      parameter (precsn = 1.0d-14)

    !-------------------------------------------------------------------
    ! Start

!      include '~/includes/general0.inc' !Definition of pi etc.

    convergeCriterion = 3.0 ** (-precision(J1zero))

    if (k == 0) THEN
      J1zero = 0.0d0
    else
      zero = k * Pi + 0.25d0
      inc = besselJ1(zero) / (besselJ0(zero) - besselJ1(zero) / zero)
      zero = zero - inc
      do
      if (ABS(inc) < convergeCriterion) exit
        inc = besselJ1(zero) / (besselJ0(zero) - besselJ1(zero) / zero)
        zero = zero - inc
      end do
      J1zero = zero
    end if

  end function besselJ1Zero

  !---------------------------------------------------------------------
  subroutine calcTrialJ0J1(arg, alpha, k, trial_J0, trial_J1)
  !---------------------------------------------------------------------
  ! Calculates J0, J1 by the recurrence method of Goldstein M and Thaler
  ! R M, 1960 (don't know the journal).

    implicit none

    real(double), intent(in)  :: arg, alpha
    integer,      intent(in)  :: k
    real(double), intent(out) :: trial_J0, trial_J1

    integer :: p
    real(double) :: summ
    real(double), allocatable :: J(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(J(0:k+1))

    J(k + 1) = 0.0d0
    J(k) = alpha

    do p = k, 1, -1
      J(p - 1) = 2.0d0 * J(p) * p / arg - J(p + 1)
    end do

    summ = J(0)
    do p = 1, int(k / 2)
      summ = summ + 2.0d0 * J(2 * p)
    end do

    trial_J0 = J(0) / summ
    trial_J1 = J(1) / summ

  end subroutine calcTrialJ0J1

  !---------------------------------------------------------------------
  subroutine PQ0_calc(arg, Psum, Qsum)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: arg
    real(double), intent(out) :: Psum, Qsum

    real(double) :: seed, multiplier, oldterm, term
    integer :: n, sig

    !-------------------------------------------------------------------
    ! Start

    seed = 1.0d0
    oldterm = seed
    Psum = seed
    Qsum = 0.0d0

    sig = -1
    n = 1
    multiplier = sig * real(2 * n - 1, double) * real(2 * n - 1, double) / real(n, double) / 8.0d0 / arg
    term = oldterm * multiplier

    do
    if (abs(term) >= abs(oldterm)) exit

      oldterm = term
      Qsum = Qsum + term
      sig = -sig
      n = n + 1
      multiplier = sig * real(2 * n - 1, double) * real(2 * n - 1, double) / real(n, double) / 8.0d0 / arg
      term = oldterm * multiplier

    if (abs(term) >= abs(oldterm)) exit

      oldterm = term
      Psum = Psum + term
      sig = -sig
      n = n + 1
      multiplier = sig * real(2 * n - 1, double) * real(2 * n - 1, double) / real(n, double) / 8.0d0 / arg
      term = oldterm * multiplier
    end do

  end subroutine PQ0_calc

  !---------------------------------------------------------------------
  subroutine PQ1_calc(arg, Psum, Qsum)
  !---------------------------------------------------------------------
         
    implicit none

    real(double), intent(in)  :: arg
    real(double), intent(out) :: Psum, Qsum

    real(double) :: seed, multiplier, oldterm, term
    integer :: n, sig

    !-------------------------------------------------------------------
    ! Start

    seed = 1.0d0
    oldterm = seed
    Psum = seed
    Qsum = 0.0d0

    sig = -1
    n = 1
    multiplier = sig * real(2 * n - 3, double) * real(2 * n + 1, double) / real(n, double) / 8.0d0 / arg
    term = oldterm * multiplier

    do
    if (abs(term) >= abs(oldterm)) exit

      oldterm = term
      Qsum = Qsum + term
      sig = -sig
      n = n + 1
      multiplier = sig * real(2 * n - 3, double) * real(2 * n + 1, double) / real(n, double) / 8.0d0 / arg
      term = oldterm * multiplier

    if (abs(term) >= abs(oldterm)) exit

      oldterm = term
      Psum = Psum + term
      sig = -sig
      n = n + 1
      multiplier = sig * real(2 * n - 3, double) * real(2 * n + 1, double) / real(n, double) / 8.0d0 / arg
      term = oldterm * multiplier
    end do

  end subroutine PQ1_calc

  !---------------------------------------------------------------------
  real(double) function sinc(x)
  !---------------------------------------------------------------------
  ! I define sinc(x) as
  !
  !	sinc(x) = sin(x)/x.
  !
  ! Have to handle the singularity at x==0. Recall that Taylor series for
  ! sin(x) is
  !
  !	sin(x) = x - x^3/6 + x^5/120 + O(x^7).
  !
  ! Thus can approximate sin(x) by 1st two terms provided
  ! abs(x)^5/120 < the numeric precision.

    implicit none

    real(double), intent(in) :: x

    real(double) :: xCutoff

    !-------------------------------------------------------------------
    ! Start

    xCutoff = (120.0d0*epsilon(1.0d0))**0.2
    if (abs(x)<xCutoff) then
      sinc = 1.0d0 - x * x / 6.0d0
    else
      sinc = sin(x) / x
    end if

  end function sinc

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module math_utils
