!+
! Module: pgplot_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to allow convenient test plots to be made.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module pgplot_aux

  use dal
!  use param  
  use errorHandling

  ! ssclib modules:
  use constants
  use dal_aux
  use interpolations
  use sort_mod
  use ssc_misc	! for decideOption.

  implicit none 

  private 

  integer, public, parameter ::&
!    localNumColourPoints = 256,&
!    numReservedClrs = 3,&
    red = 1,&
    grn = 2,&
    blu = 3
  logical, public :: coloursHaveBeenSet=.false.

  type, public :: colourType
    real :: hue(3)
  end type colourType
  type, public :: colourMapType
    real :: wedgeFrac, brightFrac(red:blu)
  end type colourMapType
  type, public :: pinType
    real :: wedgeFrac, brightFrac, worldX, worldY
  end type pinType
  type, public :: pinsType
    type(pinType), pointer :: pins(:)
    type(pinType) :: oldBoundaryPins(2)
  end type pinsType
  type, public :: localRangeType
    real :: lo, hi
  end type localRangeType
!****** could we not use IntervalT?? No because need kind 'real' for pgplot.
  type, public :: boxFieldType
    type(localRangeType) :: vpX, vpY, worldX, worldY
    character(10) :: type ! eg 'button'
    integer :: id, colourIndex
  end type boxFieldType
  type, public :: buttonType
    type(boxFieldType) :: boxField
    character(100) :: text
!    logical :: inFocus
    integer :: textColourIndex
  end type buttonType

  real, public, dimension(6) ::&
    pgTrMatrix = (/0.0, 1.0, 0.0, 0.0, 0.0, 1.0/)

  integer, parameter, public ::&
    P_POINT=-1, P_TRIANGLE=7, P_SQUARE=0, P_CIRCLE=4, P_CROSS=2,&
    P_DIAMOND=11, P_XCROSS=5, P_ASTERIX=3, P_FILLEDTRIANGLE=13,&
    P_FILLEDSQUARE=16, P_FILLEDCIRCLE=17,&
    L_SOLID=1, L_DASHED=2, L_DOTTED=4,&
    C_WHITE=1, C_RED=2, C_BLUE=3, C_GREEN=4,&
    C_BACKGROUND=0, C_FOREGROUND=1, C_HIGHLIGHT=2, C_UNFOCUSSED=3

  ! Local variables:
  !
  logical ::&
    plotInitialized=.false., plotAutoSizedX=.false., plotAutoSizedY=.false.,&
    plotLogX=.false., plotLogY=.false., worldRangesSet=.false.,&
    imageAutoScaled=.false., imageAutoSizedXY=.false.
!, coloursHaveBeenSet=.false.
  integer ::&
    numPreviousPlots=0,&
!    localNumColourPoints = 256,&
    numReservedClrs = 4
!    minCi=0,maxCi=0,ciLo=0,ciHi=0,&
  character(100) :: xLabel='', yLabel='', plotLabel='', defaultDevice='/xs'
  real ::&
    imageMaxVal=0.0, imageMinVal=0.0,&
!    worldXLo=0.0, worldXHi=1.0, worldYLo=0.0, worldYHi=1.0,&
    worldMinDataX=0.0, worldMaxDataX=1.0, worldMinDataY=0.0, worldMaxDataY=1.0

  real, parameter ::&
    defaultVpXLo=0.1, defaultVpXHi=0.99, defaultVpYLo=0.1, defaultVpYHi=0.9,&
    xPadFrac = 0.05, yPadFrac = 0.05

  public initializeBoxField
  public initializeButton

  public plotInit

  public setXLog, setYLog
  public plotAutoSizeX, plotAutoSizeY
  public drawPointPlot
  public drawLinePlot
  public drawHistogram
  private checkRangesForXYPlot
  private setWorldRangesForXYPlot
  public finishPlot

  public imageAutoScale
  public setWorldRangesForImage
  public drawImage
  public finishImage

  public inBoxField
  public vpToWorld
  public boxVpToWorld
  public drawBoxes
  public drawButtons
  public divideUpRange
  public getIndexOfSelectedBox

  public drawArc

  public readColourMap_obsolete
  public writeColourMap_obsolete
  public readColourMap
  public setColourMap
  public writeColourMap

  public initializePinType
  public deallocatePins
  public loadColourPins
  public saveColourPins
  public calcPinsWorldCoords
  public convertPinsToColourMap
  private convertColourMapToPins
  private secondDerivative

  interface setColourMap
    module procedure setColourMapFromMap, setColourMapFromName
  end interface

  interface initializeBoxField
    module procedure initializeBoxFieldScalar, initializeBoxFieldVector
  end interface
  interface initializeButton
    module procedure initializeButtonScalar, initializeButtonVector
  end interface

  interface drawPointPlot
    module procedure drawPointPlotXY, drawPointPlotY
  end interface
  interface drawLinePlot
    module procedure drawLinePlotXY, drawLinePlotY
  end interface
  interface boxVpToWorld
    module procedure boxVpToWorldScalar, boxVpToWorldVector
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine initializeBoxFieldScalar(box)
  !---------------------------------------------------------------------

    implicit none

    type(boxFieldType), intent(out) :: box

    type(boxFieldType) :: boxes(1)

    !-------------------------------------------------------------------
    ! Start

    call initializeBoxFieldVector(boxes)
    box = boxes(1)

  end subroutine initializeBoxFieldScalar

  !---------------------------------------------------------------------
  subroutine initializeBoxFieldVector(boxes)
  !---------------------------------------------------------------------

    implicit none

    type(boxFieldType), intent(out) :: boxes(:)

!    integer :: i

    !-------------------------------------------------------------------
    ! Start

    boxes%vpX%lo = 0.0
    boxes%vpX%hi = 0.0
    boxes%vpY%lo = 0.0
    boxes%vpY%hi = 0.0
    boxes%worldX%lo = 0.0
    boxes%worldX%hi = 0.0
    boxes%worldY%lo = 0.0
    boxes%worldY%hi = 0.0
    boxes%type = ''
    boxes%id = 0
    boxes%colourIndex = 1

  end subroutine initializeBoxFieldVector

  !---------------------------------------------------------------------
  subroutine initializeButtonScalar(button)
  !---------------------------------------------------------------------

    implicit none

    type(buttonType), intent(out) :: button

    type(buttonType) :: buttons(1)

    !-------------------------------------------------------------------
    ! Start

    call initializeButtonVector(buttons)
    button = buttons(1)

  end subroutine initializeButtonScalar

  !---------------------------------------------------------------------
  subroutine initializeButtonVector(buttons)
  !---------------------------------------------------------------------

    implicit none

    type(buttonType), intent(out) :: buttons(:)

!    integer :: i

    !-------------------------------------------------------------------
    ! Start

    call initializeBoxFieldVector(buttons%boxField)
    buttons%text = ''
!    buttons%inFocus = .true.
    buttons%textColourIndex = buttons%boxField%colourIndex

  end subroutine initializeButtonVector


  !---------------------------------------------------------------------
  subroutine plotInit(pgplotDevice, xLabelArg, yLabelArg, plotLabelArg&
  , numReservedClrsArg)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in), optional :: pgplotDevice,&
                                          xLabelArg,&
                                          yLabelArg,&
                                          plotLabelArg
    integer,      intent(in), optional :: numReservedClrsArg

    integer :: deviceId, pgopen, minCi, maxCi, ciLo, ciHi
    character(256) :: localPgplotDevice

    !-------------------------------------------------------------------
    ! Start:

    if (plotInitialized)&
call error('pgplotAlreadyInitialized',"")

    localPgplotDevice = decideOption(pgplotDevice, defaultDevice)

    deviceId = pgopen(trim(localPgplotDevice))
if (deviceId <= 0) call error('badPgplotDevice',"")

    if (present(numReservedClrsArg))&
      numReservedClrs = numReservedClrsArg

    call pgsvp(defaultVpXLo, defaultVpXHi, defaultVpYLo, defaultVpYHi)
    plotInitialized = .true.

    if (present(xLabelArg))&
      xLabel = xLabelArg
    if (present(yLabelArg))&
      yLabel = yLabelArg
    if (present(plotLabelArg))&
      plotLabel = plotLabelArg

!    call pgqcir(minCi, maxCi)
    call pgqcol(minCi, maxCi)
    ciLo = minCi + numReservedClrs
    ciHi = maxCi
    call pgscir(ciLo, ciHi)

  end subroutine plotInit

  !---------------------------------------------------------------------
  subroutine setXLog
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    plotLogX = .true.

  end subroutine setXLog

  !---------------------------------------------------------------------
  subroutine setYLog
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    plotLogY = .true.

  end subroutine setYLog

  !---------------------------------------------------------------------
  subroutine plotAutoSizeX(xVals, xErrs)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xVals(:)
    real(single), intent(in), optional :: xErrs(size(xVals))

    real :: newXLo, newXHi
!, xPadSize
    logical :: xRangesSet
    logical, allocatable :: xIsOk(:)

    !-------------------------------------------------------------------
    ! Start:

    if (.not.plotInitialized) then
!      call warning('pgplotNotInitialized',"You hadn't set up PGPLOT before &
!        &making a PGPLOT-dependent call. I've initialized the output device, &
!        &the viewport size, the labels and the colour index range to default &
!        &values. If that's indeed what you meant me to do, then no problem. &
!        &But if not, you might want to call plotInit before making the present &
!        &call.")
      call plotInit
!call error('pgplotNotInitialized',"")
    end if

    allocate(xIsOk(size(xVals)))
    if (plotLogX) then
      xIsOk = (xVals > 0.0)
    else
      xIsOk = .true.
    end if

    xRangesSet = .false.
    if (count(xIsOk)>0) then
      if (present(xErrs)) then
!!!        if (all(xVals==xVals(firstOk), xIsOk) .and. all(xErrs==xErrs(firstOk), xIsOk))&
!!!call error('allValidXsAndErrsAreEqual',"")

        if (plotLogX) then
          newXLo = minval(xVals, xIsOk)
        else
          newXLo = minval(xVals-xErrs, xIsOk)
        end if
        newXHi = maxval(xVals+xErrs, xIsOk)
      else
!!!        if (all(xVals==xVals(firstOk), xIsOk))&
!!!call error('allValidXsAreEqual',"")

        newXLo = minval(xVals, xIsOk)
        newXHi = maxval(xVals, xIsOk)
!print *, 'newXLo, newXHi:', newXLo, newXHi
      end if
      xRangesSet = .true.
    end if
!print *, 'xRangesSet?', xRangesSet

    if (plotAutoSizedX) then
!      call pgqwin(xLo, xHi, yLo, yHi)
      if (xRangesSet) then
        if (plotLogX .and. worldMinDataX <= 0.0) then
          worldMinDataX = newXLo
        else
          worldMinDataX = min(worldMinDataX, newXLo)
        end if
        worldMaxDataX = max(worldMaxDataX, newXHi)
        worldRangesSet = .false.
      end if

      if (plotLogX) then
        worldMinDataX = log10(worldMinDataX)
        worldMaxDataX = log10(worldMaxDataX)
      end if

    else
      if (xRangesSet) then
        if (plotLogX) then
          worldMinDataX = log10(newXLo)
          worldMaxDataX = log10(newXHi)
        else
          worldMinDataX = newXLo
          worldMaxDataX = newXHi
!print *, 'worldMinDataX, worldMaxDataX:', worldMinDataX, worldMaxDataX
        end if
      else
!*** really should be an error.
        worldMinDataX = 0.0
        worldMaxDataX = 1.0
      end if
    end if

    if (worldMinDataX >= worldMaxDataX)&
call error('badOrderXBounds',"")

!    xPadSize = (worldMaxDataX - worldMinDataX) * xPadFrac / (1.0d0 - 2.0d0 * xPadFrac)

!    worldMinDataX = worldMinDataX - xPadSize
!    worldMaxDataX = worldMaxDataX + xPadSize
!print *, 'worldMinDataX, worldMaxDataX after pad applied:', worldMinDataX, worldMaxDataX

!    call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)

    plotAutoSizedX = .true.

  end subroutine plotAutoSizeX

  !---------------------------------------------------------------------
  subroutine plotAutoSizeY(yVals, yErrs)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: yVals(:)
    real(single), intent(in), optional :: yErrs(size(yVals))

    real :: newYLo, newYHi
!, yPadSize
    logical :: yRangesSet
    logical, allocatable :: yIsOk(:)
logical :: testFlag=.false.
integer :: i

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa plotAutoSizeY'

if (testFlag) then
  do i = 1, size(yVals)
    print *, i, yVals(i)
  end do
end if

    if (.not.plotInitialized) then
!      call warning('pgplotNotInitialized',"You hadn't set up PGPLOT before &
!        &making a PGPLOT-dependent call. I've initialized the output device, &
!        &the viewport size, the labels and the colour index range to default &
!        &values. If that's indeed what you meant me to do, then no problem. &
!        &But if not, you might want to call plotInit before making the present &
!        &call.")
      call plotInit
!call error('pgplotNotInitialized',"")
    end if

    allocate(yIsOk(size(yVals)))
    if (plotLogY) then
      yIsOk = (yVals > 0.0)
    else
      yIsOk = .true.
    end if

    yRangesSet = .false.
    if (count(yIsOk)>0) then
      if (present(yErrs)) then
        if (plotLogY) then
          newYLo = minval(yVals, yIsOk)
        else
          newYLo = minval(yVals-yErrs, yIsOk)
        end if
        newYHi = maxval(yVals+yErrs, yIsOk)
      else
        newYLo = minval(yVals, yIsOk)
        newYHi = maxval(yVals, yIsOk)
if (testFlag) print *, '  bbb - newYLo, newYHi:', newYLo, newYHi
      end if
      yRangesSet = .true.
    end if

    if (plotAutoSizedY) then
!      call pgqwin(yLo, yHi, yLo, yHi)
      if (yRangesSet) then
        if (plotLogY .and. worldMinDataY <= 0.0) then
          worldMinDataY = newYLo
        else
          worldMinDataY = min(worldMinDataY, newYLo)
        end if
        worldMaxDataY = max(worldMaxDataY, newYHi)
        worldRangesSet = .false.
if (testFlag) print *, '  ccc - 2nd run - worldMinDataY, worldMaxDataY:', worldMinDataY, worldMaxDataY
      end if

      if (plotLogY) then
        worldMinDataY = log10(worldMinDataY)
        worldMaxDataY = log10(worldMaxDataY)
      end if

    else
      if (yRangesSet) then
        if (plotLogY) then
          worldMinDataY = log10(newYLo)
          worldMaxDataY = log10(newYHi)
        else
          worldMinDataY = newYLo
          worldMaxDataY = newYHi
if (testFlag) print *, '  ccc - 1st run - worldMinDataY, worldMaxDataY:', worldMinDataY, worldMaxDataY
        end if
      else
        worldMinDataY = 0.0
        worldMaxDataY = 1.0
      end if
    end if

    if (worldMinDataY >= worldMaxDataY)&
call error('badOrderYBounds',"")

!    yPadSize = (worldMaxDataY - worldMinDataY) * yPadFrac / (1.0d0 - 2.0d0 * yPadFrac)

!    worldMinDataY = worldMinDataY - yPadSize
!    worldMaxDataY = worldMaxDataY + yPadSize

!    call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)

    plotAutoSizedY = .true.
if (testFlag) print *, '  zzz plotAutoSizeY'

  end subroutine plotAutoSizeY

  !---------------------------------------------------------------------
  subroutine drawPointPlotY(yVals, pointStyle, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: yVals(:)
    integer,      intent(in), optional :: pointStyle
    logical,      intent(in), optional :: finished

    real(single), allocatable :: xVals(:)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    allocate(xVals(size(yVals)))
    xVals = (/(real(i-1, single), i=1,size(yVals))/)
    call drawPointPlotXY(xVals, yVals, pointStyle=pointStyle, finished=finished)

  end subroutine drawPointPlotY

  !---------------------------------------------------------------------
!  subroutine drawPointPlot(xVals, yVals, xErrs, yErrs, pointStyle, finished)
  subroutine drawPointPlotXY(xVals, yVals, yErrs, pointStyle, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xVals(:), yVals(size(xVals))
!    real(single), intent(in), optional :: xErrs(size(xVals)), yErrs(size(xVals))
    real(single), intent(in), optional :: yErrs(size(xVals))
    integer,      intent(in), optional :: pointStyle
    logical,      intent(in), optional :: finished

    logical :: localFinished
    integer :: localPointStyle, numOkVals
!    logical, allocatable :: xIsOk(:), yIsOk(:), valIsOk
    real, pointer ::&
      localXVals(:), localYVals(:), localYBots(:), localYTops(:)
!integer :: i

    !-------------------------------------------------------------------
    ! Start:

!!    if (.not.plotAutoSized) call plotAutoSize(xVals, yVals, xErrs, yErrs)
!!    if (.not.plotAutoSized) call plotAutoSizeXY(xVals, yVals, yErrs=yErrs)
!if (.not.(plotAutoSizedX .and. plotAutoSizedY)) then
!  if (.not.plotAutoSizedX) call plotAutoSizeX(xVals)
!  if (.not.plotAutoSizedY) call plotAutoSizeY(yVals, yErrs)
!!  call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)
!  call setWorldRangesForXYPlot
!end if

call checkRangesForXYPlot(xVals, yVals, yErrs)

    localFinished = decideOption(finished, .true.) ! in ssclib/ssc_misc.
    localPointStyle = decideOption(pointStyle, P_POINT)

    call pgsci(numPreviousPlots+1)
    numPreviousPlots = numPreviousPlots + 1

    call getLocalVals(xVals, yVals, yErrs, localXVals, localYVals&
      , localYBots, localYTops)
    numOkVals = size(localXVals)
!do i = 1, numOkVals
!  print *, i, localXVals(i), localYBots(i), localYTops(i)
!end do

    if (numOkVals > 0) then
      call pgpt(numOkVals, localXVals, localYVals, localPointStyle)
      if (present(yErrs))&
        call pgerry(numOkVals, localXVals, localYBots, localYTops, 1.0)
    end if

    if (localFinished) call finishPlot
    deallocate(localXVals, localYVals, localYBots, localYTops)

  end subroutine drawPointPlotXY

  !---------------------------------------------------------------------
  subroutine drawLinePlotY(yVals, lineStyle, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: yVals(:)
    integer,      intent(in), optional :: lineStyle
    logical,      intent(in), optional :: finished

    real(single), allocatable :: xVals(:)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    allocate(xVals(size(yVals)))
    xVals = (/(real(i-1, single), i=1,size(yVals))/)
    call drawLinePlotXY(xVals, yVals, lineStyle, finished)

  end subroutine drawLinePlotY

  !---------------------------------------------------------------------
  subroutine drawLinePlotXY(xVals, yVals, lineStyle, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xVals(:), yVals(size(xVals))
    integer,      intent(in), optional :: lineStyle
    logical,      intent(in), optional :: finished

    logical :: localFinished
    integer :: localLineStyle, oldLineStyle, numOkVals
    real, pointer ::&
      localXVals(:), localYVals(:)

    !-------------------------------------------------------------------
    ! Start:

!!    if (.not.plotAutoSized) call plotAutoSizeXY(xVals, yVals)
!!print *, 'plotAutoSizedX?', plotAutoSizedX
!!print *, 'plotAutoSizedY?', plotAutoSizedY
!if (.not.(plotAutoSizedX .and. plotAutoSizedY)) then
!!print *, minval(xVals), maxval(xVals)
!  if (.not.plotAutoSizedX) call plotAutoSizeX(xVals)
!  if (.not.plotAutoSizedY) call plotAutoSizeY(yVals)
!!  call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)
!  call setWorldRangesForXYPlot
!end if

call checkRangesForXYPlot(xVals, yVals)

    localFinished = decideOption(finished, .true.) ! in ssclib/ssc_misc.
    localLineStyle = decideOption(lineStyle, L_SOLID)

    call pgqls(oldLineStyle)
    call pgsls(localLineStyle)

    call pgsci(numPreviousPlots+1)
    numPreviousPlots = numPreviousPlots + 1

    call getLocalVals(xVals, yVals, localXVals=localXVals&
      , localYVals=localYVals)

    numOkVals = size(localXVals)
    if (numOkVals > 0)&
      call pgline(numOkVals, localXVals, localYVals)
!    numPoints = size(xVals)
!    call pgline(numPoints, xVals, yVals)

    call pgsls(oldLineStyle)
    if (localFinished) call finishPlot
    deallocate(localXVals, localYVals)

  end subroutine drawLinePlotXY

  !---------------------------------------------------------------------
  subroutine drawHistogram(binEdges, binVals, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: binVals(:)
    real(single), intent(in)           :: binEdges(size(binVals)+1)
    logical,      intent(in), optional :: finished

    logical :: localFinished
    integer :: numBins, i
    real :: someXTooLowToSeeOnPlot, someYTooLowToSeeOnPlot, baseY
    real, allocatable :: xs(:), ys(:)

    !-------------------------------------------------------------------
    ! Start:

!!    if (.not.plotAutoSized) call plotAutoSize(binEdges, binVals)
!if (.not.(plotAutoSizedX .and. plotAutoSizedY)) then
!  if (.not.plotAutoSizedX) call plotAutoSizeX(binEdges)
!  if (.not.plotAutoSizedY) call plotAutoSizeY(binVals)
!!  call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)
!  call setWorldRangesForXYPlot
!end if

call checkRangesForXYPlot(binEdges, binVals)

    localFinished = decideOption(finished, .true.) ! in ssclib/ssc_misc.

    numBins = size(binVals)
if (any(binEdges(1:numBins)>binEdges(2:))) call error('badOrderBinEdges',"")

    call pgsci(numPreviousPlots+1)
    numPreviousPlots = numPreviousPlots + 1

    allocate(xs(numBins+1), ys(numBins))

    if (plotLogX) then
      someXTooLowToSeeOnPlot = getTooLowX()
      where(binEdges <= 0.0)
        xs = someXTooLowToSeeOnPlot
      elsewhere
        xs = log10(binEdges)
      end where
    else
      xs = binEdges
    end if

    if (plotLogY) then
      someYTooLowToSeeOnPlot = getTooLowY()
      where(binVals <= 0.0)
        ys = someYTooLowToSeeOnPlot
      elsewhere
        ys = log10(binVals)
      end where
      baseY = someYTooLowToSeeOnPlot
    else
      ys = binVals
      baseY = 0.0
    end if

    call pgmove(xs(1), baseY)
    do i = 1, numBins
      call pgdraw(xs(i), ys(i))
      call pgmove(xs(i), ys(i))
      call pgdraw(xs(i+1), ys(i))
      call pgmove(xs(i+1), ys(i))
    end do
    call pgdraw(xs(numBins+1),baseY)

    if (localFinished) call finishPlot

  end subroutine drawHistogram

  !---------------------------------------------------------------------
  subroutine checkRangesForXYPlot(xs, ys, yErrs)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xs(:), ys(:)
    real(single), intent(in), optional :: yErrs(size(ys))

    !-------------------------------------------------------------------
    ! Start

    if (.not.(plotAutoSizedX .and. plotAutoSizedY)) then
      if (.not.plotAutoSizedX) call plotAutoSizeX(xs)
      if (.not.plotAutoSizedY) call plotAutoSizeY(ys, yErrs)
      call setWorldRangesForXYPlot
    else if (.not.worldRangesSet) then
      call setWorldRangesForXYPlot
    end if

  end subroutine checkRangesForXYPlot

  !---------------------------------------------------------------------
  subroutine setWorldRangesForXYPlot
  !---------------------------------------------------------------------

    implicit none

    real :: xPadSize, yPadSize,&
      worldXLo, worldXHi, worldYLo, worldYHi

    !-------------------------------------------------------------------
    ! Start:

if (worldMinDataX >= worldMaxDataX) call error('badOrderXBounds',"")
if (worldMinDataY >= worldMaxDataY) call error('badOrderYBounds',"")

    xPadSize = (worldMaxDataX - worldMinDataX) * xPadFrac / (1.0d0 - 2.0d0 * xPadFrac)
    worldXLo = worldMinDataX - xPadSize
    worldXHi = worldMaxDataX + xPadSize

    yPadSize = (worldMaxDataY - worldMinDataY) * yPadFrac / (1.0d0 - 2.0d0 * yPadFrac)
    worldYLo = worldMinDataY - yPadSize
    worldYHi = worldMaxDataY + yPadSize

    call pgswin(worldXLo, worldXHi, worldYLo, worldYHi)

    worldRangesSet = .true.

  end subroutine setWorldRangesForXYPlot

  !---------------------------------------------------------------------
  subroutine getLocalVals(xVals, yVals, yErrs, localXVals, localYVals&
  , localYBots, localYTops)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xVals(:), yVals(size(xVals))
!    real(single), intent(in), optional :: xErrs(size(xVals)), yErrs(size(xVals))
    real(single), intent(in), optional :: yErrs(size(xVals))
    real,         pointer              :: localXVals(:),&
                                          localYVals(:)
    real,         pointer,    optional :: localYBots(:),&
                                          localYTops(:)

    integer :: numPoints, numOkVals
    logical, allocatable :: xIsOk(:), yIsOk(:), valIsOk(:)
    real, allocatable :: localYErrs(:)
    real :: someYTooLowToSeeOnPlot
!integer :: i

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(xVals)
    allocate(xIsOk(numPoints), yIsOk(numPoints), valIsOk(numPoints))
    if (plotLogX) then
      xIsOk = (xVals > 0.0)
    else
      xIsOk = .true.
    end if
    if (plotLogY) then
      yIsOk = (yVals > 0.0)
    else
      yIsOk = .true.
    end if
    valIsOk = (xIsOk .and. yIsOk)

    numOkVals = count(valIsOk)
    allocate(localXVals(numOkVals), localYVals(numOkVals))
    localXVals = pack(xVals, valIsOk)
    localYVals = pack(yVals, valIsOk)

    if (present(yErrs)) then
if (.not.present(localYBots) .or. .not.present(localYTops)) call error('bug',"pgplot_aux:")
      allocate(localYErrs(numOkVals))
      allocate(localYBots(numOkVals), localYTops(numOkVals))
      localYErrs = pack(yErrs, valIsOk)
      localYBots = localYVals - localYErrs
      localYTops = localYVals + localYErrs
!do i = 1, numOkVals
!  print *, i, localYErrs(i), localYBots(i), localYTops(i)
!end do
      if (plotLogY) then
        localYTops = log10(localYTops)
        someYTooLowToSeeOnPlot = getTooLowY()
        where(localYBots>0.0)
          localYBots = log10(localYBots)
        elsewhere
          localYBots = someYTooLowToSeeOnPlot
        end where 
      end if
    end if

    if (plotLogX) localXVals = log10(localXVals)
    if (plotLogY) localYVals = log10(localYVals)

  end subroutine getLocalVals

  !---------------------------------------------------------------------
  subroutine finishPlot
  !---------------------------------------------------------------------

    implicit none

    character(10) :: xOpt, yOpt

    !-------------------------------------------------------------------
    ! Start:

    if (plotLogX) then
      xOpt = 'BCNTSL'
    else
      xOpt = 'BCNTS'
    end if
    if (plotLogY) then
      yOpt = 'BCNTSL'
    else
      yOpt = 'BCNTS'
    end if
    call pgsci(1)
    call pgsfs(2)
    call pgbox(xOpt, 0.0, 0, yOpt, 0.0, 0)
    call pglab(xLabel, yLabel, plotLabel)
    call pgend
    plotInitialized = .false.

  end subroutine finishPlot

  !---------------------------------------------------------------------
  real function getTooLowX()
  !---------------------------------------------------------------------

    implicit none

    real :: minx, maxx, miny, maxy

    !-------------------------------------------------------------------
    ! Start:

    call pgqwin(minx, maxx, miny, maxy)
    getTooLowX = minx - (maxx - minx)

  end function getTooLowX

  !---------------------------------------------------------------------
  real function getTooLowY()
  !---------------------------------------------------------------------

    implicit none

    real :: minx, maxx, miny, maxy

    !-------------------------------------------------------------------
    ! Start:

    call pgqwin(minx, maxx, miny, maxy)
    getTooLowY = miny - (maxy - miny)

  end function getTooLowY


  !---------------------------------------------------------------------
  subroutine imageAutoScale(array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: array(:,:)

!    integer :: xiSize, yiSize
!    real :: xLo, xHi, yLo, yHi

    !-------------------------------------------------------------------
    ! Start:

!    if (.not.plotInitialized)&
!call error('pgplotNotInitialized',"")

    imageMinVal = minval(array)
    imageMaxVal = maxval(array)

!    xiSize = size(array,1)
!    yiSize = size(array,2)

!    xLo = 0.5
!    xHi = xiSize + 0.5
!    yLo = 0.5
!    yHi = yiSize + 0.5

!    call pgswin(xLo, xHi, yLo, yHi)
!    call pgwnad(xLo, xHi, yLo, yHi)

    imageAutoScaled = .true.

  end subroutine imageAutoScale

  !---------------------------------------------------------------------
  subroutine setWorldRangesForImage(array, box)
  !---------------------------------------------------------------------
  ! This prepares pgplot for plotting an image via the routines pgimag or
  ! pggray. Specifically, this routine sets up the correct world XY
  ! coordinates; intensity scaling is performed by imageAutoScale.

    implicit none

    real(single),       intent(in)           :: array(:,:)
    type(boxFieldType), intent(in), optional :: box

    integer :: xiSize, yiSize
    real ::&
      worldXLo, worldXHi, worldYLo, worldYHi,&
      presentVpXMin, presentVpXMax, presentVpYMin, presentVpYMax,&
      newWorldXLo, newWorldXHi, newWorldYLo, newWorldYHi

    !-------------------------------------------------------------------
    ! Start:

    if (.not.plotInitialized) then
!      call warning('pgplotNotInitialized',"You hadn't set up PGPLOT before &
!        &making a PGPLOT-dependent call. I've initialized the output device, &
!        &the viewport size, the labels and the colour index range to default &
!        &values. If that's indeed what you meant me to do, then no problem. &
!        &But if not, you might want to call plotInit before making the present &
!        &call.")
      call plotInit
!call error('pgplotNotInitialized',"")
    end if

    xiSize = size(array,1)
    yiSize = size(array,2)

    worldXLo = 0.5
    worldXHi = xiSize + 0.5
    worldYLo = 0.5
    worldYHi = yiSize + 0.5

    if (present(box)) then ! we need to stretch the world coordinate limits
    ! so that the image is plotted not to fill the viewport, but only the box.
if (box%vpX%lo >= box%vpX%hi) call error('badBoxXBounds',"")
if (box%vpY%lo >= box%vpY%hi) call error('badBoxYBounds',"")
      call pgqvp(0, presentVpXMin, presentVpXMax, presentVpYMin, presentVpYMax)

      newWorldXLo = fracVal(worldXLo, worldXHi&
        , (presentVpXMin - box%vpX%lo) / (box%vpX%hi - box%vpX%lo))
      newWorldXHi = fracVal(worldXLo, worldXHi&
        , (presentVpXMax - box%vpX%lo) / (box%vpX%hi - box%vpX%lo))
      newWorldYLo = fracVal(worldYLo, worldYHi&
        , (presentVpYMin - box%vpY%lo) / (box%vpY%hi - box%vpY%lo))
      newWorldYHi = fracVal(worldYLo, worldYHi&
        , (presentVpYMax - box%vpY%lo) / (box%vpY%hi - box%vpY%lo))

      call pgswin(newWorldXLo, newWorldXHi, newWorldYLo, newWorldYHi)

    else
      call pgswin(worldXLo, worldXHi, worldYLo, worldYHi)
!      call pgwnad(xLo, xHi, yLo, yHi)
    end if

    imageAutoSizedXY = .true.

  end subroutine setWorldRangesForImage

  !---------------------------------------------------------------------
  subroutine drawImage(array, box, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single),       intent(in)           :: array(:,:)
    type(boxFieldType), intent(in), optional :: box
    logical,            intent(in), optional :: finished

    logical :: localFinished
    integer :: xiSize, yiSize

    !-------------------------------------------------------------------
    ! Start:

    if (.not.imageAutoScaled) call imageAutoScale(array)

    if (.not.(imageAutoSizedXY)) then
      call setWorldRangesForImage(array, box)
    end if

    localFinished = decideOption(finished, .true.) ! in ssclib/ssc_misc.

    xiSize = size(array,1)
    yiSize = size(array,2)

    if (coloursHaveBeenSet) then
      call pgimag(real(array), xiSize, yiSize, 1, xiSize, 1, yiSize&
        , imageMaxVal, imageMinVal, pgTrMatrix)
    else
      call pggray(real(array), xiSize, yiSize, 1, xiSize, 1, yiSize&
        , imageMaxVal, imageMinVal, pgTrMatrix)
    end if

    if (localFinished) call finishImage

  end subroutine drawImage

  !---------------------------------------------------------------------
  subroutine finishImage
  !---------------------------------------------------------------------

    implicit none

    character(10) :: xOpt, yOpt

    !-------------------------------------------------------------------
    ! Start:

    xOpt = 'BC'
    yOpt = 'BC'

    call pgsci(1)
    call pgsfs(2)
    call pgbox(xOpt, 0.0, 0, yOpt, 0.0, 0)
    call pglab(xLabel, yLabel,'')
    call pgend
    plotInitialized = .false.

  end subroutine finishImage

  !---------------------------------------------------------------------
  logical function inBoxField(worldX, worldY, box)
  !---------------------------------------------------------------------

    implicit none

    real,               intent(in) :: worldX, worldY
    type(boxFieldType), intent(in) :: box

    !-------------------------------------------------------------------
    ! Start

    inBoxField = .false. ! default
    if (  worldX>=box%worldX%lo .and. worldX<box%worldX%hi&
    .and. worldY>=box%worldY%lo .and. worldY<box%worldY%hi)&
      inBoxField = .true.

  end function inBoxField

  !---------------------------------------------------------------------
  subroutine vpToWorld(vpX, vpY, worldX, worldY)
  !---------------------------------------------------------------------
  ! PGPLOT defines minimum and maximum values, in X and Y directions, of
  ! two related coordinate systems: the viewport (vp) and world coordinate
  ! systems. The offset and scaling between the two systems is such that
  ! worldMin is aligned with vpMin and worldMax with vpMax. Eg graphically:
  !
  !	vp:     0     vpMin                      vpMax              1
  !	        |-------|--------------------------|----------------|
  !	
  !	    ------------|--------------------------|--------------------
  !	world:      worldMin                   worldMax
  !
  ! The present subroutine converts between the two.

    implicit none

    real, intent(in)  :: vpX, vpY
    real, intent(out) :: worldX, worldY

    real ::&
      presentVpXMin, presentVpXMax, presentVpYMin, presentVpYMax,&
      presentWorldXMin, presentWorldXMax, presentWorldYMin, presentWorldYMax

    !-------------------------------------------------------------------
    ! Start:

    if (.not.plotInitialized) then
!      call warning('pgplotNotInitialized',"You hadn't set up PGPLOT before &
!        &making a PGPLOT-dependent call. I've initialized the output device, &
!        &the viewport size, the labels and the colour index range to default &
!        &values. If that's indeed what you meant me to do, then no problem. &
!        &But if not, you might want to call plotInit before making the present &
!        &call.")
      call plotInit
!call error('pgplotNotInitialized',"")
    end if

    call pgqvp(0, presentVpXMin, presentVpXMax, presentVpYMin, presentVpYMax)
if (presentVpXMin >= presentVpXMax) call error('badViewportXBounds',"")
if (presentVpYMin >= presentVpYMax) call error('badViewportYBounds',"")

    call pgqwin(presentWorldXMin, presentWorldXMax&
      , presentWorldYMin, presentWorldYMax)

    worldX = fracVal(presentWorldXMin, presentWorldXMax&
      , (vpX - presentVpXMin) / (presentVpXMax - presentVpXMin))
    worldY = fracVal(presentWorldYMin, presentWorldYMax&
      , (vpY - presentVpYMin) / (presentVpYMax - presentVpYMin))

  end subroutine vpToWorld

  !---------------------------------------------------------------------
  subroutine boxVpToWorldScalar(box)
  !---------------------------------------------------------------------
  ! It is convenient to describe boxes in terms of viewport (vp)
  ! coordinates; however if interactions with a cursor are desired, it is
  ! better to have the box coordinates in the world system, because
  ! cursor coordinates are returned in that system. (Of course, when one
  ! draws the box, one also needs world coords.) Hence this conversion
  ! routine.

    implicit none

    type(boxFieldType), intent(inout) :: box

    !-------------------------------------------------------------------
    ! Start:

    call vpToWorld(box%vpX%lo, box%vpY%lo, box%worldX%lo, box%worldY%lo)
    call vpToWorld(box%vpX%hi, box%vpY%hi, box%worldX%hi, box%worldY%hi)

  end subroutine boxVpToWorldScalar

  !---------------------------------------------------------------------
  subroutine boxVpToWorldVector(boxes)
  !---------------------------------------------------------------------
  ! It is convenient to describe boxes in terms of viewport (vp)
  ! coordinates; however if interactions with a cursor are desired, it is
  ! better to have the box coordinates in the world system, because
  ! cursor coordinates are returned in that system. (Of course, when one
  ! draws the box, one also needs world coords.) Hence this conversion
  ! routine.

    implicit none

    type(boxFieldType), intent(inout) :: boxes(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    do i = 1, size(boxes)
      call vpToWorld(boxes(i)%vpX%lo, boxes(i)%vpY%lo&
        , boxes(i)%worldX%lo, boxes(i)%worldY%lo)
      call vpToWorld(boxes(i)%vpX%hi, boxes(i)%vpY%hi&
        , boxes(i)%worldX%hi, boxes(i)%worldY%hi)
    end do

  end subroutine boxVpToWorldVector

  !---------------------------------------------------------------------
  subroutine drawBoxes(boxes)
  !---------------------------------------------------------------------

    implicit none

    type(boxFieldType), intent(in) :: boxes(:)

    integer :: i, oldFillStyle, oldColourIndex
!    real :: x, y

    !-------------------------------------------------------------------
    ! Start:

    if (.not.plotInitialized) then
!      call warning('pgplotNotInitialized',"You hadn't set up PGPLOT before &
!        &making a PGPLOT-dependent call. I've initialized the output device, &
!        &the viewport size, the labels and the colour index range to default &
!        &values. If that's indeed what you meant me to do, then no problem. &
!        &But if not, you might want to call plotInit before making the present &
!        &call.")
      call plotInit
!call error('pgplotNotInitialized',"")
    end if

    call pgqfs(oldFillStyle)
    call pgsfs(2) ! = unfilled.
    call pgqci(oldColourIndex)
!    call pgsci(fgCi)

    do i = 1, size(boxes)
      call pgsci(boxes(i)%colourIndex)
      call pgrect(boxes(i)%worldX%lo, boxes(i)%worldX%hi&
        ,         boxes(i)%worldY%lo, boxes(i)%worldY%hi)
    end do

    call pgsfs(oldFillStyle)
    call pgsci(oldColourIndex)

  end subroutine drawBoxes

  !---------------------------------------------------------------------
  subroutine drawButtons(buttons)
  !---------------------------------------------------------------------

    implicit none

    type(buttonType), intent(in) :: buttons(:)

    integer :: i, oldFillStyle, oldColourIndex
    real :: x, y

    !-------------------------------------------------------------------
    ! Start:

    if (.not.plotInitialized) then
!      call warning('pgplotNotInitialized',"You hadn't set up PGPLOT before &
!        &making a PGPLOT-dependent call. I've initialized the output device, &
!        &the viewport size, the labels and the colour index range to default &
!        &values. If that's indeed what you meant me to do, then no problem. &
!        &But if not, you might want to call plotInit before making the present &
!        &call.")
      call plotInit
!call error('pgplotNotInitialized',"")
    end if

    call pgqfs(oldFillStyle)
    call pgsfs(2) ! = unfilled.
    call pgqci(oldColourIndex)
!    call pgsci(fgCi)

    do i = 1, size(buttons)
      call pgsci(buttons(i)%boxField%colourIndex)
      call pgrect(buttons(i)%boxField%worldX%lo, buttons(i)%boxField%worldX%hi&
        ,         buttons(i)%boxField%worldY%lo, buttons(i)%boxField%worldY%hi)

      call pgsci(buttons(i)%textColourIndex)
      x = fracVal(buttons(i)%boxField%worldX%lo, buttons(i)%boxField%worldX%hi, 0.05)
      y = fracVal(buttons(i)%boxField%worldY%lo, buttons(i)%boxField%worldY%hi, 0.45)
      call pgtext(x, y, buttons(i)%text)
    end do

    call pgsfs(oldFillStyle)
    call pgsci(oldColourIndex)

  end subroutine drawButtons

  !---------------------------------------------------------------------
  subroutine divideUpRange(lo, hi, spaceFrac, ranges)
  !---------------------------------------------------------------------
  ! Useful for calculating Y coordinates for a column of buttons. A
  ! graphical explanation of the various quantities follows:
  !
  !	hi -     - hi
  !	    |   |
  !	    |   | ranges(3)
  !	    |   |
  !	    |    - lo
  !	    |   |
  !	    |    - hi
  !	    |   |
  !	    |   | ranges(2)
  !	    |   |
  !	    |    - lo
  !	    |   |
  !	    |    - hi
  !	    |   |
  !	    |   | ranges(1)
  !	    |   |
  !	lo -     - lo
  !
  ! Let a = ranges(i)%hi - ranges(i)%lo, b = the space between ranges and
  ! s = the spaceFrac, which is b/(a+b). NOTE that spaceFrac is NOT the
  ! ratio between the space between ranges and the size of each range, it
  ! is the ratio between the space between ranges and the TOTAL size of
  ! the range PLUS a space. Thus sensible values of spaceFrac are between
  ! 0 and 1.
  !
  ! Given only n=size(ranges), lo, hi and s, a little algebra shows that
  !
  !	             1 - s
  !	a = (hi-lo) ------- .
  !	             n - s
  !
  ! The ranges are then simply calculated using fracVal.

    implicit none

    real(single),         intent(in)  :: lo, hi, spaceFrac
    type(localRangeType), intent(out) :: ranges(:)

    integer :: numRanges, i
    real :: a

    !-------------------------------------------------------------------
    ! Start:

if (lo >= hi) call error('badRangeLimits',"")

    numRanges = size(ranges)
    a = (hi - lo) * (1.0 - spaceFrac) / (numRanges - spaceFrac)

    ranges%lo = fracVal(lo, hi-a, (/(i/real(numRanges-1), i=0,numRanges-1)/))
    ranges%hi = fracVal(lo+a, hi, (/(i/real(numRanges-1), i=0,numRanges-1)/))

  end subroutine divideUpRange

  !---------------------------------------------------------------------
  subroutine getIndexOfSelectedBox(boxes, cursorWorldX, cursorWorldY&
    , aBoxWasSelected, selectedBox)
  !---------------------------------------------------------------------

    implicit none

    type(boxFieldType), intent(in)  :: boxes(:)
    real,               intent(in)  :: cursorWorldX, cursorWorldY
    logical,            intent(out) :: aBoxWasSelected
    integer,            intent(out) :: selectedBox

    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    aBoxWasSelected = .false. ! default
    selectedBox = 0 ! impossible default.

    do i = 1, size(boxes)
      if (inBoxField(cursorWorldX, cursorWorldY, boxes(i))) then
        aBoxWasSelected = .true.
        selectedBox = i
    exit
      end if
    end do
        
!***** NOTE! The cursor position may lie within more than one box, if some of the boxes intersect. But here we will just choose the first matching box.

  end subroutine getIndexOfSelectedBox

  !---------------------------------------------------------------------
  subroutine drawArc(xCent, yCent, radius, startAngleDeg, stopAngleDeg)
  !---------------------------------------------------------------------
!*** calling this lots is going to be very slow. There are clever ways one could think of to speed it up: spline interpolation of a saved table of sines/cosines for example. But I'll leave this to future generations!

    implicit none

    real, intent(in) :: xCent, yCent, radius, startAngleDeg, stopAngleDeg

    real, parameter :: maxSegmentPixels=3.0 ! in device coords (pixels).
    real ::&
      startAngle, stopAngle,&
      xLoPixels, xHiPixels, yLoPixels, yHiPixels,&
      xLoWorld, xHiWorld, yLoWorld, yHiWorld,&
      xRadiusPixels, yRadiusPixels, minRadiusPixels, maxAngle
    integer :: numAngles, i
    real, allocatable :: angles(:), x(:), y(:)

    !-------------------------------------------------------------------
    ! Start

    startAngle = startAngleDeg * PiOn180
    stopAngle  = stopAngleDeg  * PiOn180

    call pgqvp(3, xLoPixels, xHiPixels, yLoPixels, yHiPixels)
    call pgqwin(xLoWorld, xHiWorld, yLoWorld, yHiWorld)
    xRadiusPixels = radius * (xHiPixels - xLoPixels) / (xHiWorld - xLoWorld)
    yRadiusPixels = radius * (yHiPixels - yLoPixels) / (yHiWorld - yLoWorld)
    minRadiusPixels = min(xRadiusPixels, yRadiusPixels)
    maxAngle = maxSegmentPixels / minRadiusPixels
    numAngles = ceiling(abs(stopAngle-startAngle)/maxAngle)
    allocate(angles(numAngles), x(numAngles), y(numAngles))
    angles = fracVal(startAngle, stopAngle&
      , (/(i/real(numAngles-1),i=0,numAngles-1)/))
    x = xCent + radius * cos(angles)
    y = yCent + radius * sin(angles)
    call pgline(numAngles, x, y)

  end subroutine drawArc


  !---------------------------------------------------------------------
  subroutine readColourMap_obsolete(inSetName, colourMap)
  !---------------------------------------------------------------------

    implicit none

    character(*),     intent(in) :: inSetName
    type(colourType), pointer    :: colourMap(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    integer :: numColours

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(inSetName, READ)
    tab = table(set, 'CLRMAP')
    numColours = numberOfRows(tab)
    allocate(colourMap(numColours))
    call readColDataToFixed(tab, 'RED_FRAC', colourMap%hue(red)) ! This sub is in
    call readColDataToFixed(tab, 'GRN_FRAC', colourMap%hue(grn)) ! ssclib/dal_aux.
    call readColDataToFixed(tab, 'BLU_FRAC', colourMap%hue(blu))
    call release(tab)
    call release(set)

  end subroutine readColourMap_obsolete

  !---------------------------------------------------------------------
  subroutine writeColourMap_obsolete(outSetName, colourMap)
  !---------------------------------------------------------------------

    implicit none

    character(*),     intent(in) :: outSetName
    type(colourType), intent(in) :: colourMap(:)

    type(DataSetT) :: set
    type(TableT) :: tab

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(outSetName, CREATE)
    tab = addTable(set, 'CLRMAP', size(colourMap))
    call writeColData(tab, 'RED_FRAC', colourMap%hue(red)) ! This sub is in
    call writeColData(tab, 'GRN_FRAC', colourMap%hue(grn)) ! ssclib/dal_aux.
    call writeColData(tab, 'BLU_FRAC', colourMap%hue(blu))
    call release(tab)
    call release(set)

  end subroutine writeColourMap_obsolete

  !---------------------------------------------------------------------
  subroutine readColourMap(inSetName, colourMap)
  !---------------------------------------------------------------------
  ! After calling the present routine you should set the colours by doing
  ! something like the following:
  !
  !    call pgqcol(minCi, maxCi)
  !    ciLo = minCi + numReservedClrs
  !    ciHi = maxCi
  !    call pgscir(ciLo, ciHi)
  !    call pgctab(colourMap%wedgeFrac, colourMap%brightFrac(red)&
  !      , colourMap%brightFrac(grn), colourMap%brightFrac(blu)&
  !      , size(colourMap), 1.0, 0.5)
  !---------------------------------------------------------------------

    implicit none

    character(*),        intent(in) :: inSetName
    type(colourMapType), pointer    :: colourMap(:)

    type(pinsType) :: hue(red:blu)

    !-------------------------------------------------------------------
    ! Start

    call loadColourPins(inSetName, hue)
    call convertPinsToColourMap(hue, colourMap)
    call deallocatePins(hue)

  end subroutine readColourMap

  !---------------------------------------------------------------------
  subroutine setColourMapFromMap(colourMap)
  !---------------------------------------------------------------------

    implicit none

    type(colourMapType), intent(in) :: colourMap(:)

    integer :: minCi, maxCi, ciLo, ciHi

    !-------------------------------------------------------------------
    ! Start

    if (.not.plotInitialized) call plotInit

    call pgqcol(minCi, maxCi)
    ciLo = minCi + numReservedClrs
    ciHi = maxCi
    call pgscir(ciLo, ciHi)
    call pgctab(colourMap%wedgeFrac, colourMap%brightFrac(red)&
      , colourMap%brightFrac(grn), colourMap%brightFrac(blu)&
      , size(colourMap), -1.0, 0.5)

     coloursHaveBeenSet=.true.

  end subroutine setColourMapFromMap

  !---------------------------------------------------------------------
  subroutine setColourMapFromName(inSetName)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: inSetName

    type(colourMapType), pointer :: colourMap(:)

    !-------------------------------------------------------------------
    ! Start

    call readColourMap(inSetName, colourMap)
    call setColourMapFromMap(colourMap)
    deallocate(colourMap)

  end subroutine setColourMapFromName

  !---------------------------------------------------------------------
  subroutine writeColourMap(outSetName, colourMap)
  !---------------------------------------------------------------------

    implicit none

    character(*),        intent(in) :: outSetName
    type(colourMapType), intent(in) :: colourMap(:)

    type(pinsType) :: hue(red:blu)

    !-------------------------------------------------------------------
    ! Start

    call convertColourMapToPins(colourMap, hue)
    call saveColourPins(outSetName, hue)
    call deallocatePins(hue)

  end subroutine writeColourMap


  !---------------------------------------------------------------------
  subroutine initializePinType(newPins)
  !---------------------------------------------------------------------

    implicit none

    type(pinType), intent(out) :: newPins(:)

    !-------------------------------------------------------------------
    ! Start

    newPins%wedgeFrac=0.0
    newPins%brightFrac=0.0
    newPins%worldX=0.0
    newPins%worldY=0.0

  end subroutine initializePinType

  !---------------------------------------------------------------------
  subroutine deallocatePins(hue)
  !---------------------------------------------------------------------

    implicit none

    type(pinsType), intent(inout) :: hue(red:blu)

    integer :: hueIndex

    !-------------------------------------------------------------------
    ! Start

    do hueIndex = red, blu
      deallocate(hue(hueIndex)%pins)
    end do

  end subroutine deallocatePins

  !---------------------------------------------------------------------
  subroutine loadColourPins(setName, hue)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in)  :: setName
    type(pinsType), intent(out) :: hue(red:blu)

    type(DataSetT) :: set
    integer :: hueIndex, numPins
    character(5) :: tabNames(red:blu) = (/'RED  ', 'GREEN', 'BLUE '/)
    type(TableT) :: tab
    type(ColumnT) :: brightCol, wedgeCol
    real(single), pointer :: bright(:), wedge(:)

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(setName, READ)
    do hueIndex = red, blu
      tab = table(set, tabNames(hueIndex))
      numPins = numberOfRows(tab)
      allocate(hue(hueIndex)%pins(numPins))
      brightCol = column(tab, 'BRIGHT_FRAC')
      wedgeCol  = column(tab, 'WEDGE_FRAC')
      bright => real32Data(brightCol)
      wedge  => real32Data(wedgeCol)
      hue(hueIndex)%pins%brightFrac = bright
      hue(hueIndex)%pins%wedgeFrac  = wedge
      call release(wedgeCol)
      call release(brightCol)
      call release(tab)
    end do
    call release(set)

  end subroutine loadColourPins

  !---------------------------------------------------------------------
  subroutine saveColourPins(setName, hue)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in) :: setName
    type(pinsType), intent(in) :: hue(red:blu)

    type(DataSetT) :: set
    integer :: hueIndex
    character(5) :: tabNames(red:blu) = (/'RED  ', 'GREEN', 'BLUE '/)
    type(TableT) :: tab
    type(ColumnT) :: brightCol, wedgeCol
    real(single), pointer :: bright(:), wedge(:)

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(setName, CREATE)
    do hueIndex = red, blu
      tab = addTable(set, trim(tabNames(hueIndex)), size(hue(hueIndex)%pins))
      brightCol = addColumn(tab, 'BRIGHT_FRAC', REAL32)
      wedgeCol  = addColumn(tab, 'WEDGE_FRAC',  REAL32&
        , comment='Fraction of available colour index range.')
      bright => real32Data(brightCol)
      wedge  => real32Data(wedgeCol)
      bright = hue(hueIndex)%pins%brightFrac
      wedge  = hue(hueIndex)%pins%wedgeFrac
      call release(wedgeCol)
      call release(brightCol)
      call release(tab)
    end do
    call release(set)

  end subroutine saveColourPins

  !---------------------------------------------------------------------
  subroutine calcPinsWorldCoords(hueFnBoxes, hue)
  !---------------------------------------------------------------------

    implicit none

    type(boxFieldType), intent(in)    :: hueFnBoxes(red:blu)
    type(pinsType),     intent(inout) :: hue(red:blu)

    integer :: hueIndex

    !-------------------------------------------------------------------
    ! Start

    do hueIndex = red, blu
      hue(hueIndex)%pins%worldX = fracVal(hueFnBoxes(hueIndex)%worldX%lo&
        , hueFnBoxes(hueIndex)%worldX%hi, hue(hueIndex)%pins%brightFrac)
      hue(hueIndex)%pins%worldY = fracVal(hueFnBoxes(hueIndex)%worldY%lo&
        , hueFnBoxes(hueIndex)%worldY%hi, hue(hueIndex)%pins%wedgeFrac)
      ! fracVal is in ssclib/ssc_misc.
    end do

  end subroutine calcPinsWorldCoords

  !---------------------------------------------------------------------
  subroutine convertPinsToColourMap(hue, colourMap)
  !---------------------------------------------------------------------

    implicit none

    type(pinsType),      intent(in) :: hue(red:blu)
    type(colourMapType), pointer    :: colourMap(:)

    integer ::&
      numPins(red:blu), numMapPoints, startI, finisI, hueIndex, hueIndex2, i
    type(colourMapType), allocatable :: tempClrMap(:)
    real :: oldWedgeFrac
    integer, allocatable :: indices(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if(testFlag) print *, '  aaa convertPinsToColourMap'
    numPins(red) = size(hue(red)%pins)
    numPins(grn) = size(hue(grn)%pins)
    numPins(blu) = size(hue(blu)%pins)

    numMapPoints = 2 + (numPins(red)-2 + numPins(grn)-2 + numPins(blu)-2)
    allocate(colourMap(numMapPoints))

    colourMap(           1)%wedgeFrac = 0.0 ! always.
    colourMap(           1)%brightFrac(red) = hue(red)%pins(1)%brightFrac
    colourMap(           1)%brightFrac(grn) = hue(grn)%pins(1)%brightFrac
    colourMap(           1)%brightFrac(blu) = hue(blu)%pins(1)%brightFrac
    colourMap(numMapPoints)%wedgeFrac = 1.0 ! always.
    colourMap(numMapPoints)%brightFrac(red) = hue(red)%pins(numPins(red))%brightFrac
    colourMap(numMapPoints)%brightFrac(grn) = hue(grn)%pins(numPins(grn))%brightFrac
    colourMap(numMapPoints)%brightFrac(blu) = hue(blu)%pins(numPins(blu))%brightFrac

    ! Each of the 3 hues of pins may have a sequence of wedgeFrac values which is unrelated to that of the others, but the single list of colourMap wedgeFrac values must be the union of the sequences for the three hues. The first thing is just to load the wedgeFrac sequences from all three pins hues into colourMap. The corresponding brightFrac values are obtained by interpolation. Finally the colourMap values are sorted in order of increasing wedgeFrac value.
    !
    finisI = 1
    do hueIndex = red, blu
      startI = finisI + 1
      finisI = startI + numPins(hueIndex) - 3
    if (finisI < startI) cycle ! because there are no 'internal' pins for this hue.
if(testFlag) print *, '    hueIndex, startI, finisI:', hueIndex, startI, finisI

      colourMap(startI:finisI)%wedgeFrac&
        = hue(hueIndex)%pins(2:numPins(hueIndex)-1)%wedgeFrac

      colourMap(startI:finisI)%brightFrac(hueIndex)&
        = hue(hueIndex)%pins(2:numPins(hueIndex)-1)%brightFrac

      do i = 1, 2
        hueIndex2 = red + modulo(hueIndex+i-red, 3)
if(testFlag) print *, '      i, hueIndex2:', i, hueIndex2
        call interpolate(hue(hueIndex2)%pins%brightFrac&
          , colourMap(startI:finisI)%brightFrac(hueIndex2)&
          , colourMap(startI:finisI)%wedgeFrac&
          , hue(hueIndex2)%pins%wedgeFrac, style=INTERP_LINEAR&
        ) ! in ssclib/interpolations.
      end do
    end do

    allocate(indices(numMapPoints), tempClrMap(numMapPoints))
    call quickSortStatic(colourMap%wedgeFrac, indices)
    tempClrMap = colourMap
    do i = 1, numMapPoints
      oldWedgeFrac = colourMap(i)%wedgeFrac
      colourMap(i) = tempClrMap(indices(i))
if(testFlag) print *, '    i, indices(i), old wedgeFrac, new wedgeFrac:'&
  , i, indices(i), oldWedgeFrac, colourMap(i)%wedgeFrac
    end do

if(testFlag) print *, '  zzz convertPinsToColourMap'

  end subroutine convertPinsToColourMap

  !---------------------------------------------------------------------
  subroutine convertColourMapToPins(colourMap, hue)
  !---------------------------------------------------------------------

    implicit none

    type(colourMapType), intent(in)  :: colourMap(:)
    type(pinsType),      intent(out) :: hue(red:blu)

    real, parameter :: minBendSize = 0.001

    integer :: numColourPoints,&
! numPins(red:blu),&
      hueIndex, i, j
    type(pinsType) :: tempHue(red:blu)
    real :: tempXs(3), tempYs(3)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if(testFlag) print *, '  aaa convertColourMapToPins'

    numColourPoints = size(colourMap)
    do hueIndex = red, blu
      allocate(tempHue(hueIndex)%pins(numColourPoints))
      call initializePinType(tempHue(hueIndex)%pins)

      i = 1
      j = 1
      tempHue(hueIndex)%pins(i)%wedgeFrac  = colourMap(j)%wedgeFrac
      tempHue(hueIndex)%pins(i)%brightFrac = colourMap(j)%brightFrac(hueIndex)

      do j = 2, numColourPoints-1
        tempXs = colourMap(j-1:j+1)%wedgeFrac
        tempYs = colourMap(j-1:j+1)%brightFrac(hueIndex)
!        if (secondDerivative(colourMap(j-1:j+1)%wedgeFrac&
!          , colourMap(j-1:j+1)%brightFrac(hueIndex)) > minBendSize) then
        if (secondDerivative(tempXs, tempYs) > minBendSize) then

          i = i + 1
          tempHue(hueIndex)%pins(i)%wedgeFrac  = colourMap(j)%wedgeFrac
          tempHue(hueIndex)%pins(i)%brightFrac = colourMap(j)%brightFrac(hueIndex)
        end if
      end do

      i = i + 1
      j = numColourPoints
      tempHue(hueIndex)%pins(i)%wedgeFrac  = colourMap(j)%wedgeFrac
      tempHue(hueIndex)%pins(i)%brightFrac = colourMap(j)%brightFrac(hueIndex)

!      numPins(hueIndex) = i

      allocate(hue(hueIndex)%pins(i))
      hue(hueIndex)%pins = tempHue(hueIndex)%pins(1:i)
    end do

if(testFlag) print *, '  zzz convertColourMapToPins'

  end subroutine convertColourMapToPins

  !---------------------------------------------------------------------
  real(single) function secondDerivative(xs, ys)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: xs(3), ys(3)

    !-------------------------------------------------------------------
    ! Start

    if (any(xs(1:2) >= xs(2:3)))&
call error('badOrderXs',"")

    secondDerivative = ((xs(3)-xs(2))*(ys(1)-ys(2))+(xs(2)-xs(1))*(ys(3)-ys(2)))&
      / (xs(2)-xs(1)) / (xs(3)-xs(2)) / (xs(3)-xs(1)) * 2.0

  end function secondDerivative

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module pgplot_aux
