!
! Module: regridders
!
! Authors:  Ian Stewart - SSC/LUX
!        :  Masaaki Sakano - SSC/LUX
!          
! Description: This module contains routines to regrid between arrays of different coordinate systems.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module regridders


!********* check out the stuff in non_sas/mycaltools/src/l_regridders some time - is this better?


  use types
  use errorHandling

  use array_utils	! for smallestTrueIndex, aryEdgesInfoT.
  use constants
  use geometric_types	! for Point2dT type
  use polygon		! for type defs, areaOfPoly, isConvex, pointIsInsidePoly
  use sort_mod
  use ssc_misc		! for isOdd, fracVal

  implicit none 

  private 

  public calcOldPixelCornersAffine
  public regridCartesian
  private calcIntersectPolyGrid
  private verticalSlice
  private horizontalSlice
  public regridCartesianToPolar
  public regridPolarToCartesian

  interface calcOldPixelCornersAffine
    module procedure calcOldPixelCornersAffineSingle, calcOldPixelCornersAffineDouble
  end interface

  interface regridCartesian
    module procedure regridScalar, regridScalarDouble, regridVector
  end interface

  interface loadSquare
    module procedure loadSquareSingle, loadSquarePos
  end interface

  integer(int8), parameter ::&
    INSIDE_WEDGE   = 0,&
    ON_RIGHT_RAY   = 1,&
    UPPER_RIGHT    = 2,&
    LOWER_RIGHT    = 3,&
    LOWER_LEFT     = 4,&
    UPPER_LEFT     = 5,&
    ON_LEFT_RAY    = 6,&
    INSIDE_CIRCLE  = 7,&
    ON_CIRCLE      = 8,&
    OUTSIDE_CIRCLE = 9,&
    INSIDE_INNER   = 10,&
    ON_INNER       = 11,&
    ANNULUS        = 12,&
    ON_OUTER       = 13,&
    OUTSIDE_OUTER  = 14,&
    ON_CENTRE      = 15,&
    OUTSIDE_WEDGE  = 16,&
    ON_LOK_RAY     = 17,&
    ON_HIK_RAY     = 18

contains 

  !---------------------------------------------------------------------
  subroutine calcOldPixelCornersAffineDouble(oldPixelCorners &
     , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)
  !---------------------------------------------------------------------
  !
  ! NOTE: Size of oldPixelCorners has to be larger by 1
  !  than those specified in oldAryEdgesInfo%arySize.
  !
  ! Descriptions: 
  !  This subroutine gives an array oldPixelCorners as an argument
  !  to pass to regridCartesian(), when an Affine transformation (A and B)
  !  as given below is the coordinate transformation used in regridding,
  !
  !   ( x_new )            ( x_old )   
  !   (       ) =    A  *  (       )  +  B
  !   ( y_new )            ( y_old )   
  !               ( a  b ) ( x_old )     ( e ) 
  !             = (      ) (       )  +  (   )
  !               ( c  d ) ( y_old )     ( f ) 
  !
  ! where the matrix A (=mtrxLinTrans) is a component
  ! for the linear transformation and B (=vecTranslate) is
  ! for the translation (a.k.a. parallel move).
  ! Note 
  !   ( a  b )   ( mtrxLinTrans(1,1)  mtrxLinTrans(1,2) )
  !   (      ) = (                                      )
  !   ( c  d )   ( mtrxLinTrans(2,1)  mtrxLinTrans(2,2) )
  !
  ! The returned oldPixelCorners can be directly passed to regridCartesian().
  ! 
  ! Note oldPixelCorners can contain the values which are out of
  ! the boundary of newAryEdgesInfo, such as zero or negative values.
  ! 
  ! Examples:
  !  (A) When oldAry(1:5,1:5), newAry(1:5,1:5), identical transformation.
  !    oldPixelCorners(1,1)%x,y == (0.5, 0.5) 
  !    oldPixelCorners(6,6)%x,y == (5.5, 5.5) 
  !  (B) When oldAry(1:5,1:5), newAry(-1:6,0:8), identical transformation.
  !    oldPixelCorners(1,1)%x,y == (2.5, 1.5) 
  !    oldPixelCorners(3,1)%x,y == (4.5, 1.5) 
  !    # because the respective indices of 3rd and 2nd for X and Y
  !    # in newAry correspond to (1,1) in oldAry.

    implicit none

    type(Point2dT), intent(out) :: oldPixelCorners(:,:)	! defined in geometric_types
    type(AryEdgesInfoT), intent(in) :: oldAryEdgesInfo, newAryEdgesInfo	! defined in array_utils
    real(double),   intent(in)  :: mtrxLinTrans(2,2), vecTranslate(2)

    real(double), dimension(2) :: indicesOld, indicesNew, indicesInSub, vecOld, vecNew
    type(AryEdgesInfoT) :: tmpAryEdgesInfo
    integer :: i, j, sizeOfOldPixCorn(2), tmpUbound(2)
    integer(int32) :: countOutOfBoundary
logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    sizeOfOldPixCorn = (/ size(oldPixelCorners,1), size(oldPixelCorners,2) /)
    if ((oldAryEdgesInfo%arySize(1) .ne. sizeOfOldPixCorn(1)-1) .or. &
        (oldAryEdgesInfo%arySize(2) .ne. sizeOfOldPixCorn(2)-1)) then
      call printAryEdgesInfo(oldAryEdgesInfo, 'DEBUG')
      print *,'size=',size(oldPixelCorners,1),size(oldPixelCorners,2)
      call error('inconsistentDimOldPixelCorners', 'Contact the code developer.')
    end if

    tmpAryEdgesInfo = newAryEdgesInfo
    tmpAryEdgesInfo%lboundIndex = 1
    tmpUbound(:) = getUbound(tmpAryEdgesInfo)	! Defined in array_utils.
    countOutOfBoundary = 0
if (testFlag) print *,'DEBUG:'
if (testFlag) call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
    do i=1, sizeOfOldPixCorn(1)
      do j=1, sizeOfOldPixCorn(2)
        indicesOld = (/ oldAryEdgesInfo%lboundIndex(1)+i-1.5d0 &
                      , oldAryEdgesInfo%lboundIndex(2)+j-1.5d0 /)
        vecOld = getAxesFromIndices(indicesOld, oldAryEdgesInfo)	! Get axes.  Defined in array_utils.
        vecNew = matmul(mtrxLinTrans, vecOld) + vecTranslate
if (.true.) then
        !newAryEdgesInfo%lboundIndex = 1	! against intent(in)
        indicesInSub = getIndicesFromAxes(vecNew, tmpAryEdgesInfo)	! Axes -> indices conversion on the New array.  Defined in array_utils.
else
        indicesNew = getIndicesFromAxes(vecNew, newAryEdgesInfo)	! Axes -> indices conversion on the New array.  Defined in array_utils.
        call calcIndicesFromIndices(indicesNew, indicesInSub, newAryEdgesInfo%lboundIndex)	! to get indicesInSub, namely the indices applied in the array variable in Subroutine (lbound==1).
end if
if (testFlag) print *,'DEBUG:vec=',vecNew,indicesNew, indicesInSub
        oldPixelCorners(i,j)%x = indicesInSub(1)
        oldPixelCorners(i,j)%y = indicesInSub(2)
        !oldPixelCorners(i,j)%x = indicesNew(1)
        !oldPixelCorners(i,j)%y = indicesNew(2)

        if ((oldPixelCorners(i,j)%x < tmpAryEdgesInfo%lboundIndex(1)-0.5) .or. &
            (oldPixelCorners(i,j)%x > tmpUbound(1)+0.5) .or. &
            (oldPixelCorners(i,j)%y < tmpAryEdgesInfo%lboundIndex(2)-0.5) .or. &
            (oldPixelCorners(i,j)%y > tmpUbound(2)+0.5) ) then
!print *,'oldPix=',i,j,oldPixelCorners(i,j)%x,oldPixelCorners(i,j)%y
          countOutOfBoundary = countOutOfBoundary + 1
        end if
      end do
    end do

    if (countOutOfBoundary > 0) then
       write(errStr, '(a,i8,a,es14.6,a)') "In ssclib/regridders: " &
          , countOutOfBoundary, " pixels in the old frame out of " &
          , real(i, double)*j, " fall out of the boundary in the new frame."
       call message(AppMsg, Verbose)
    end if

  end subroutine calcOldPixelCornersAffineDouble

  !---------------------------------------------------------------------
  subroutine calcOldPixelCornersAffineSingle(oldPixelCorners &
     , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)
  !---------------------------------------------------------------------
  ! See calcOldPixelCornersAffineDouble for detail.

    implicit none

    type(Point2dT), intent(out) :: oldPixelCorners(:,:)	! defined in geometric_types
    type(AryEdgesInfoT), intent(in) :: oldAryEdgesInfo, newAryEdgesInfo	! defined in array_utils
    real(single),   intent(in)  :: mtrxLinTrans(2,2), vecTranslate(2)

    !-------------------------------------------------------------------
    ! Start

    call calcOldPixelCornersAffineDouble(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo &
       , real(mtrxLinTrans, double), real(vecTranslate, double))

  end subroutine calcOldPixelCornersAffineSingle


  !---------------------------------------------------------------------
  subroutine regridScalar(oldPixelCorners, oldImage, newImage, status&
  , testFlagArg, averagingStyle)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in)            :: oldImage(:,:)
    type(Point2dT), intent(in)            :: oldPixelCorners(size(oldImage,1)+1&
                                                            ,size(oldImage,2)+1)
    real(single),   intent(out)           :: newImage(:,:)
    integer,        intent(out), optional :: status
    logical(bool),  intent(in),  optional :: testFlagArg
    character(*),   intent(in),  optional :: averagingStyle

    real(single), allocatable :: oldImages(:,:,:), newImages(:,:,:)
logical(bool) :: testFlag

    !-------------------------------------------------------------------
    ! Start

if (present(testFlagArg)) then
  testFlag = testFlagArg
else
  testFlag = .false.
end if

    allocate(oldImages(1, size(oldImage,1), size(oldImage,2)))
    allocate(newImages(1, size(newImage,1), size(newImage,2)))
    oldImages(1,:,:) = oldImage

    if (present(status)) then
      call regridVector(oldPixelCorners, oldImages, newImages, status&
        , testFlag, averagingStyle)
    else
      call regridVector(oldPixelCorners, oldImages, newImages&
        , testFlagArg=testFlag, averagingStyle=averagingStyle)
    end if

    newImage = newImages(1,:,:)

  end subroutine regridScalar

  !---------------------------------------------------------------------
  subroutine regridScalarDouble(oldPixelCorners, oldImage, newImage, status&
  , testFlagArg, averagingStyle)
  !---------------------------------------------------------------------

    implicit none

    real(single),       intent(in)            :: oldImage(:,:)
    type(Point2dDbleT), intent(in)            :: oldPixelCorners(&
                                                   size(oldImage,1)+1,&
                                                   size(oldImage,2)+1)
    real(single),       intent(out)           :: newImage(:,:)
    integer,            intent(out), optional :: status
    logical(bool),      intent(in),  optional :: testFlagArg
    character(*),       intent(in),  optional :: averagingStyle

    real(single), allocatable :: oldImages(:,:,:), newImages(:,:,:)
    type(Point2dT), allocatable :: oldPixelCornersSingle(:,:)
logical(bool) :: testFlag

    !-------------------------------------------------------------------
    ! Start

if (present(testFlagArg)) then
  testFlag = testFlagArg
else
  testFlag = .false.
end if

    allocate(oldImages(1, size(oldImage,1), size(oldImage,2)))
    allocate(newImages(1, size(newImage,1), size(newImage,2)))
    allocate(oldPixelCornersSingle(size(oldImage,1)+1,size(oldImage,2)+1))
    oldImages(1,:,:) = oldImage
    oldPixelCornersSingle%x = oldPixelCorners%x
    oldPixelCornersSingle%y = oldPixelCorners%y

    call regridVector(oldPixelCornersSingle, oldImages, newImages, status&
      , testFlag, averagingStyle)

    newImage = newImages(1,:,:)

  end subroutine regridScalarDouble

  !---------------------------------------------------------------------
  subroutine regridVector(oldPixelCorners, oldImages, newImages, status&
  , testFlagArg, averagingStyle)
  !---------------------------------------------------------------------
  ! This sub is intended to allow rebinning of an image from one cartesian
  ! coordinate system to another. Now, we define here an image as a two-
  ! dimensional array of uniform rectangular pixels. If we change the
  ! coordinate system, this image becomes distorted. It is nice to be able
  ! to convert it once again to an array of rectangular pixels, but this
  ! time in the new coordinate system. This involves taking each of the
  ! old, distorted pixels and dividing its contents up among the new
  ! pixels. It is assumed here that (i) the distorted pixel still has
  ! straight 'sides', ie it is a quadrilateral (a polygon for generality);
  ! (ii) that the value within the old pixel is evenly distributed through
  ! the pixel (this assumption MAY BE INVALID for extremely nonlinear
  ! distortions); (iii) that the distorted pixel is not folded over - ie
  ! that no two of its sides cross; (iv) that the old pixels are not
  ! greatly larger than the new - the present algorithm still works ok in
  ! such a regime, but the result will look 'steppy.' In this case an
  ! interpolation algorithm would yield smoother-looking results.
  !
  ! The relation between the old and new coordinate systems is here
  ! entirely contained within the input array oldPixelCorners. This gives
  ! the coordinates of each pixel in the array oldImage, expressed in the
  ! wcs 'pixel' system of newImage. What does 'wcs pixel system' mean? It
  ! means that the width and height of pixels in newImage are both equal
  ! to 1.0 and the centre of the pixel newImage(1,1) is at (1.0, 1.0).
  !
  ! NOTE: newImages is NOT defined if any error is detected before processing.
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in)            :: oldImages(:,:,:)
    type(Point2dT), intent(in)            :: oldPixelCorners(&
                                               size(oldImages,2)+1,&
                                               size(oldImages,3)+1&
                                             )
    real(single),   intent(out)           :: newImages(:,:,:)
    integer,        intent(out), optional :: status
    logical(bool),  intent(in),  optional :: testFlagArg
    character(*),   intent(in),  optional :: averagingStyle

    type(Point2dT) :: oldPoly(4)
    integer :: parity
    integer :: oldXi, oldXsize, oldYi, oldYsize,&
      newXsize, newYsize, foreignStatus, numImages
    type(Point2dT) :: minCornerNew, maxCornerNew
    real(single) :: areaOldPoly
logical(bool) :: testFlag
    character(20) :: localAveragingStyle

    integer, parameter ::&
      ZERO_IMG_BIT      = 0,&
      NO_OVERLAP_BIT    = 1,&
      ZERO_IMG_SIZE_BIT = 2,&
      WRONG_NUM_IMG_BIT = 3

    !-------------------------------------------------------------------
    ! Start

if (present(testFlagArg)) then
  testFlag = testFlagArg
else
  testFlag = .false.
end if

if (testFlag) print *, '      aaa regrid'

    status = 0

    if (present(averagingStyle)) then
      localAveragingStyle = averagingStyle
    else
      localAveragingStyle = 'counts'
    end if

    numImages = size(oldImages, 1)
    oldXsize  = size(oldImages, 2)
    oldYsize  = size(oldImages, 3)
    newXsize = size(newImages, 2)
    newYsize = size(newImages, 3)

    if (newXsize .le. 0 .or. newYsize .le. 0) then
      write(errStr, '("ssclib/regridders: ZERO_IMG_SIZE_BIT detected: XSize=",i6," YSize=",i6)')&
         newXsize, newYsize
      call message(AppMsg, Noisy)
      status = ibset(status, ZERO_IMG_SIZE_BIT)
  return
    end if

    if (size(newImages, 1) .ne. numImages) then
      call message(AppMsg, Noisy, 'ssclib/regridders: WRONG_NUM_IMG_BIT detected.')
      status = ibset(status, WRONG_NUM_IMG_BIT)
  return
    end if

    minCornerNew%x = 0.5
    minCornerNew%y = 0.5
    maxCornerNew%x = newXsize + 0.5
    maxCornerNew%y = newYsize + 0.5

    if (all(oldImages == 0.0)) then
      call message(AppMsg, Noisy, 'ssclib/regridders: ZERO_IMG_BIT detected.')
      status = ibset(status, ZERO_IMG_BIT)
  return
    end if

    if  (minval(oldPixelCorners%x) >= maxCornerNew%x&
    .or. maxval(oldPixelCorners%x) <= minCornerNew%x&
    .or. minval(oldPixelCorners%y) >= maxCornerNew%y&
    .or. maxval(oldPixelCorners%y) <= minCornerNew%y) then
      status = ibset(status, NO_OVERLAP_BIT)
      call message(AppMsg, Noisy, 'ssclib/regridders: NO_OVERLAP_BIT detected.')
  return
    end if

    newImages = 0.0

    ! Check to see whether poly vertices are in clockwise or
    ! anticlockwise order:
    oldPoly(1) = oldPixelCorners(1, 1)
    oldPoly(2) = oldPixelCorners(1, 2)
    oldPoly(3) = oldPixelCorners(2, 2)
    oldPoly(4) = oldPixelCorners(2, 1)
    areaOldPoly = areaOfPolygon(oldPoly) ! in polygon
    if (areaOldPoly > 0.0) then
      parity = 1
    else
      parity = -1
    end if

    ! Starting point: (i) a real array of size (oldXsize, oldYsize) that
    ! contains the old image values; (ii) an array of type Point2dT and
    ! size (oldXsize+1, oldYsize+1) which contains the xy values of each
    ! pixel corner of the old image in the new coord system; (iii) an empty
    ! real array of size (newXsize, newYsize) which will contain the
    ! regridded image, with associated minXY and maxXY values.

    if (parity > 0.0) then
      ! Process each old image pixel:
      do oldXi = 1, oldXsize
        do oldYi = 1, oldYsize
        if (all(oldImages(:, oldXi, oldYi) == 0.0)) cycle

          ! Load the coordinates of the pixel corners into a polygon:
          oldPoly(1) = oldPixelCorners(oldXi,     oldYi)
          oldPoly(2) = oldPixelCorners(oldXi,     oldYi + 1)
          oldPoly(3) = oldPixelCorners(oldXi + 1, oldYi + 1)
          oldPoly(4) = oldPixelCorners(oldXi + 1, oldYi)

          if  (minval(oldPoly%x) >= maxCornerNew%x&
          .or. maxval(oldPoly%x) <= minCornerNew%x&
          .or. minval(oldPoly%y) >= maxCornerNew%y&
          .or. maxval(oldPoly%y) <= minCornerNew%y)&
        cycle

!*** check for no lines crossing??

          call calcIntersectPolyGrid(oldPoly, oldImages(:, oldXi, oldYi)&
            , minCornerNew, maxCornerNew, localAveragingStyle, newImages&
            , foreignStatus, testFlag)

!*** absorb foreignStatus into status
          if (foreignStatus < 0) then
print *, 'calcIntersectPolyGrid err code =', foreignStatus
print *, 'parity > 0.0'
print *, 'oldXi, oldYi:', oldXi, oldYi
            status = -2
  return
          else if (foreignStatus > 0) then
            status = ior(status, ishft(foreignStatus, 3))
!***** or -3???
          end if
        end do
      end do

    else ! parity < 0
      ! Process each old image pixel:
      do oldXi = 1, oldXsize
        do oldYi = 1, oldYsize
        if (all(oldImages(:, oldXi, oldYi) == 0.0)) cycle

          ! Load the coordinates of the pixel corners into a polygon:
          oldPoly(4) = oldPixelCorners(oldXi,     oldYi)
          oldPoly(3) = oldPixelCorners(oldXi,     oldYi + 1)
          oldPoly(2) = oldPixelCorners(oldXi + 1, oldYi + 1)
          oldPoly(1) = oldPixelCorners(oldXi + 1, oldYi)

          if  (minval(oldPoly%x) >= maxCornerNew%x&
          .or. maxval(oldPoly%x) <= minCornerNew%x&
          .or. minval(oldPoly%y) >= maxCornerNew%y&
          .or. maxval(oldPoly%y) <= minCornerNew%y)&
        cycle

!*** check for no lines crossing??

          call calcIntersectPolyGrid(oldPoly, oldImages(:, oldXi, oldYi)&
            , minCornerNew, maxCornerNew, localAveragingStyle, newImages&
            , foreignStatus, testFlag)

!*** check foreignStatus into status
          if (foreignStatus < 0) then
print *, 'calcIntersectPolyGrid err code =', foreignStatus
print *, 'parity <= 0.0'
print *, 'oldXi, oldYi:', oldXi, oldYi
            status = -2
  return
          else if (foreignStatus > 0) then
            status = ior(status, ishft(foreignStatus, 3))
!***** or -3???
          end if

        end do
      end do
    end if
if (testFlag) print *, '      zzz regrid'

  end subroutine regridVector

  !---------------------------------------------------------------------
  subroutine calcIntersectPolyGrid(oldPoly, values, minCornerNew&
  , maxCornerNew, averagingStyle, newImages, status, testFlag)
  !---------------------------------------------------------------------
  ! This sub takes a polygonal bin (perhaps, but not necessarily, a
  ! distorted image pixel) which contains a real value and divides the
  ! value up (on a per area basis) between the pixels of the array
  ! newImage(). Although the indices of newImage() extend between 1
  ! and size(newImage, 1) or size(newImage, 2), only pixels with x index
  ! between lowNewXi and hiNewXi and y index between lowNewYi and hiNewYi
  ! are added to. The left and right world-coordinate edges of the
  ! newImage() array are stored in minNew and maxNew.

    implicit none

    type(Point2dT), intent(in)    :: oldPoly(:),&
                                     minCornerNew,&
                                     maxCornerNew
    real(single),   intent(in)    :: values(:)
    character(*),   intent(in)    :: averagingStyle	! No correction if 'rate', otherwise (such as, 'count') the original total counts are preserved.
    real(single),   intent(inout) :: newImages(:,:,:)
    integer,        intent(out)   :: status
logical(bool), intent(in) :: testFlag

    type(Point2dT), allocatable :: slicePolys(:,:), chunkPolys(:,:)
    ! The dimensions given are the maximum possible limits. The first index
    ! is the number of the sub-poly resulting from the intersection of poly
    ! with the two verticals, the second index is the number of the point
    ! within that sub-poly.

    integer, allocatable :: numSlicePolyPoints(:), numChunkPolyPoints(:)
    integer ::&
      xi, yi, numSlicePolys, numPts,&
      numChunkPolys, sliceStatus, numSlicePoly, numChunkPoly,&
      loNewXi, hiNewXi, loNewYi, hiNewYi, newXsize, newYsize, numImages
    real(single) ::&
      areaOldPoly, normArea, overlapArea, realNumCols, realNumRows,&
      xLo, xHi, yLo, yHi
logical :: localTestFlag

    !-------------------------------------------------------------------
    ! Start

!**** do something with averagingStyle! Default action of the routine at the moment seems to be 'rate'.

    status = 0
localTestFlag = testFlag

    numImages = size(newImages, 1)
    newXsize  = size(newImages, 2)
    newYsize  = size(newImages, 3)
    realNumCols = real(newXsize)
    realNumRows = real(newYsize)
    numPts = size(oldPoly)

    allocate(numSlicePolyPoints(numPts), numChunkPolyPoints(numPts))

    if (size(values) .ne. numImages) then
      status = -4
  return
    end if

    ! Calculate minigrid limits:
    loNewXi = max(nint(minval(oldPoly%x)), 1)
    hiNewXi = min(nint(maxval(oldPoly%x)), newXsize)
    loNewYi = max(nint(minval(oldPoly%y)), 1)
    hiNewYi = min(nint(maxval(oldPoly%y)), newYsize)

    ! Make sure poly vertices are in clockwise order:
    areaOldPoly = areaOfPolygon(oldPoly) ! in polygon.
    if (areaOldPoly < 0.0) then
      status = -1
  return
    end if

    select case(trim(averagingStyle))
    case('rate')
      normArea = 1.0 ! ie, area of new poly
    case default
      normArea = areaOldPoly
    end select
    ! if (trim(averagingStyle) == 'rate') then
    !   normArea = 1.0 ! ie, area of new poly
    ! else
    !   normArea = areaOldPoly
    ! end if

    allocate(slicePolys(numPts, 2*numPts))
    allocate(chunkPolys(numPts, 2*numPts))

    ! Slice up the polygon vertically:
    do xi = loNewXi, hiNewXi
      xLo = fracVal(minCornerNew%x, maxCornerNew%x, (xi - 1) / realNumCols)
      xHi = fracVal(minCornerNew%x, maxCornerNew%x,  xi      / realNumCols)

      call verticalSlice(&
        oldPoly,&
        xLo, xHi, slicePolys, &
        numSlicePolys, numSlicePolyPoints, sliceStatus&
      )

      if (sliceStatus .lt. 0) then
        status = -2
  return
      end if

      if (numSlicePolys .gt. 1) status = ibset(status, 0)

    if (numSlicePolys .le. 0) cycle

      ! Slice up the polygon horizontally:
      do yi = loNewYi, hiNewYi
        yLo = fracVal(minCornerNew%y, maxCornerNew%y, (yi - 1) / realNumRows)
        yHi = fracVal(minCornerNew%y, maxCornerNew%y,  yi      / realNumRows)
        do numSlicePoly = 1, numSlicePolys
          call horizontalSlice(&
            slicePolys(numSlicePoly, 1:numSlicePolyPoints(numSlicePoly)), &
            yLo, yHi, chunkPolys, numChunkPolys, numChunkPolyPoints, &
            sliceStatus, localTestFlag&
          )

          if (sliceStatus < 0) then
print *, 'xi, yi, numSlicePoly:', xi, yi, numSlicePoly
            status = sliceStatus - 3
  return
          else if (sliceStatus > 0) then
            status = ior(status, ishft(sliceStatus, 2))
          end if

          if (numChunkPolys > 1) status = ibset(status, 1)

        if (numChunkPolys <= 0) cycle

          do numChunkPoly = 1, numChunkPolys
            overlapArea = areaOfPolygon(&
              chunkPolys(numChunkPoly, 1:numChunkPolyPoints(numChunkPoly))&
            ) ! in polygon.

            newImages(:, xi, yi) = newImages(:, xi, yi)&
              + (overlapArea / normArea) * values

          end do ! loop over sub-polys in chunk
        end do ! loop over sub-polys in slice
      end do ! loop over horizontal bands
    end do ! loop over vertical bands

  end subroutine calcIntersectPolyGrid

  !---------------------------------------------------------------------
  subroutine verticalSlice(poly, xLo, xHi, subPolys&
  , nSubs, nPoints, status)
  !---------------------------------------------------------------------
  ! NOTE that it is ABSOLUTELY REQUIRED in this subroutine that the points
  ! in the poly are numbered in clockwise order and that none of the sides
  ! cross each other.
  !
  ! The sub has been copied to horizontalSlice with minimal alterations.
  ! Therefore if bugs are found in one, the other should also be altered.

    implicit none

    type(Point2dT), intent(in)  :: poly(:)
    real(single),   intent(in)  :: xLo, xHi
    type(Point2dT), intent(out) :: subPolys(size(poly),2*size(poly))
    ! The dimensions given are the maximum possible limits. The first index
    ! is the number of the sub-poly resulting from the intersection of poly
    ! with the two verticals, the second index is the number of the point
    ! within that sub-poly.
    integer,             intent(out) :: nPoints(size(poly)), nSubs, status

    type :: intersectionType
      real :: y
      logical :: exists, isAvailable
    end type intersectionType

    type(intersectionType), allocatable :: interLo(:), interHi(:)
    integer, allocatable :: zone(:)
    integer :: i, j, k, nVertices, precedingVertex,&
      state, indexVec(1)
    integer, parameter :: LOW = 1, HIGH = 2, MIDDLE = 3
    real(single) :: xFrac
    logical, allocatable :: vertexIsAvailable(:), mask(:)

    !-------------------------------------------------------------------
    ! Start

    status = 0
    nPoints = 0 ! vector default
    nSubs = 0 ! default

  if (minval(poly%x) .ge. xHi .or. maxval(poly%x) .le. xLo) return

    nVertices = size(poly)
    allocate(interLo(nVertices), interHi(nVertices))
    allocate(zone(nVertices))
    allocate(vertexIsAvailable(nVertices), mask(nVertices))

! Find out into which zone each poly vertex falls:

    do i = 1, nVertices
      if (poly(i)%x .le. xLo) then
        zone(i) = LOW
      else if (poly(i)%x .ge. xHi) then
        zone(i) = HIGH
      else
        zone(i) = MIDDLE
      end if
    end do

! Find which sides of the polygon intersect the left and right verticals:

    interLo%isAvailable = .false.
    interHi%isAvailable = .false.
    interLo%exists = .false.
    interHi%exists = .false.
    do i = 1, nVertices
      j = 1 + modulo(i, nVertices) ! ie j=1 for i=nVertices, j=i+1 else.

      if ( (poly(i)%x .le. xLo .and. poly(j)%x .gt. xLo)&
      .or. (poly(i)%x .gt. xLo .and. poly(j)%x .le. xLo)) then
        xFrac = (xLo - poly(i)%x) / (poly(j)%x - poly(i)%x)
        interLo(i)%y = fracVal(poly(i)%y, poly(j)%y, xFrac)
        interLo(i)%exists      = .true.
        interLo(i)%isAvailable = .true.
      end if
      if ( (poly(i)%x .ge. xHi .and. poly(j)%x .lt. xHi)&
      .or. (poly(i)%x .lt. xHi .and. poly(j)%x .ge. xHi)) then
        xFrac = (xHi - poly(i)%x) / (poly(j)%x - poly(i)%x)
        interHi(i)%y = fracVal(poly(i)%y, poly(j)%y, xFrac)
        interHi(i)%exists      = .true.
        interHi(i)%isAvailable = .true.
      end if
    end do

    ! And, just for neatness really, change a glancing touch onto one of
    ! the verticals from a non-MIDDLE point to a MIDDLE point. If this is
    ! not done, the intersection polygon will have 2 points at that place.

    k = 1 + modulo(1, nVertices)
    do i = 1, nVertices
      j = k				! ie j = i + 1, modulo nVertices
      k = 1 + modulo(i + 1, nVertices)	! ie k = i + 2, modulo nVertices
      if (zone(i) .eq. MIDDLE .and. zone(k) .eq. MIDDLE) then
        if (poly(j)%x .eq. xLo) then
          if (isConvex(poly(i), poly(j), poly(k))) then
            zone(j) = MIDDLE
            interLo(i)%exists = .false.
            interLo(j)%exists = .false.
            interLo(i)%isAvailable = .false.
            interLo(j)%isAvailable = .false.
          end if
        else if (poly(j)%x .eq. xHi) then
          if (isConvex(poly(i), poly(j), poly(k))) then
            zone(j) = MIDDLE
            interHi(i)%exists = .false.
            interHi(j)%exists = .false.
            interHi(i)%isAvailable = .false.
            interHi(j)%isAvailable = .false.
          end if
        end if
      end if
    end do

if (any(interLo%exists .and. interHi%exists .and. (zone .eq. MIDDLE))) then
!***** BUG!!
status = -1
return
end if

    vertexIsAvailable = .false. ! vector default
    where(zone .eq. MIDDLE) vertexIsAvailable = .true.

    ! Finally, because it is possible for poly()pos%x to be just on the
    ! other side of a vertical to its neighbouring vertices, thereby
    !generating two neighbouring intercepts with that vertical, but these
    !intercepts can nevertheless due to numerical rounding errors be at
    !the same y value, the following is done to erase any such paired,
    !same-y intercepts:

    do i = 1, nVertices
      j = 1 + modulo(i, nVertices)	! ie j = i + 1, modulo nVertices

      if (interLo(i)%exists .and. interLo(j)%exists&
      .and. interLo(i)%y .eq. interLo(j)%y) then
        vertexIsAvailable(j) = .not. vertexIsAvailable(j)
        interLo(i)%exists = .false.
        interLo(j)%exists = .false.
        interLo(i)%isAvailable = .false.
        interLo(j)%isAvailable = .false.
      end if

      if (interHi(i)%exists .and. interHi(j)%exists&
      .and. interHi(i)%y .eq. interHi(j)%y) then
        vertexIsAvailable(j) = .not. vertexIsAvailable(j)
        interHi(i)%exists = .false.
        interHi(j)%exists = .false.
        interHi(i)%isAvailable = .false.
        interHi(j)%isAvailable = .false.
      end if
    end do

! Now construct the intersection of the verticals and the poly (could be a number of subpolys):

    do while(any(interLo%isAvailable) .or. any(interHi%isAvailable)&
    .or. any(vertexIsAvailable))

      ! Start a new subpoly:

      nSubs = nSubs + 1
      nPoints(nSubs) = 1

      if (any(interLo%isAvailable)) then ! find it:
        do i = 1, nVertices
          if (interLo(i)%isAvailable) then
            precedingVertex = i
        exit
          end if
        end do
        interLo(precedingVertex)%isAvailable = .false.
        subPolys(nSubs, nPoints(nSubs))%x = xLo
        subPolys(nSubs, nPoints(nSubs))%y = interLo(precedingVertex)%y

        if (zone(precedingVertex) .eq. LOW) then
          state = MIDDLE
        else
          state = LOW
        end if

      else if (any(interHi%isAvailable)) then
        do i = 1, nVertices
          if (interHi(i)%isAvailable) then
            precedingVertex = i
        exit
          end if
        end do
        interHi(precedingVertex)%isAvailable = .false.
        subPolys(nSubs, nPoints(nSubs))%x = xHi
        subPolys(nSubs, nPoints(nSubs))%y = interHi(precedingVertex)%y

        if (zone(precedingVertex) .eq. HIGH) then
          state = MIDDLE
        else
          state = HIGH
        end if

      else ! there must be an available vertex.
        do i = 1, nVertices
          if (vertexIsAvailable(i)) then
            precedingVertex = i
        exit
          end if
        end do
        vertexIsAvailable(precedingVertex) = .false.
        subPolys(nSubs, nPoints(nSubs)) = poly(precedingVertex)

        state = MIDDLE
      end if

      ! Now follow it around until the end:
      do
        select case(state)
        case(LOW)
        ! Move up until next higher intercept, after which state ought to
        ! be MIDDLE.
          mask = (interLo%exists&
          .and. interLo%y .gt. subPolys(nSubs, nPoints(nSubs))%y)
if (.not. any(mask)) then
!***** BUG!!
status = -2
return
end if
          indexVec = minloc(interLo%y, mask)
          precedingVertex = indexVec(1)
      if (.not. interLo(precedingVertex)%isAvailable) exit ! because this
      ! subpoly is now finished.

if (zone(precedingVertex) .ne. LOW) then
!***** BUG!!
status = -3
return
end if

          nPoints(nSubs) = nPoints(nSubs) + 1
          interLo(precedingVertex)%isAvailable = .false.
          subPolys(nSubs, nPoints(nSubs))%x = xLo
          subPolys(nSubs, nPoints(nSubs))%y = interLo(precedingVertex)%y

          state = MIDDLE

        case(HIGH)
        ! Move down until next lower intercept, after which state ought to
        ! be MIDDLE.
          mask = (interHi%exists&
          .and. interHi%y .lt. subPolys(nSubs, nPoints(nSubs))%y)
if (.not. any(mask)) then
!***** BUG!!
status = -4
return
end if
          indexVec = maxloc(interHi%y, mask)
          precedingVertex = indexVec(1)
      if (.not. interHi(precedingVertex)%isAvailable) exit ! because this
      ! subpoly is now finished.

if (zone(precedingVertex) .ne. HIGH) then
!***** BUG!!
status = -5
return
end if

          nPoints(nSubs) = nPoints(nSubs) + 1
          interHi(precedingVertex)%isAvailable = .false.
          subPolys(nSubs, nPoints(nSubs))%x = xHi
          subPolys(nSubs, nPoints(nSubs))%y = interHi(precedingVertex)%y

          state = MIDDLE

        case(MIDDLE)
          if (interLo(precedingVertex)%isAvailable) then
            nPoints(nSubs) = nPoints(nSubs) + 1
            interLo(precedingVertex)%isAvailable = .false.
            subPolys(nSubs, nPoints(nSubs))%x = xLo
            subPolys(nSubs, nPoints(nSubs))%y = interLo(precedingVertex)%y

            state = LOW

          else&
          if (interHi(precedingVertex)%isAvailable) then
            nPoints(nSubs) = nPoints(nSubs) + 1
            interHi(precedingVertex)%isAvailable = .false.
            subPolys(nSubs, nPoints(nSubs))%x = xHi
            subPolys(nSubs, nPoints(nSubs))%y = interHi(precedingVertex)%y

            state = HIGH

          else
            precedingVertex = 1 + modulo(precedingVertex, nVertices)
      if (.not. vertexIsAvailable(precedingVertex)) exit ! because this
      ! subpoly is now finished.

            nPoints(nSubs) = nPoints(nSubs) + 1
            vertexIsAvailable(precedingVertex) = .false.
            subPolys(nSubs, nPoints(nSubs)) = poly(precedingVertex)
          end if

        case default
!***** BUG!!
status = -6
return
        end select
      end do ! complete sub-poly
    end do ! no more available vertices or intersections.

  end subroutine verticalSlice

  !---------------------------------------------------------------------
  subroutine horizontalSlice(poly, yLo, yHi, subPolys&
!!!  , nSubs, nPoints, status)
, nSubs, nPoints, status, testFlag)
  !---------------------------------------------------------------------
  ! NOTE that it is ABSOLUTELY REQUIRED in this subroutine that the points
  ! in the poly are numbered in clockwise order and that none of the sides
  ! cross each other.
  !
  ! The sub has been copied from verticalSlice with minimal alterations.
  ! Therefore if bugs are found in one, the other should also be altered.

    implicit none

    type(Point2dT), intent(in)  :: poly(:)
    real(single),   intent(in)  :: yLo, yHi
    type(Point2dT), intent(out) :: subPolys(:,:)
    ! The dimensions given are the maximum possible limits. The first index
    ! is the number of the sub-poly resulting from the intersection of poly
    ! with the two verticals, the second index is the number of the point
    ! within that sub-poly.
    integer,             intent(out) :: nPoints(:), nSubs, status
logical, intent(in) :: testFlag

    type :: intersectionType
      real :: x
      logical :: exists, isAvailable
    end type intersectionType

    type(intersectionType), allocatable :: interLo(:), interHi(:)
    integer, allocatable :: zone(:)
    integer :: &
      i, j, k, nVertices, precedingVertex, state, indexVec(1)
    integer, parameter :: LOW = 1, HIGH = 2, MIDDLE = 3
    real(single) :: yFrac
    logical, allocatable :: vertexIsAvailable(:), mask(:)

    !-------------------------------------------------------------------
    ! Start

    status = 0
    nPoints = 0 ! vector default
    nSubs = 0 ! default

  if (minval(poly%y) .ge. yHi .or. maxval(poly%y) .le. yLo) return

    nVertices = size(poly)
    allocate(interLo(nVertices), interHi(nVertices))
    allocate(zone(nVertices))
    allocate(vertexIsAvailable(nVertices), mask(nVertices))

if (testFlag) then
  do i = 1, nVertices
    print *, '  poly(',i,')%x, poly(',i,')%y:', poly(i)%x, poly(i)%y
  end do
  print *, 'yLo, yHi:', yLo, yHi
  print *, ''
end if

! Find out into which zone each poly vertex falls:

    do i = 1, nVertices
      if (poly(i)%y .le. yLo) then
        zone(i) = LOW
      else if (poly(i)%y .ge. yHi) then
        zone(i) = HIGH
      else
        zone(i) = MIDDLE
      end if
    end do

if (testFlag) then
  print *, 'Find out into which zone each poly vertex falls:'
  do i = 1, nVertices
    print *, '  i, zone(i):', i, zone(i)
  end do
  print *, ''
end if

! Find which sides of the polygon intersect the lower and upper horizontals:

    interLo%isAvailable = .false.
    interHi%isAvailable = .false.
    interLo%exists = .false.
    interHi%exists = .false.
    do i = 1, nVertices
      j = 1 + modulo(i, nVertices) ! ie j=1 for i=nVertices, j=i+1 else.

      if ( (poly(i)%y .le. yLo .and. poly(j)%y .gt. yLo)&
      .or. (poly(i)%y .gt. yLo .and. poly(j)%y .le. yLo)) then
        yFrac = (yLo - poly(i)%y) / (poly(j)%y - poly(i)%y)
if (testFlag) print *, 'lo - i, j, yFrac:', i, j, yFrac
        interLo(i)%x = fracVal(poly(i)%x, poly(j)%x, yFrac)
        interLo(i)%exists      = .true.
        interLo(i)%isAvailable = .true.
      end if
      if ( (poly(i)%y .ge. yHi .and. poly(j)%y .lt. yHi)&
      .or. (poly(i)%y .lt. yHi .and. poly(j)%y .ge. yHi)) then
        yFrac = (yHi - poly(i)%y) / (poly(j)%y - poly(i)%y)
if (testFlag) print *, 'hi - i, j, yFrac:', i, j, yFrac
if (testFlag) print *, 'yFrac, 1-(1-yFrac)), yFrac+(1-yFrac), (1-(1-yFrac))+(1-yFrac):'&
, yFrac, 1.-(1.-yFrac), yFrac+(1.-yFrac), (1.-(1.-yFrac))+(1.-yFrac)
        interHi(i)%x = fracVal(poly(i)%x, poly(j)%x, yFrac)
        interHi(i)%exists      = .true.
        interHi(i)%isAvailable = .true.
      end if
    end do

if (testFlag) then
  print *, 'Find which sides of the polygon intersect the left and right verticals:'
  do i = 1, nVertices
    print *, '  i, interLo(i)%isAvailable, interLo(i)%exists, interLo(i)%x:'&
      , i, interLo(i)%isAvailable, interLo(i)%exists, interLo(i)%x
    print *, '  i, interHi(i)%isAvailable, interHi(i)%exists, interHi(i)%x:'&
      , i, interHi(i)%isAvailable, interHi(i)%exists, interHi(i)%x
  end do
  print *, ''
end if

    ! And, just for neatness really, change a glancing touch onto one of
    ! the horizontals from a non-MIDDLE point to a MIDDLE point. If this
    ! is not done, the intersection polygon will have 2 points at that place.

    k = 1 + modulo(1, nVertices)
    do i = 1, nVertices
      j = k				! ie j = i + 1, modulo nVertices
      k = 1 + modulo(i + 1, nVertices)	! ie k = i + 2, modulo nVertices
      if (zone(i) .eq. MIDDLE .and. zone(k) .eq. MIDDLE) then
        if (poly(j)%y .eq. yLo) then
          if (isConvex(poly(i), poly(j), poly(k))) then
            zone(j) = MIDDLE
            interLo(i)%exists = .false.
            interLo(j)%exists = .false.
            interLo(i)%isAvailable = .false.
            interLo(j)%isAvailable = .false.
          end if
        else if (poly(j)%y .eq. yHi) then
          if (isConvex(poly(i), poly(j), poly(k))) then
            zone(j) = MIDDLE
            interHi(i)%exists = .false.
            interHi(j)%exists = .false.
            interHi(i)%isAvailable = .false.
            interHi(j)%isAvailable = .false.
          end if
        end if
      end if
    end do

if (testFlag) then
  print *, 'And, just for neatness really, change a glancing touch onto &
    &one of the horizontals from a non-MIDDLE point to a MIDDLE point.'
  do i = 1, nVertices
    print *, '  i, zone(i):', i, zone(i)
    print *, '  i, interLo(i)%isAvailable, interLo(i)%exists, interLo(i)%x:'&
      , i, interLo(i)%isAvailable, interLo(i)%exists, interLo(i)%x
    print *, '  i, interHi(i)%isAvailable, interHi(i)%exists, interHi(i)%x:'&
      , i, interHi(i)%isAvailable, interHi(i)%exists, interHi(i)%x
  end do
  print *, ''
end if

if (any(interLo%exists .and. interHi%exists .and. (zone .eq. MIDDLE))) then
!***** BUG!!
status = -1
return
end if

    vertexIsAvailable = .false. ! vector default
    where(zone .eq. MIDDLE) vertexIsAvailable = .true.

    ! Finally, because it is possible for poly()pos%y to be just on the
    ! other side of a horizontal to its neighbouring vertices, thereby
    ! generating two neighbouring intercepts with that horizontal, but
    ! these intercepts can nevertheless due to numerical rounding errors
    ! be at the same x value, the following is done to erase any such
    ! paired, same-x intercepts:

    do i = 1, nVertices
      j = 1 + modulo(i, nVertices)	! ie j = i + 1, modulo nVertices

      if (interLo(i)%exists .and. interLo(j)%exists&
      .and. interLo(i)%x .eq. interLo(j)%x) then
        vertexIsAvailable(j) = .not. vertexIsAvailable(j)
        interLo(i)%exists = .false.
        interLo(j)%exists = .false.
        interLo(i)%isAvailable = .false.
        interLo(j)%isAvailable = .false.
      end if

      if (interHi(i)%exists .and. interHi(j)%exists&
      .and. interHi(i)%x .eq. interHi(j)%x) then
        vertexIsAvailable(j) = .not. vertexIsAvailable(j)
        interHi(i)%exists = .false.
        interHi(j)%exists = .false.
        interHi(i)%isAvailable = .false.
        interHi(j)%isAvailable = .false.
      end if
    end do

if (testFlag) then
  print *, 'Erase any paired, same-x intercepts:'
  do i = 1, nVertices
    print *, '  i, zone(i):', i, zone(i)
    print *, '  i, interLo(i)%isAvailable, interLo(i)%exists, interLo(i)%x:'&
      , i, interLo(i)%isAvailable, interLo(i)%exists, interLo(i)%x
    print *, '  i, interHi(i)%isAvailable, interHi(i)%exists, interHi(i)%x:'&
      , i, interHi(i)%isAvailable, interHi(i)%exists, interHi(i)%x
  end do
  print *, ''
end if

! Now construct the intersection of the horizontals and the poly (could result in a number of subpolys):

    do while(any(interLo%isAvailable) .or. any(interHi%isAvailable)&
    .or. any(vertexIsAvailable))

      ! Start a new sub-poly:

      nSubs = nSubs + 1
      nPoints(nSubs) = 1
if (testFlag) then
  print *, 'nSubs, nPoints(nSubs):', nSubs, nPoints(nSubs)
end if

      if (any(interLo%isAvailable)) then ! find it:
        do i = 1, nVertices
          if (interLo(i)%isAvailable) then
            precedingVertex = i
        exit
          end if
        end do
        interLo(precedingVertex)%isAvailable = .false.
        subPolys(nSubs, nPoints(nSubs))%y = yLo
        subPolys(nSubs, nPoints(nSubs))%x = interLo(precedingVertex)%x

        if (zone(precedingVertex) .eq. LOW) then
          state = MIDDLE
        else
          state = LOW
        end if

if (testFlag) print *, 'pvtx, zone(pvtx), state:', precedingVertex, zone(precedingVertex), state
      else if (any(interHi%isAvailable)) then
        do i = 1, nVertices
          if (interHi(i)%isAvailable) then
            precedingVertex = i
        exit
          end if
        end do
        interHi(precedingVertex)%isAvailable = .false.
        subPolys(nSubs, nPoints(nSubs))%y = yHi
        subPolys(nSubs, nPoints(nSubs))%x = interHi(precedingVertex)%x

        if (zone(precedingVertex) .eq. HIGH) then
          state = MIDDLE
        else
          state = HIGH
        end if

if (testFlag) print *, 'pvtx, zone(pvtx), state:', precedingVertex, zone(precedingVertex), state
      else ! there must be an available vertex.
        do i = 1, nVertices
          if (vertexIsAvailable(i)) then
            precedingVertex = i
        exit
          end if
        end do
        vertexIsAvailable(precedingVertex) = .false.
        subPolys(nSubs, nPoints(nSubs)) = poly(precedingVertex)
if (testFlag) then
  print *, '>>>'
  print *, 'poly(pvtx)%x, poly(pvtx)%y:'&
    , poly(precedingVertex)%x, poly(precedingVertex)%y
  print *, 'subPolys(i, j)%x, subPolys(i, j)%y:'&
    , subPolys(nSubs, nPoints(nSubs))%x&
    , subPolys(nSubs, nPoints(nSubs))%y
  print *, '<<<'
end if
        state = MIDDLE
if (testFlag) print *, 'pvtx, state:', precedingVertex, state
      end if

      ! Now follow it around until the end:
      do
        select case(state)
        case(LOW)
        ! Move down until next lower intercept, after which state ought to
        ! be MIDDLE.
! NOTE!! This is the REVERSE to verticalSlice.
          mask = (interLo%exists&
          .and. interLo%x .lt. subPolys(nSubs, nPoints(nSubs))%x)
if (.not. any(mask)) then
!***** BUG!!
status = -2
return
end if

          indexVec = maxloc(interLo%x, mask)
          precedingVertex = indexVec(1)

      if (.not. interLo(precedingVertex)%isAvailable) exit

if (zone(precedingVertex) .ne. LOW) then
!***** BUG!!
status = -3
return
end if

          nPoints(nSubs) = nPoints(nSubs) + 1
          interLo(precedingVertex)%isAvailable = .false.
          subPolys(nSubs, nPoints(nSubs))%y = yLo
          subPolys(nSubs, nPoints(nSubs))%x = interLo(precedingVertex)%x

          state = MIDDLE

        case(HIGH)
        ! Move up until next higher intercept, after which state ought to
        ! be MIDDLE.
! NOTE!! This is the REVERSE to verticalSlice.
          mask = (interHi%exists&
          .and. interHi%x .gt. subPolys(nSubs, nPoints(nSubs))%x)
if (.not. any(mask)) then
!***** BUG!!
print *, 'here'
status = -4
return
end if
          indexVec = minloc(interHi%x, mask)
          precedingVertex = indexVec(1)
      if (.not. interHi(precedingVertex)%isAvailable) exit

if (zone(precedingVertex) .ne. HIGH) then
!***** BUG!!
status = -5
return
end if

          nPoints(nSubs) = nPoints(nSubs) + 1
          interHi(precedingVertex)%isAvailable = .false.
          subPolys(nSubs, nPoints(nSubs))%y = yHi
          subPolys(nSubs, nPoints(nSubs))%x = interHi(precedingVertex)%x

          state = MIDDLE

        case(MIDDLE)
          if (interLo(precedingVertex)%isAvailable) then
            nPoints(nSubs) = nPoints(nSubs) + 1
            interLo(precedingVertex)%isAvailable = .false.
            subPolys(nSubs, nPoints(nSubs))%y = yLo
            subPolys(nSubs, nPoints(nSubs))%x = interLo(precedingVertex)%x

            state = LOW

          else&
          if (interHi(precedingVertex)%isAvailable) then
            nPoints(nSubs) = nPoints(nSubs) + 1
            interHi(precedingVertex)%isAvailable = .false.
            subPolys(nSubs, nPoints(nSubs))%y = yHi
            subPolys(nSubs, nPoints(nSubs))%x = interHi(precedingVertex)%x

            state = HIGH

          else
            precedingVertex = 1 + modulo(precedingVertex, nVertices)
      if (.not. vertexIsAvailable(precedingVertex)) exit

            nPoints(nSubs) = nPoints(nSubs) + 1
            vertexIsAvailable(precedingVertex) = .false.
            subPolys(nSubs, nPoints(nSubs)) = poly(precedingVertex)

          end if

        case default
!***** BUG!!
          status = -6
  return
        end select
      end do ! complete sub-poly
    end do ! no more available vertices or intersections.

  end subroutine horizontalSlice

!  !---------------------------------------------------------------------
!  subroutine verticalSlice_new(poly, xLo, xHi, subPolys, status)
!  !---------------------------------------------------------------------
!
!    implicit none 
!
!    type(Point2dT_new),       intent(in)  :: poly
!    real(single),              intent(in)  :: xLo, xHi
!    type(localPolyVectorType), pointer     :: subPolys(:)
!!***** REMEMBER to deallocate both the %poly pointers of subPolys and also subPolys itself.
!    integer,                   intent(out) :: status
!
!    type :: crossingType
!      logical :: available(4), onLowBound(4), outwards(4)
!      integer :: numCrossingsThisSide ! 0 to 4
!      real :: crossVal(4)
!    end type crossingType
!    type(crossingType) :: crossing(size(poly))
!    type(localPolyVectorType), allocatable :: tempSubPolys(:)
!
!    !-------------------------------------------------------------------
!    ! Start:
!
!    numVertices = size(poly)
!    if (numVertices <= 0) then
!      allocate(subPolys(0))
!  return
!    end if
!
!!***** load the crossing vector
!
!    ! Deal with the case that poly intersects neither bound:
!
!    if (.not. any(crossing%numCrossingsThisSide > 0)) then
!      if (all(poly%x >= xLo .and. poly%x <= xHi)) then
!        allocate(subPolys(1))
!        call copyPoly_new(poly, subPolys(1)%poly)
!      else
!        allocate(subPolys(0))
!      end if
!
!  return
!    end if
!
!!vvvvv
!    ! Check (*****pro tem??) that crossings are paired, each pair being
!    ! exit-entry of the same bound.
!
!    numCrossings = sum(crossing%numCrossingsThisSide)
!if (isOdd(numCrossings)) call error('','')
!
!    nextSideWithCrossing&
!      = smallestTrueIndex(logical(crossing%numCrossingsThisSide > 0, boolean))
!    lastCrossingWasOnLowBound = crossing(nextSideWithCrossing)%onLowBound(1)
!    lastCrossingWasOutwards   = crossing(nextSideWithCrossing)%isOutwards(1)
!
!    do i = 2, crossing(nextSideWithCrossing)%numCrossingsThisSide
!      if (lastCrossingWasOutwards&
!      .eqv. crossing(nextSideWithCrossing)%isOutwards(i))&
!call error('','') ! because outwards and inwards should alternate.
!
!      if (lastCrossingWasOutwards) then
!        if (lastCrossingWasOnLowBound&
!        .neqv. crossing(nextSideWithCrossing)%onLowBound(i))&
!call error('','')
!      end if
!
!      lastCrossingWasOnLowBound = crossing(nextSideWithCrossing)%onLowBound(i)
!      lastCrossingWasOutwards   = crossing(nextSideWithCrossing)%isOutwards(i)
!    end do
!
!    do while(any(crossing(nextSideWithCrossing+1:)%numCrossingsThisSide > 0))
!
!      nextSideWithCrossing = smallestTrueIndex(logical(&
!        (crossing(nextSideWithCrossing+1:)%numCrossingsThisSide > 0), boolean&
!      ))
!
!      do i = 1, crossing(nextSideWithCrossing)%numCrossingsThisSide
!        if (lastCrossingWasOutwards&
!        .eqv. crossing(nextSideWithCrossing)%isOutwards(i))&
!call error('','') ! because outwards and inwards should alternate.
!
!        if (lastCrossingWasOutwards) then
!          if (lastCrossingWasOnLowBound&
!          .neqv. crossing(nextSideWithCrossing)%onLowBound(i))&
!call error('','')
!        end if
!
!        lastCrossingWasOnLowBound = crossing(nextSideWithCrossing)%onLowBound(i)
!        lastCrossingWasOutwards   = crossing(nextSideWithCrossing)%isOutwards(i)
!      end do
!    end do
!
!!^^^^^
!
!    ! Now find the results of cutting off those bits of the poly which
!    ! lie outside the bounds:
!
!    maxNumVertices = numVertices + sum(crossing%numCrossingsThisSide)
!    allocate(tempSubPolys(maxNumVertices))
!    ! Can't be more sub-polys than this, in fact it is an over-estimate by
!    ! a factor of three.
!
!    numSubPolys = 0
!    do while(any(crossing%numCrossingsThisSide > 0 .and. crossing%available))
!      numSubPolys = numSubPolys + 1
!      allocate(tempSubPolys(numSubPolys)%poly(maxNumVertices))
!
!      !*** Find next available outward crossing:
!
!!!!      nextSideWithCrossing = smallestTrueIndex(logical((), boolean))
!
!      numPoints = 1
!      tempSubPolys(numSubPolys)%poly(numPoints)&
!        = copyPos(crossing(outwardSideNum)(outwardIndex))
!
!      !*** flag this unavailable;
!
!      !*** find the paired inward crossing;
!
!      numPoints = numPoints + 1
!      tempSubPolys(numSubPolys)%poly(numPoints)&
!        = copyPos(crossing(inwardSideNum)(inwardIndex))
!
!      !*** flag this unavailable;
!
!      sideNum = inwardSideNum
!      do while !*** not back at starting point
!        !*** work around the poly until a crossing is reached. If this is not outwards and available, BUG ERROR!
!
!        if (inwardIndex) then
!        end if
!
!        sideNum = sideNum + 1
!        if (sideNum > numVertices) sideNum = 1
!
!        if () then
!        else
!        end if
!
!      end do
!    end do
!
!
!
!
!
!
!
!    deallocate(tempSubPolys)
!
!  end subroutine verticalSlice_new

  !---------------------------------------------------------------------
  subroutine regridCartesianToPolar(cartesianImage, angles, radii&
  , centreXPixels, centreYPixels&
  , pixelSizeX, pixelSizeY, preserveDensity, polarImage)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: cartesianImage(:,:),&
                                 angles(:),&
                                 radii(:),&
                                 centreXPixels,&
                                 centreYPixels,&
                                 pixelSizeX,&
                                 pixelSizeY
    logical,      intent(in)  :: preserveDensity
    real(single), intent(out) :: polarImage(size(angles),size(radii))

    real(single), allocatable :: tempCartesianImage(:,:)

    !-------------------------------------------------------------------

    ! Start:

    allocate(tempCartesianImage(size(cartesianImage,1),size(cartesianImage,2)))
    tempCartesianImage = cartesianImage
    call cartesianPolarKernel(angles, radii, centreXPixels, centreYPixels&
      , pixelSizeX, pixelSizeY&
      , preserveDensity, .true., tempCartesianImage, polarImage)

  end subroutine

  !---------------------------------------------------------------------
  subroutine regridPolarToCartesian (polarImage, angles, radii&
  , centreXPixels, centreYPixels&
  , pixelSizeX, pixelSizeY, preserveDensity, cartesianImage)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: angles(:),&
                                 radii(:),&
                                 polarImage(size(angles),size(radii)),&
                                 centreXPixels,&
                                 centreYPixels,&
                                 pixelSizeX,&
                                 pixelSizeY
    logical,      intent(in)  :: preserveDensity
    real(single), intent(out) :: cartesianImage(:,:)

    real(single), allocatable :: tempPolarImage(:,:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(tempPolarImage(size(polarImage,1),size(polarImage,2)))
    tempPolarImage = polarImage
    call cartesianPolarKernel(angles, radii, centreXPixels, centreYPixels&
      , pixelSizeX, pixelSizeY&
      , preserveDensity, .false., cartesianImage, tempPolarImage)

  end subroutine regridPolarToCartesian

  !---------------------------------------------------------------------
  subroutine cartesianPolarKernel(angles, radii, centreXPixels, centreYPixels&
  , pixelSizeX, pixelSizeY, preserveDensity, cartToPol&
  , cartesianImage, polarImage)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)    :: angles(:),&
                                   radii(:),&
                                   centreXPixels,&
                                   centreYPixels,&
                                   pixelSizeX,&
                                   pixelSizeY
    logical,      intent(in)    :: preserveDensity, cartToPol 
    real(single), intent(inout) :: cartesianImage(:,:),&
                                   polarImage(size(angles),size(radii))

    integer ::&
      numAngles, numRadii, angleNum, radiusNum, polyNum, i,&
      wedgeStatus, annulusStatus, circleStatus
    integer(int16) :: xSize, ySize, xi, yi
    real(single), allocatable ::&
      radiiOfCart(:,:), anglesOfCart(:,:), cartXEdges(:), cartYEdges(:),&
      polarPixelArea(:,:), cosAngles(:), sinAngles(:)
    type(Point2DT), allocatable :: xyOfCart(:,:)
    real(single) ::&
      cartPixelArea, annulusArea, maxPolygonRadius, overlapArea, normArea
    type(CurvedSideT) :: square(4)
    type(CurvedSideT), pointer :: polygon(:) => null()
    type(PolycurveVectorT), pointer :: polyVec(:) => null()

    !-------------------------------------------------------------------
    ! Start:

    xSize = size(cartesianImage,1)
    ySize = size(cartesianImage,2)
    numAngles = size(angles)
    numRadii  = size(radii)
  if (xSize <= 0 .or. ySize <= 0 .or. numAngles <= 2 .or. numRadii <= 0) return
    if (any(angles(2:) <= angles(:numAngles-1)))&
call error('nonSortedAngles',"ssclib:cartesianPolarKernel: The values of the &
&input vector argument angles() must be ordered in an increasing sequence.")

    if (any(angles(2:) - angles(:numAngles-1) >= Pi))&
call error('angleGapTooBig',"ssclib:cartesianPolarKernel: Two successive &
&values of the input vector argument angles() are separated by more than &
&pi. This is not permitted.")

    if (angles(1) .ne. 0.0)&
call error('nonZeroFirstAngle',"ssclib:cartesianPolarKernel: The first element &
&of the input vector argument angles() must equal zero.")

    if (any(radii <= 0.0))&
call error('nonPositiveRadii',"ssclib:cartesianPolarKernel: All values of the &
&input vector argument radii() must be > 0.")

    if (any(radii(2:) <= radii(:numRadii-1)))&
call error('nonSortedRadii',"ssclib:cartesianPolarKernel: The values of the &
&input vector argument radii() must be ordered in an increasing sequence.")

    allocate(cartXEdges(0:xSize), cartYEdges(0:ySize))
    cartXEdges = (/(i+0.5-centreXPixels, i=0,xSize)/) * pixelSizeX
    cartYEdges = (/(i+0.5-centreYPixels, i=0,ySize)/) * pixelSizeY
!*** use tanToPixels??

    allocate(xyOfCart(0:xSize,0:ySize))
    xyOfCart%x = spread(cartXEdges,ySize+1,2)
    xyOfCart%y = spread(cartYEdges,xSize+1,1)

    allocate(radiiOfCart(0:xSize,0:ySize))
    allocate(anglesOfCart(0:xSize,0:ySize))
    radiiOfCart = sqrt(xyOfCart%x * xyOfCart%x + xyOfCart%y * xyOfCart%y)
    anglesOfCart = atan2(xyOfCart%y, xyOfCart%x)

    cartPixelArea = pixelSizeX * pixelSizeY

    allocate(cosAngles(numAngles), sinAngles(numAngles))
    cosAngles = cos(angles)
    sinAngles = sin(angles)

    allocate(polarPixelArea(numAngles,numRadii))
    annulusArea = 0.5 * radii(1)*radii(1)
    do angleNum = 1, numAngles
      polarPixelArea(angleNum, 1)&
        = annulusArea * (angles(angleNum+1) - angles(angleNum))
    end do
    polarPixelArea(numAngles, 1)&
      = annulusArea * (TwoPi - angles(numAngles))

    do radiusNum = 2, numRadii
      annulusArea = 0.5 * (radii(radiusNum)*radii(radiusNum)&
        - radii(radiusNum-1)*radii(radiusNum-1))

      do angleNum = 1, numAngles-1
        polarPixelArea(angleNum, radiusNum)&
          = annulusArea * (angles(angleNum+1) - angles(angleNum))
      end do
      polarPixelArea(numAngles, radiusNum)&
        = annulusArea * (TwoPi - angles(numAngles))
    end do

    square%isStraight = .true.

    do angleNum = 1, numAngles
      do xi = 1, xSize
        YILOOP: do yi = 1, ySize

          square%radius = loadSquare(radiiOfCart,  xi, yi)
          square%angle  = loadSquare(anglesOfCart, xi, yi)
          square%pos    = loadSquare(xyOfCart,     xi, yi)

          call calcPartOfPolygonInWedge(square, angles(angleNum)&
            , angles(angleNum+1), cosAngles(angleNum), sinAngles(angleNum)&
            , cosAngles(angleNum+1), sinAngles(angleNum+1), polygon&
            , wedgeStatus)

          if (wedgeStatus > 0) then
!*** report
          else if (wedgeStatus < 0) then
!*** report
            deallocate(polygon)
        cycle YILOOP
          end if

          if (size(polygon) <= 0) then
            deallocate(polygon)
        cycle YILOOP
          end if

          maxPolygonRadius = maxval(polygon%radius)
          do radiusNum = numRadii, 2, -1
          if (maxPolygonRadius <= radii(radiusNum-1)) cycle

            call calcPartOfPolygonInAnnulus(polygon, radii(radiusNum-1)&
              , radii(radiusNum), polyVec, annulusStatus)

            if (annulusStatus > 0) then
!*** report
            else if (annulusStatus < 0) then
!*** report
              call deallocatePolycurveVectorT(polyVec)
              deallocate(polygon)
        cycle YILOOP
            end if

            if (size(polyVec) <= 0) then
              call deallocatePolycurveVectorT(polyVec)
              deallocate(polygon)
        cycle YILOOP
            end if

            overlapArea = 0.0
            do polyNum = 1, size(polyVec)
              overlapArea = overlapArea&
                + areaOfPolycurve(polyVec(polyNum)%poly)
            end do

            if (cartToPol) then
              if (preserveDensity) then
                normArea = polarPixelArea(angleNum, radiusNum)
              else
                normArea = cartPixelArea
              end if
              polarImage(angleNum, radiusNum) = polarImage(angleNum, radiusNum)&
                + (overlapArea / normArea) * cartesianImage(xi,yi)
            else
              if (preserveDensity) then
                normArea = cartPixelArea
              else
                normArea = polarPixelArea(angleNum, radiusNum)
              end if
              cartesianImage(xi,yi) = cartesianImage(xi,yi)&
                + (overlapArea / normArea) * polarImage(angleNum, radiusNum)
            end if

            call deallocatePolycurveVectorT(polyVec)
          end do

          ! Handle inside the 1st radius separately:
          radiusNum = 1

          call calcPartOfPolygonInCircle(polygon, radii(1), polyVec&
            , circleStatus)

          if (circleStatus > 0) then
!*** report
          else if (circleStatus < 0) then
!*** report
            call deallocatePolycurveVectorT(polyVec)
            deallocate(polygon)
        cycle YILOOP
          end if

          if (size(polyVec) <= 0) then
            call deallocatePolycurveVectorT(polyVec)
            deallocate(polygon)
        cycle YILOOP
          end if

          overlapArea = 0.0
          do polyNum = 1, size(polyVec)
            overlapArea = overlapArea&
              + areaOfPolycurve(polyVec(polyNum)%poly)
          end do

          if (cartToPol) then
            if (preserveDensity) then
              normArea = polarPixelArea(angleNum, radiusNum)
            else
              normArea = cartPixelArea
            end if
            polarImage(angleNum, radiusNum) = polarImage(angleNum, radiusNum)&
              + (overlapArea / normArea) * cartesianImage(xi,yi)
          else
            if (preserveDensity) then
              normArea = cartPixelArea
            else
              normArea = polarPixelArea(angleNum, radiusNum)
            end if
            cartesianImage(xi,yi) = cartesianImage(xi,yi)&
              + (overlapArea / normArea) * polarImage(angleNum, radiusNum)
          end if

          call deallocatePolycurveVectorT(polyVec)
          deallocate(polygon)
        end do YILOOP
      end do
    end do

  end subroutine cartesianPolarKernel

  !---------------------------------------------------------------------
  function loadSquareSingle(arrayVals, xi, yi) result(squareVals)
  !---------------------------------------------------------------------

    implicit none 

    real(single),   intent(in) :: arrayVals(:,:)
    integer(int16), intent(in) :: xi, yi
    real(single)               :: squareVals(4)

    !-------------------------------------------------------------------
    ! Start:

    squareVals(1) = arrayVals(xi-1,yi-1)
    squareVals(2) = arrayVals(xi-1,yi)
    squareVals(3) = arrayVals(xi  ,yi)
    squareVals(4) = arrayVals(xi  ,yi-1)

  end function loadSquareSingle

  !---------------------------------------------------------------------
  function loadSquarePos(arrayVals, xi, yi) result(squareVals)
  !---------------------------------------------------------------------

    implicit none 

    type(Point2dT), intent(in) :: arrayVals(:,:)
    integer(int16), intent(in) :: xi, yi
    type(Point2dT)             :: squareVals(4)

    !-------------------------------------------------------------------
    ! Start:

    squareVals(1) = arrayVals(xi-1,yi-1)
    squareVals(2) = arrayVals(xi-1,yi)
    squareVals(3) = arrayVals(xi  ,yi)
    squareVals(4) = arrayVals(xi  ,yi-1)

  end function loadSquarePos

  !---------------------------------------------------------------------
  subroutine calcPartOfPolygonInWedge(polygonIn, angleLo, angleHi&
  , cosAngleLo, sinAngleLo, cosAngleHi, sinAngleHi, polygonOut, status)
  !---------------------------------------------------------------------
  ! Note! The calling routine should deallocate the pointer polygonOut
  ! after use.

    implicit none 

    type(CurvedSideT), intent(inout) :: polygonIn(:)
    real(single),      intent(in)    :: angleLo, angleHi,&
                                        cosAngleLo, sinAngleLo,&
                                        cosAngleHi, sinAngleHi
    type(CurvedSideT), pointer       :: polygonOut(:)
    integer,           intent(out)   :: status

    type(CurvedSideT) :: polygonMid(12), tempPolygonOut(12)
    integer :: numPtsIn, numPtsMid, si, siNext, numPtsOut, pi
    logical :: includeCentre

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    if (angleLo >= angleHi)&
call error('nonOrderedAngles',"ssclib:calcPartOfPolygonInWedge: Re the input &
&angle arguments: angleLo must be < angleHi.")

    numPtsIn = size(polygonIn)

    call calculateFields(polygonIn%angle, angleLo, angleHi, polygonIn%field)
    where(polygonIn%pos%x == 0.0 .and. polygonIn%pos%y == 0.0)&
      polygonIn%field = ON_CENTRE

    if (.not. any(polygonIn%field == INSIDE_WEDGE)) then
      if (.not. ((any(polygonIn%field == UPPER_LEFT&
      .or.            polygonIn%field == ON_LEFT_RAY)&
      .and.       any(polygonIn%field == ON_RIGHT_RAY&
      .or.            polygonIn%field == UPPER_RIGHT&
      .or.            polygonIn%field == LOWER_RIGHT))&
      .or.       (any(polygonIn%field == ON_LEFT_RAY&
      .or.            polygonIn%field == UPPER_LEFT&
      .or.            polygonIn%field == LOWER_LEFT)&
      .and.       any(polygonIn%field == UPPER_RIGHT&
      .or.            polygonIn%field == ON_RIGHT_RAY)))) then
        allocate(polygonOut(0))
  return
      end if
    end if

    if (all(polygonIn%field == INSIDE_WEDGE&
    .or.         polygonIn%field == ON_LEFT_RAY&
    .or.         polygonIn%field == ON_RIGHT_RAY)) then
      allocate(polygonOut(numPtsIn))
      polygonOut = polygonIn
  return
    end if

    ! Next construct a chain of CurvedSideT points comprising the vertices of the original polyCurveIn, plus any intersection points between this and the wedge, in order as they are encountered clockwise around the polyCurve.

!    polygonMid%centre%x = 0.0
!    polygonMid%centre%y = 0.0

    numPtsMid = 1
    do si = 1, numPtsIn
      siNext = 1 + modulo(si, numPtsIn)

      polygonMid(numPtsMid) = polygonIn(si)

      call calcIntersectsLineWedge(polygonIn(si), polygonIn(siNext)&
        , angleLo, angleHi, cosAngleLo, sinAngleLo, cosAngleHi, sinAngleHi&
        , polygonMid, numPtsMid, status)

      if (status < 0) then
        allocate(polygonOut(0))
  return
      end if
    end do

    ! Now calculate the polygon:

    if (.not. any(polygonMid(1:numPtsMid)%field == INSIDE_WEDGE)&
    .and. count(polygonMid(1:numPtsMid)%field == ON_LEFT_RAY&
    .or.        polygonMid(1:numPtsMid)%field == ON_RIGHT_RAY) <= 1) then
    ! There is no overlap.
      allocate(polygonOut(0))
  return
    end if

    includeCentre = .false. ! default
    if (pointIsInsidePoly(0.0, 0.0, polygonMid(1:numPtsMid)%pos, .true.))&
      includeCentre = .true.

    numPtsOut = 0
    do pi = 1, numPtsMid
      if (.not. (polygonMid(pi)%field == INSIDE_WEDGE&
      .or.       polygonMid(pi)%field == ON_LEFT_RAY&
      .or.       polygonMid(pi)%field == ON_RIGHT_RAY))&
    cycle

      numPtsOut = numPtsOut + 1
      tempPolygonOut(numPtsOut) = polygonMid(pi)

      if (polygonMid(pi)%field == ON_RIGHT_RAY .and. includeCentre) then
      ! only works if the polygonMid is convex!
        numPtsOut = numPtsOut + 1

        tempPolygonOut(numPtsOut)%angle = 0.0
        tempPolygonOut(numPtsOut)%radius = 0.0
        tempPolygonOut(numPtsOut)%pos%x = 0.0
        tempPolygonOut(numPtsOut)%pos%y = 0.0
        tempPolygonOut(numPtsOut)%field = ON_CENTRE
        tempPolygonOut(numPtsOut)%isStraight = .true.
      end if
    end do

    allocate(polygonOut(numPtsOut))
    polygonOut = tempPolygonOut(1:numPtsOut)

  end subroutine calcPartOfPolygonInWedge

  !---------------------------------------------------------------------
  subroutine calculateFields(polyAngles, angleLo, angleHi, polyFields)
  !---------------------------------------------------------------------

    implicit none

    real(single),  intent(in)  :: polyAngles(:), angleLo, angleHi
    integer(int8), intent(out) :: polyFields(size(polyAngles))

    real(single), allocatable :: tempPolyAngles(:)
    integer :: numPolyAngles
    real(single) :: angleMid, angleHalf

    !-------------------------------------------------------------------
    ! Start:

    numPolyAngles = size(polyAngles)
    allocate(tempPolyAngles(numPolyAngles))
    angleMid  = (angleHi + angleLo) / 2.0
    angleHalf = (angleHi - angleLo) / 2.0
    tempPolyAngles = modulo(real(polyAngles - angleMid, double), TwoPi)

    where(tempPolyAngles > TwoPi - angleHalf .or. tempPolyAngles < angleHalf)&
      polyFields = INSIDE_WEDGE
    where(tempPolyAngles == angleHalf)&
      polyFields = ON_RIGHT_RAY
    where(tempPolyAngles > angleHalf .and. tempPolyAngles < PiOn2)&
      polyFields = UPPER_RIGHT
    where(tempPolyAngles >= PiOn2 .and. tempPolyAngles < Pi)&
      polyFields = LOWER_RIGHT
    where(tempPolyAngles >= Pi .and. tempPolyAngles < 3.0d0*PiOverTwo)&
      polyFields = LOWER_LEFT
    where(tempPolyAngles >= 3.0d0*PiOverTwo&
    .and. tempPolyAngles < TwoPi - angleHalf)&
      polyFields = UPPER_LEFT
    where(tempPolyAngles == TwoPi - angleHalf)&
      polyFields = ON_LEFT_RAY

  end subroutine calculateFields

  !---------------------------------------------------------------------
  subroutine calcIntersectsLineWedge(vertexLo, vertexHi&
  , angleLo, angleHi, cosAngleLo, sinAngleLo, cosAngleHi, sinAngleHi&
  , ptChain, numPts, status)
  !---------------------------------------------------------------------

    implicit none 

    type(CurvedSideT), intent(in)    :: vertexLo, vertexHi
    real(single),      intent(in)    :: angleLo, angleHi,&
                                        cosAngleLo, sinAngleLo,&
                                        cosAngleHi, sinAngleHi
    type(CurvedSideT), intent(inout) :: ptChain(12)
    integer,           intent(inout) :: numPts
    integer,           intent(out)   :: status

    real(single) :: k(2), e, denom, x, y, angleLoK, angleHiK
    type(Point2DT) :: v, o, cHat
    logical :: loIntersectExists, hiIntersectExists
    integer(int8) :: localFieldLo, localFieldHi

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    ! Get rid of the cases for which there can be no intercept:

    if (vertexLo%field == ON_CENTRE .or. vertexHi%field == ON_CENTRE)&
  return

    select case(vertexLo%field)
    case(INSIDE_WEDGE)
      select case(vertexHi%field)
      case(INSIDE_WEDGE, ON_RIGHT_RAY, ON_LEFT_RAY)
      end select
    case(ON_RIGHT_RAY)
      select case(vertexHi%field)
      case(INSIDE_WEDGE, ON_RIGHT_RAY, UPPER_RIGHT, LOWER_RIGHT, ON_LEFT_RAY)
      end select
    case(UPPER_RIGHT)
      select case(vertexHi%field)
      case(ON_RIGHT_RAY, UPPER_RIGHT, LOWER_RIGHT)
      end select
    case(LOWER_RIGHT)
      select case(vertexHi%field)
      case(ON_RIGHT_RAY, UPPER_RIGHT, LOWER_RIGHT, LOWER_LEFT)
      end select
    case(LOWER_LEFT)
      select case(vertexHi%field)
      case(LOWER_RIGHT, LOWER_LEFT, UPPER_LEFT, ON_LEFT_RAY)
      end select
    case(UPPER_LEFT)
      select case(vertexHi%field)
      case(LOWER_LEFT, UPPER_LEFT, ON_LEFT_RAY)
      end select
    case(ON_LEFT_RAY)
      select case(vertexHi%field)
      case(INSIDE_WEDGE, ON_RIGHT_RAY, LOWER_LEFT, UPPER_LEFT, ON_LEFT_RAY)
      end select
    end select

    ! Now work out the intercepts:

    k = 0.0 ! just to give them some value - don't like
    ! floating variables, messy untidy things.

    ! Let us denote the vector between the origin and vertexLo%pos by _o_, the vector from vertexLo%pos to vertexHi%pos by _v_. If we extend the side _v_ in both directions then this line may intersect with a ray from the origin at zero or one points. If the ray direction is denoted by c^ then the equation for the intersection point is _c_ = e*c^ = _o_ + k*_v_, for some constants e and k. In terms of the vector components we have 2 equations: 
    !
    !	e*c^_x - k*v_x = o_x
    !	e*c^_y - k*v_y = o_y.
    !
    ! These equations can be inverted to find e and k as follows:
    !
    !	      v_y*o_x - v_x*o_y
    !	e = ---------------------,
    !	     c^_x*v_y - c^_y*v_x
    !
    !	     c^_y*o_x - c^_x*o_y
    !	k = ---------------------.
    !	     c^_x*v_y - c^_y*v_x
    !
    ! Note that if the denominator equals zero, this means that _v_ and _c_ are parallel and there is no intersection point; if e < zero it means that the line _v_ does not intercept the ray c^ in the positive half-plane and the solution should therefore be discarded.

    v%x = vertexHi%pos%x - vertexLo%pos%x
    v%y = vertexHi%pos%y - vertexLo%pos%y
    o = vertexLo%pos

    cHat%x = cosAngleLo
    cHat%y = sinAngleLo

    loIntersectExists = .false. ! default
    denom = cHat%x * v%y - cHat%y * v%x
    if (denom .ne. 0.0) then
      e = (v%y * o%x - v%x * o%y) / denom
      if (e > 0.0) then
        k(1) = (cHat%y * o%x - cHat%x * o%y) / denom
        loIntersectExists = .true.
      end if
    end if

    cHat%x = cosAngleHi
    cHat%y = sinAngleHi

    hiIntersectExists = .false. ! default
    denom = cHat%x * v%y - cHat%y * v%x
    if (denom .ne. 0.0) then
      e = (v%y * o%x - v%x * o%y) / denom
      if (e > 0.0) then
        k(2) = (cHat%y * o%x - cHat%x * o%y) / denom
        hiIntersectExists = .true.
      end if
    end if

  if (.not. (loIntersectExists .or. hiIntersectExists)) return

    if (.not. loIntersectExists) then
      if (k(2) <= 0.0) then
        select case(vertexLo%field)
        case(INSIDE_WEDGE)
          select case(vertexHi%field)
          case(LOWER_LEFT)
            status = 1
            k(2) = epsilon(1.0)
          case(LOWER_RIGHT)
            status = 2
            k(2) = epsilon(1.0)
          case(UPPER_RIGHT)
            status = 3
            k(2) = epsilon(1.0)
          case default
            status = -1
          end select
        case(LOWER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = -2
          case(UPPER_RIGHT)
            status = -3
          case(ON_RIGHT_RAY)
            status = -4
          case default
            status = -5
          end select
        case(LOWER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = -6
          case default
            status = -7
          end select
        case(UPPER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = 4
            k(2) = epsilon(1.0)
          case(LOWER_LEFT)
            ! fine
          case default
            status = -8
          end select
        case(ON_RIGHT_RAY)
          select case(vertexHi%field)
          case(LOWER_LEFT)
            ! fine
          case default
            status = -9
          end select
        case default
          status = -10
        end select
      else if (k(2) > 0.0 .and. k(2) < 1.0) then
        select case(vertexLo%field)
        case(INSIDE_WEDGE)
          select case(vertexHi%field)
          case(LOWER_LEFT)
            ! fine
          case(LOWER_RIGHT)
            ! fine
          case(UPPER_RIGHT)
            ! fine
          case default
            status = -11
          end select
        case(LOWER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            ! fine
          case(UPPER_RIGHT)
            status = 5
!            k(2) = 1.0 + epsilon(1.0)
  return
          case(ON_RIGHT_RAY)
            status = 6
!            k(2) = 1.0
  return
          case default
            status = -12
          end select
        case(LOWER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            ! fine
          case default
            status = -13
          end select
        case(UPPER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            ! fine
          case(LOWER_LEFT)
            status = 7
!            k(2) = -epsilon(1.0)
  return
          case default
            status = -14
          end select
        case(ON_RIGHT_RAY)
          select case(vertexHi%field)
          case(LOWER_LEFT)
            status = 8
!            k(2) = 0.0
  return
          case default
            status = -15
          end select
        case default
          status = -16
        end select
      else ! k(2) >= 1.0
        select case(vertexLo%field)
        case(INSIDE_WEDGE)
          select case(vertexHi%field)
          case(LOWER_LEFT)
            status = -17
          case(LOWER_RIGHT)
            status = -18
          case(UPPER_RIGHT)
            status = 9
            k(2) = 1.0 - epsilon(1.0)
          case default
            status = -19
          end select
        case(LOWER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = 10
            k(2) = 1.0 - epsilon(1.0)
          case(UPPER_RIGHT)
            ! fine
          case(ON_RIGHT_RAY)
            ! fine
          case default
            status = -20
          end select
        case(LOWER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = 11
            k(2) = 1.0 - epsilon(1.0)
          case default
            status = -21
          end select
        case(UPPER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = 12
            k(2) = 1.0 - epsilon(1.0)
          case(LOWER_LEFT)
            status = -22
          case default
            status = -23
          end select
        case(ON_RIGHT_RAY)
          select case(vertexHi%field)
          case(LOWER_LEFT)
            status = -24
          case default
            status = -25
          end select
        case default
          status = -26
        end select
      end if

if (status < 0) return

      if (k(2) > 0.0 .and. k(2) < 1.0) then
        numPts = numPts + 1
        x = o%x + k(2) * v%x
        y = o%y + k(2) * v%y
        ptChain(numPts)%angle = angleHi
        ptChain(numPts)%radius = sqrt(x * x + y * y)
        ptChain(numPts)%pos%x = x
        ptChain(numPts)%pos%y = y
        ptChain(numPts)%field = ON_RIGHT_RAY
        ptChain(numPts)%isStraight = .true.
      end if

  return
    end if ! .not. loIntersectExists

    if (.not. hiIntersectExists) then
      if (k(1) <= 0.0) then
        select case(vertexLo%field)
        case(INSIDE_WEDGE)
          select case(vertexHi%field)
          case(LOWER_RIGHT)
            status = 13
            k(1) = epsilon(1.0)
          case(LOWER_LEFT)
            status = 14
            k(1) = epsilon(1.0)
          case(UPPER_LEFT)
            status = 15
            k(1) = epsilon(1.0)
          case default
            status = -27
          end select
        case(LOWER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = -28
          case(UPPER_LEFT)
            status = -29
          case(ON_LEFT_RAY)
            status = -30
          case default
            status = -31
          end select
        case(LOWER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = -32
          case default
            status = -33
          end select
        case(UPPER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = 16
            k(1) = epsilon(1.0)
          case(LOWER_RIGHT)
            ! fine
          case default
            status = -34
          end select
        case(ON_LEFT_RAY)
          select case(vertexHi%field)
          case(LOWER_RIGHT)
            ! fine
          case default
            status = -35
          end select
        case default
          status = -36
        end select
      else if (k(1) > 0.0 .and. k(1) < 1.0) then
        select case(vertexLo%field)
        case(INSIDE_WEDGE)
          select case(vertexHi%field)
          case(LOWER_RIGHT)
            ! fine
          case(LOWER_LEFT)
            ! fine
          case(UPPER_LEFT)
            ! fine
          case default
            status = -37
          end select
        case(LOWER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            ! fine
          case(UPPER_LEFT)
            status = 17
!            k(1) = 1.0 + epsilon(1.0)
  return
          case(ON_LEFT_RAY)
            status = 18
!            k(1) = 1.0
  return
          case default
            status = -38
          end select
        case(LOWER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            ! fine
          case default
            status = -39
          end select
        case(UPPER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            ! fine
          case(LOWER_RIGHT)
            status = 19
!            k(1) = -epsilon(1.0)
  return
          case default
            status = -40
          end select
        case(ON_LEFT_RAY)
          select case(vertexHi%field)
          case(LOWER_RIGHT)
            status = 20
!            k(1) = 0.0
  return
          case default
            status = -41
          end select
        case default
          status = -42
        end select
      else ! k(1) >= 1.0
        select case(vertexLo%field)
        case(INSIDE_WEDGE)
          select case(vertexHi%field)
          case(LOWER_RIGHT)
            status = -43
          case(LOWER_LEFT)
            status = -44
          case(UPPER_LEFT)
            status = 21
            k(1) = 1.0 - epsilon(1.0)
          case default
            status = -45
          end select
        case(LOWER_RIGHT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = 22
            k(1) = 1.0 - epsilon(1.0)
          case(UPPER_LEFT)
            ! fine
          case(ON_LEFT_RAY)
            ! fine
          case default
            status = -46
          end select
        case(LOWER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = 23
            k(1) = 1.0 - epsilon(1.0)
          case default
            status = -47
          end select
        case(UPPER_LEFT)
          select case(vertexHi%field)
          case(INSIDE_WEDGE)
            status = 24
            k(1) = 1.0 - epsilon(1.0)
          case(LOWER_RIGHT)
            status = -48
          case default
            status = -49
          end select
        case(ON_LEFT_RAY)
          select case(vertexHi%field)
          case(LOWER_RIGHT)
            status = -50
          case(LOWER_LEFT)
            status = -51
          case default
            status = -52
          end select
        case default
          status = -53
        end select
      end if

  if (status < 0) return

      if (k(1) > 0.0 .and. k(1) < 1.0) then
        numPts = numPts + 1
        x = o%x + k(1) * v%x
        y = o%y + k(1) * v%y
        ptChain(numPts)%angle = angleLo
        ptChain(numPts)%radius = sqrt(x * x + y * y)
        ptChain(numPts)%pos%x = x
        ptChain(numPts)%pos%y = y
        ptChain(numPts)%field = ON_LEFT_RAY
        ptChain(numPts)%isStraight = .true.
      end if

  return
    end if ! .not. hiIntersectExists

    ! If have reached here then both intersect points exist:

    select case(vertexLo%field)
    case(UPPER_LEFT, LOWER_LEFT, LOWER_RIGHT, UPPER_RIGHT)
      localFieldLo = OUTSIDE_WEDGE
    case default
      localFieldLo = vertexLo%field
    end select

    select case(vertexHi%field)
    case(UPPER_LEFT, LOWER_LEFT, LOWER_RIGHT, UPPER_RIGHT)
      localFieldHi = OUTSIDE_WEDGE
    case default
      localFieldHi = vertexHi%field
    end select

    ! Sort the k values:
    if (k(1) > k(2)) then
      call swap(k(1), k(2))
!      ON_LOK_RAY = ON_RIGHT_RAY
!      ON_HIK_RAY = ON_LEFT_RAY
if (localFieldLo==ON_RIGHT_RAY) localFieldLo = ON_LOK_RAY
if (localFieldHi==ON_RIGHT_RAY) localFieldHi = ON_LOK_RAY
if (localFieldLo==ON_LEFT_RAY)  localFieldLo = ON_HIK_RAY
if (localFieldHi==ON_LEFT_RAY)  localFieldHi = ON_HIK_RAY

      angleLoK = angleHi
      angleHiK = angleLo
    else
!      ON_LOK_RAY = ON_LEFT_RAY
!      ON_HIK_RAY = ON_RIGHT_RAY
if (localFieldLo==ON_RIGHT_RAY) localFieldLo = ON_HIK_RAY
if (localFieldHi==ON_RIGHT_RAY) localFieldHi = ON_HIK_RAY
if (localFieldLo==ON_LEFT_RAY)  localFieldLo = ON_LOK_RAY
if (localFieldHi==ON_LEFT_RAY)  localFieldHi = ON_LOK_RAY

      angleLoK = angleLo
      angleHiK = angleHi
    end if

    ! Catch all the funny cases, repair if possible, error if not. Note that a more detailed treatment would consider all four external quadrants separately but this would be much more complicated and probably not worth the effort.

    if (k(2) <= 0.0) then ! k(1) must also be <= 0 because we just
    ! sorted them.
      select case(localFieldLo)
      case(OUTSIDE_WEDGE)
        select case(localFieldHi)
        case(OUTSIDE_WEDGE)
          ! fine
        case(ON_LOK_RAY)
          status = -54
        case(ON_HIK_RAY)
          status = -55
        case(INSIDE_WEDGE)
          status = -56
        case default
          status = -57
        end select
      case(ON_LOK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = -58
        else
          status = -59
        end if
      case(ON_HIK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 25
          k(2) = 0.0
        else
          status = -60
        end if
      case(INSIDE_WEDGE)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 26
          k(2) = epsilon(1.0)
        else
          status = -61
        end if
      case default
        status = -62
      end select

    else if (k(2) > 0.0 .and. k(2) < 1.0 .and. k(1) <= 0.0) then
      select case(localFieldLo)
      case(OUTSIDE_WEDGE)
        select case(localFieldHi)
        case(OUTSIDE_WEDGE)
          status = -63
        case(ON_LOK_RAY)
          status = -64
        case(ON_HIK_RAY)
          status = 27
          k(1) = epsilon(1.0)
          k(2) = 1.0
        case(INSIDE_WEDGE)
          status = 28
          k(1) = epsilon(1.0)
          k(2) = 1.0 + epsilon(1.0)
        case default
          status = -65
        end select
      case(ON_LOK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 29
          k(1) = 0.0
        else
          status = -66
        end if
      case(ON_HIK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 30
          k(2) = 0.0
        else
          status = -67
        end if
      case(INSIDE_WEDGE)
        if (localFieldHi==OUTSIDE_WEDGE) then
          ! fine
        else
          status = -68
        end if
      case default
        status = -69
      end select

    else if (k(2) > 0.0 .and. k(2) < 1.0&
    .and.    k(1) > 0.0 .and. k(1) < 1.0) then
      select case(localFieldLo)
      case(OUTSIDE_WEDGE)
        select case(localFieldHi)
        case(OUTSIDE_WEDGE)
          ! fine
        case(ON_LOK_RAY)
          status = -70
        case(ON_HIK_RAY)
          status = 31
          k(2) = 1.0
        case(INSIDE_WEDGE)
          status = 32
          k(2) = 1.0 + epsilon(1.0)
        case default
          status = -71
        end select
      case(ON_LOK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 33
          k(1) = 0.0
        else
          status = -72
        end if
      case(ON_HIK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = -73
        else
          status = -74
        end if
      case(INSIDE_WEDGE)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 34
          k(1) = -epsilon(1.0)
        else
          status = -75
        end if
      case default
        status = -76
      end select

    else if (k(2) >= 1.0 .and. k(1) <= 0.0) then
      select case(localFieldLo)
      case(OUTSIDE_WEDGE)
        select case(localFieldHi)
        case(OUTSIDE_WEDGE)
          status = 35
          k(1) = epsilon(1.0)
          k(2) = 1.0 - epsilon(1.0)
        case(ON_LOK_RAY)
          status = -77
        case(ON_HIK_RAY)
          status = 36
          k(1) = epsilon(1.0)
          k(2) = 1.0
        case(INSIDE_WEDGE)
          status = 37
          k(1) = epsilon(1.0)
        case default
          status = -78
        end select
      case(ON_LOK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 38
          k(1) = 0.0
          k(2) = 1.0 - epsilon(1.0)
        else
          status = -79
        end if
      case(ON_HIK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = -80
        else
          status = -81
        end if
      case(INSIDE_WEDGE)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 39
          k(2) = 1.0 - epsilon(1.0)
        else
          status = -82
        end if
      case default
        status = -83
      end select

    else if (k(2) >= 1.0&
    .and.    k(1) > 0.0 .and. k(1) < 1.0) then
      select case(localFieldLo)
      case(OUTSIDE_WEDGE)
        select case(localFieldHi)
        case(OUTSIDE_WEDGE)
          status = -84
        case(ON_LOK_RAY)
          status = 40
          k(1) = 1.0
        case(ON_HIK_RAY)
          status = 41
          k(2) = 1.0
        case(INSIDE_WEDGE)
          ! fine
        case default
          status = -85
        end select
      case(ON_LOK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 42
          k(1) = 0.0
          k(2) = 1.0 - epsilon(1.0)
        else
          status = -86
        end if
      case(ON_HIK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = -87
        else
          status = -88
        end if
      case(INSIDE_WEDGE)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = 43
          k(1) = -epsilon(1.0)
          k(2) = 1.0 - epsilon(1.0)
        else
          status = -89
        end if
      case default
        status = -90
      end select

    else! k(2) >= 1.0 .and. k(1) >= 1.0
      select case(localFieldLo)
      case(OUTSIDE_WEDGE)
        select case(localFieldHi)
        case(OUTSIDE_WEDGE)
          ! fine
        case(ON_LOK_RAY)
          status = 44
          k(1) = 1.0
        case(ON_HIK_RAY)
          status = -91
        case(INSIDE_WEDGE)
          status = 45
          k(1) = 1.0 - epsilon(1.0)
        case default
          status = -92
        end select
      case(ON_LOK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = -93
        else
          status = -94
        end if
      case(ON_HIK_RAY)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = -95
        else
          status = -96
        end if
      case(INSIDE_WEDGE)
        if (localFieldHi==OUTSIDE_WEDGE) then
          status = -97
        else
          status = -98
        end if
      case default
        status = -99
      end select
    end if

  if (status < 0) return

    ! Now calculate the details of the intersection points:

    if (k(1) > 0.0 .and. k(1) < 1.0) then
      numPts = numPts + 1
      x = o%x + k(1) * v%x
      y = o%y + k(1) * v%y
      ptChain(numPts)%angle = angleLoK
      ptChain(numPts)%radius = sqrt(x * x + y * y)
      ptChain(numPts)%pos%x = x
      ptChain(numPts)%pos%y = y
      ptChain(numPts)%field = ON_LOK_RAY
      ptChain(numPts)%isStraight = .true.
    end if
    if (k(2) > 0.0 .and. k(2) < 1.0) then
      numPts = numPts + 1
      x = o%x + k(2) * v%x
      y = o%y + k(2) * v%y
      ptChain(numPts)%angle = angleHiK
      ptChain(numPts)%radius = sqrt(x * x + y * y)
      ptChain(numPts)%pos%x = x
      ptChain(numPts)%pos%y = y
      ptChain(numPts)%field = ON_HIK_RAY
      ptChain(numPts)%isStraight = .true.
    end if

  end subroutine calcIntersectsLineWedge

  !---------------------------------------------------------------------
  subroutine calcPartOfPolygonInAnnulus(polygon, radiusLo, radiusHi, polyVec&
  , status)
  !---------------------------------------------------------------------
  ! Note! The calling routine should deallocate the pointer polyCurve after use.

    implicit none 

    type(CurvedSideT),      intent(inout) :: polygon(:)
    real(single),           intent(in)    :: radiusLo, radiusHi
    type(PolycurveVectorT), pointer       :: polyVec(:)
    integer,                intent(out)   :: status

    type(CurvedSideT) :: ptChain(48) ! The absolute maximum conceivable number
    ! of points which can be in any single one of the output polycurves. This is deduced as follows: the maximum number of points in the input polygon is 12: if we assume that each of these sides crosses the concentric circles 4 times, we get 12x4=48.

    type(PolycurveVectorT) :: tempPolyVec(48) ! This is the maximum number of pieces the intersection area between the input polygon and the annulus can fall into.

!*** actually, should make these allocatable really... otherwise the whole thing is tied to squares as input to calcPartOfPolygonInWedge.

    integer ::&
      numInPts, numMidPts, numTempSides(48), si, siNext,&
      pi, numPolys, oldPi, niLast, polyNum, ni, niNext
    logical, allocatable :: ptAlreadyInPoly(:), mask(:)

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    if (radiusLo >= radiusHi)&
call error('nonOrderedRadii',"ssclib:calcPartOfPolygonInAnnulus:  Re the input &
&radius arguments: radiusLo must be < radiusHi.")

    ! First work out where the vertices of the polygon are in relation to the annulus:

    polygon%field = INSIDE_INNER ! default
    where(polygon%radius > radiusHi)&
      polygon%field = OUTSIDE_OUTER
    where(polygon%radius == radiusHi)&
      polygon%field = ON_OUTER
    where(polygon%radius < radiusHi .and. polygon%radius > radiusLo)&
      polygon%field = ANNULUS
    where(polygon%radius == radiusLo)&
      polygon%field = ON_INNER

    ! Now construct a chain of points comprising the vertices of the original polygon, plus any intersection points between the polygon and the two circles, in order as they are encountered clockwise around the polygon.

    ptChain%centre%x = 0.0
    ptChain%centre%y = 0.0

    numInPts = size(polygon)
    numMidPts = 1
    do si = 1, numInPts
      siNext = 1 + modulo(si, numInPts)

      ptChain(numMidPts) = polygon(si)

      call calcIntersectLineAnnulus(polygon(si), polygon(siNext)&
        , radiusLo, radiusHi, ptChain, numMidPts, status)

      if (status < 0) then
        allocate(polyVec(0))
  return
      end if
    end do

    ! Now calculate the polyCurves:

    if (all(ptChain(1:numMidPts)%field == OUTSIDE_OUTER)&
    .or. (.not. any(ptChain(1:numMidPts)%field == ANNULUS&
    .or. ptChain(1:numMidPts)%field == ON_INNER&
    .or. ptChain(1:numMidPts)%field == INSIDE_INNER)&
    .and. count(ptChain(1:numMidPts)%field == ON_OUTER) <= 1)) then

      if (pointIsInsidePoly(0.0, 0.0, ptChain(1:numMidPts)%pos, .true.)) then
      ! Result is just a simple annulus.

        allocate(polyVec(2))

        allocate(polyVec(1)%poly(1))
        polyVec(1)%poly%pos%x = radiusHi
        polyVec(1)%poly%pos%y = 0.0
        polyVec(1)%poly%centre%x = 0.0
        polyVec(1)%poly%centre%y = 0.0
        polyVec(1)%poly%isStraight = .false.
        polyVec(1)%poly%angle = 0.0
        polyVec(1)%poly%radius = radiusHi
        polyVec(1)%poly%field = ON_OUTER
        polyVec(1)%poly%arcIsClockwise = .true.
        polyVec(1)%poly%centreLiesToRight = .false.

        allocate(polyVec(2)%poly(1))
        polyVec(2)%poly%pos%x = radiusLo
        polyVec(2)%poly%pos%y = 0.0
        polyVec(2)%poly%centre%x = 0.0
        polyVec(2)%poly%centre%y = 0.0
        polyVec(2)%poly%isStraight = .false.
        polyVec(2)%poly%angle = 0.0
        polyVec(2)%poly%radius = radiusLo
        polyVec(2)%poly%field = ON_INNER
        polyVec(2)%poly%arcIsClockwise = .false.
        polyVec(2)%poly%centreLiesToRight = .true.

      else ! there is no intersection.
        allocate(polyVec(0))
      end if
  return
    end if

    allocate(ptAlreadyInPoly(numMidPts), mask(numMidPts))
    ptAlreadyInPoly = .false.
    numPolys = 0
    do
      mask = ((ptChain(1:numMidPts)%field == ON_OUTER&
        .or.   ptChain(1:numMidPts)%field == ANNULUS&
        .or.   ptChain(1:numMidPts)%field == ON_INNER)&
        .and. .not. ptAlreadyInPoly)

    if (.not. any(mask)) exit

      numPolys = numPolys + 1
      allocate(tempPolyVec(numPolys)%poly(numMidPts))
      pi = smallestTrueIndex(mask)

      numTempSides(numPolys) = 1
      tempPolyVec(numPolys)%poly(numTempSides(numPolys)) = ptChain(pi)
      ptAlreadyInPoly(pi) = .true.
      oldPi = pi
      do
        pi = pi + 1
        if (pi > numMidPts) pi = 1

        if (.not. (ptChain(pi)%field == ON_OUTER&
        .or.       ptChain(pi)%field == ANNULUS&
        .or.       ptChain(pi)%field == ON_INNER)) then

          niLast = numTempSides(numPolys) - 1

          if (tempPolyVec(numPolys)%poly(niLast)%field == ON_OUTER) then
            ! Can assume tempPolyVec(numPolys)%poly(niLast)%arcIsClockwise

            mask = (ptChain(1:numMidPts)%field == ON_OUTER)

            ! Fetch the next (in a cyclic sense) true index of mask, ie the next point which lies ON_OUTER. Note that this only works if ptChain is convex. Note that 'cyclic' can mean that the 'next' point can actually be the same point again. No problem.

            if (.not. any(mask(oldPi:))) then ! cycle to the beginning:
              pi = smallestTrueIndex(mask)
            else
              mask(:oldPi) = .false.
              pi = smallestTrueIndex(mask)
            end if

          else&
          if (tempPolyVec(numPolys)%poly(niLast)%field == ON_INNER) then
            ! Can assume the arc is anticlockwise.

            mask = (ptChain(1:numMidPts)%field == ON_INNER)

            ! Fetch the previous (in a cyclic sense) true index of mask, ie the nearest previous point which lies ON_INNER. Note that this only works if ptChain is convex. Note that 'cyclic' can mean that the 'previous' point can actually be the same point again. No problem.

            if (.not. any(mask(:oldPi))) then ! cycle to the end:
              pi = largestTrueIndex(mask)
            else
              mask(oldPi:) = .false.
              pi = largestTrueIndex(mask)
            end if

          else
call error('regridders_bug0',"ssclib:calcPartOfPolygonInAnnulus: Please contact the code developer.")
! can't go straight from ANNULUS to OUTSIDE.
          end if
        end if

      if (ptAlreadyInPoly(pi)) exit

        numTempSides(numPolys) = numTempSides(numPolys) + 1
        tempPolyVec(numPolys)%poly(numTempSides(numPolys)) = ptChain(pi)
        ptAlreadyInPoly(pi) = .true.
        oldPi = pi
      end do
    end do

    ! Add centreLiesToRight info:

    do polyNum = 1, numPolys
    if (numTempSides(polyNum) <= 1) cycle
      do ni = 1, numTempSides(polyNum)
        niNext = 1 + modulo(ni, numTempSides(polyNum))

        if (originLiesToRight(tempPolyVec(polyNum)%poly(ni    )%pos&
                            , tempPolyVec(polyNum)%poly(niNext)%pos))&
        then
          tempPolyVec(polyNum)%poly(ni)%centreLiesToRight = .true.
        else
          tempPolyVec(polyNum)%poly(ni)%centreLiesToRight = .false.
        end if
      end do
    end do

    ! Transfer to output:

    if (numPolys == 1) then
      if (.not.&
        any(tempPolyVec(1)%poly(:numTempSides(1))%field==ON_INNER&
        .or.tempPolyVec(1)%poly(:numTempSides(1))%field==INSIDE_INNER)) then

        allocate(polyVec(2))

        allocate(polyVec(1)%poly(numTempSides(1)))
        polyVec(1)%poly = tempPolyVec(1)%poly(:numTempSides(1))

        ! Add inner cutout circle:

        allocate(polyVec(2)%poly(1))
        polyVec(2)%poly%pos%x = radiusLo
        polyVec(2)%poly%pos%y = 0.0
        polyVec(2)%poly%centre%x = 0.0
        polyVec(2)%poly%centre%y = 0.0
        polyVec(2)%poly%isStraight = .false.
        polyVec(2)%poly%angle = 0.0
        polyVec(2)%poly%radius = radiusLo
        polyVec(2)%poly%field = ON_INNER
        polyVec(2)%poly%arcIsClockwise = .false.
        polyVec(2)%poly%centreLiesToRight = .true.

  return
      end if
    end if

    allocate(polyVec(numPolys))
    do polyNum = 1, numPolys
      if (numTempSides(polyNum) == 1&
      .and. tempPolyVec(polyNum)%poly(1)%isStraight) then
        allocate(polyVec(polyNum)%poly(0))

      else
        allocate(polyVec(polyNum)%poly(numTempSides(polyNum)))
        polyVec(polyNum)%poly&
          = tempPolyVec(polyNum)%poly(:numTempSides(polyNum))
      end if
    end do

  end subroutine calcPartOfPolygonInAnnulus

  !---------------------------------------------------------------------
  subroutine calcIntersectLineAnnulus(vertexLo, vertexHi&
  , radiusLo, radiusHi, ptChain, numPts, status)
  !---------------------------------------------------------------------

    implicit none 

    type(CurvedSideT), intent(in)    :: vertexLo, vertexHi
    real(single),      intent(in)    :: radiusLo, radiusHi
    type(CurvedSideT), intent(inout) :: ptChain(20)
    integer,           intent(inout) :: numPts
    integer,           intent(out)   :: status

    type(Point2DT) :: v, o
    real(single) :: a, b, c, rootContents, root
    real(single) :: intersectAngle, k(4)
    ! integer :: circleStatus

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    k = 0.0 ! just to give it some value - don't like
            ! floating variables, messy untidy things.

    if (  (vertexLo%field == INSIDE_INNER .or. vertexLo%field == ON_INNER)&
    .and. (vertexHi%field == INSIDE_INNER .or. vertexHi%field == ON_INNER))&
  return ! because can't be any intersections.

    ! See intersectLineCircle for theory.

    v%x = vertexHi%pos%x - vertexLo%pos%x
    v%y = vertexHi%pos%y - vertexLo%pos%y
    o = vertexLo%pos

    a = v%x * v%x + v%y * v%y
    if (a == 0.0)&
call error('identicalPoints',"ssclib:calcIntersectLineAnnulus: The input &
&arguments vertexLo and vertexHi must be the same point because there is no &
&separation between them.")

    b = 2.0 * (o%x * v%x + o%y * v%y)
    c = o%x * o%x + o%y * o%y + radiusHi * radiusHi

    rootContents = b * b - 4.0 * a * c
    if (rootContents < 0.0) then
      if (.not. (vertexHi%field == OUTSIDE_OUTER&
           .and. vertexLo%field == OUTSIDE_OUTER))&
call error('regridders_bug1',"ssclib:calcIntersectLineAnnulus: Please contact the code developer.")

  return
    end if

    if (rootContents == 0.0) then
      if (.not. ((vertexHi%field == OUTSIDE_OUTER&
        .or.      vertexHi%field == ON_OUTER)&
        .and.    (vertexLo%field == OUTSIDE_OUTER&
        .or.      vertexLo%field == ON_OUTER)))&
call error('regridders_bug2',"ssclib:calcIntersectLineAnnulus: Please contact the code developer.")

  return ! otherwise, ignore the existence of this kissing point.
    end if

    ! Having reached here means that rootContents > 0.0. No error testing is done on the number of roots here because any combination of vertex fields can in principle give rise to 2 points of intersection.

    root = sqrt(rootContents)
    k(1) = (-b + root) / 2.0 / a
    k(2) = (-b - root) / 2.0 / a

    ! Catch all the funny cases, repair if possible, error if not:
    call fixUpLineCircleKs(vertexLo%field, vertexHi%field&
      , OUTSIDE_OUTER, ON_OUTER, k(1:2), status)

  if (status < 0) return

    ! Find k values for the inner circle:

    c = o%x * o%x + o%y * o%y + radiusLo * radiusLo

    rootContents = b * b - 4.0 * a * c
    if (rootContents < 0.0) then
      if ( vertexHi%field == ON_INNER .or. vertexHi%field == INSIDE_INNER&
      .or. vertexLo%field == ON_INNER .or. vertexLo%field == INSIDE_INNER)&
call error('regridders_bug3',"ssclib:calcIntersectLineAnnulus: Please contact the code developer.")

      call addTwoCircleIntercepts(radiusHi, o, v, ON_OUTER, k(1:2)&
        , ptChain, numPts)
  return
    end if

    if (rootContents == 0.0) then
      if ( vertexHi%field == INSIDE_INNER&
      .or. vertexLo%field == INSIDE_INNER)&
call error('regridders_bug4',"ssclib:calcIntersectLineAnnulus: Please contact the code developer.")

      call addTwoCircleIntercepts(radiusHi, o, v, ON_OUTER, k(1:2)&
        , ptChain, numPts)

  return ! ignore the existence of the inner kissing point.
    end if

    ! Having reached here means that rootContents > 0.0, ie k(1) and k(2) exist. No error testing is done on the number of roots here because any combination of vertex fields can in principle give rise to 2 points of intersection.

    root = sqrt(rootContents)
    k(3) = (-b + root) / 2.0 / a
    k(4) = (-b - root) / 2.0 / a

    ! Sort the k values:
    if (k(3) > k(4)) call swap(k(3), k(4))

    ! Catch all the funny cases, repair if possible, error if not:

    if (k(4) <= 0.0) then ! k(3) must also be <= 0 because we just
    ! sorted them.
      select case(vertexLo%field)
      case(OUTSIDE_OUTER,ON_OUTER)
        select case(vertexHi%field)
        case(OUTSIDE_OUTER)
          ! fine
        case(ON_OUTER)
          status = -1
        case(ANNULUS)
          status = -2
        case(ON_INNER)
          status = -3
        case(INSIDE_INNER)
          status = -4
        end select
      case(ANNULUS)
        select case(vertexHi%field)
        case(OUTSIDE_OUTER,ON_OUTER)
          ! fine
        case(ANNULUS)
          ! fine
        case(ON_INNER)
          status = -5
        case(INSIDE_INNER)
          status = -6
        end select
      case(ON_INNER)
        select case(vertexHi%field)
        case(OUTSIDE_OUTER,ON_OUTER)
          status = 18
          k(4) = 0.0
        case(ANNULUS)
          ! fine
        end select
      case(INSIDE_INNER)
        select case(vertexHi%field)
        case(OUTSIDE_OUTER,ON_OUTER)
          status = 19
          k(4) = epsilon(1.0)
        case(ANNULUS)
          status = 20
          k(4) = epsilon(1.0)
        end select
      end select
    else if (k(4) > 0.0 .and. k(4) < 1.0) then
      if (k(3) <= 0.0) then
        select case(vertexLo%field)
        case(OUTSIDE_OUTER,ON_OUTER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -7
          case(ANNULUS)
            status = -8
          case(ON_INNER)
            status = -9
          case(INSIDE_INNER)
            status = -10
          end select
        case(ANNULUS)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            if (k(4) > -k(3)) then
              status = 21
              k(3) = epsilon(1.0)
            else
              status = 22
              k(4) = -epsilon(1.0)
            end if
          case(ANNULUS)
!*** should sort out cases as above (check also similar cases in this decision tree).
            status = 23
            k(3) = epsilon(1.0)
          case(ON_INNER)
            status = 24
            k(3) = epsilon(1.0)
            k(4) = 1.0
          case(INSIDE_INNER)
            status = 25
            k(3) = epsilon(1.0)
            k(4) = 1.0 + epsilon(1.0)
          end select
        case(ON_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            ! fine
          case(ANNULUS)
            ! fine
          end select
        case(INSIDE_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            ! fine
          case(ANNULUS)
            ! fine
          end select
        end select
      else ! k(3) > 0 and < 1.
        select case(vertexLo%field)
        case(OUTSIDE_OUTER,ON_OUTER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            ! fine
          case(ANNULUS)
            ! fine
          case(ON_INNER)
            status = 26
            k(4) = 1.0
          case(INSIDE_INNER)
            status = 27
            k(4) = 1.0 + epsilon(1.0)
          end select
        case(ANNULUS)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            ! fine
          case(ANNULUS)
            ! fine
          case(ON_INNER)
            status = 28
            k(4) = 1.0
          case(INSIDE_INNER)
            status = 29
            k(4) = 1.0 + epsilon(1.0)
          end select
        case(ON_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = 30
            k(3) = 0.0
          case(ANNULUS)
            status = 31
            k(3) = 0.0
          end select
        case(INSIDE_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = 32
            k(3) = -epsilon(1.0)
          case(ANNULUS)
            status = 33
            k(3) = -epsilon(1.0)
          end select
        end select
      end if ! k(3) <= 0
    else ! k(4) >= 1.0
      if (k(3) <= 0.0) then
        select case(vertexLo%field)
        case(OUTSIDE_OUTER,ON_OUTER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -11
          case(ANNULUS)
            status = -12
          case(ON_INNER)
            status = -13
          case(INSIDE_INNER)
            status = -14
          end select
        case(ANNULUS)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -15
          case(ANNULUS)
            status = 34
            k(3) = epsilon(1.0)
            k(4) = 1.0 - epsilon(1.0)
          case(ON_INNER)
            status = 35
            k(3) = epsilon(1.0)
          case(INSIDE_INNER)
            status = 36
            k(3) = epsilon(1.0)
          end select
        case(ON_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -16
          case(ANNULUS)
            status = 37
            k(4) = 1.0 - epsilon(1.0)
          end select
        case(INSIDE_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -17
          case(ANNULUS)
            status = 38
            k(4) = 1.0 - epsilon(1.0)
          end select
        end select
      else if (k(3) > 0.0 .and. k(3) < 1.0) then
        select case(vertexLo%field)
        case(OUTSIDE_OUTER,ON_OUTER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -18
          case(ANNULUS)
            if (k(4) - 1 > 1 - k(3)) then
              status = 39
              k(3) = 1.0 + epsilon(1.0)
            else
              status = 40
              k(4) = 1.0 - epsilon(1.0)
            end if
          case(ON_INNER)
            ! fine
          case(INSIDE_INNER)
            ! fine
          end select
        case(ANNULUS)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -19
          case(ANNULUS)
            status = 41
            k(4) = 1.0 - epsilon(1.0)
          case(ON_INNER)
            ! fine
          case(INSIDE_INNER)
            ! fine
          end select
        case(ON_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -20
          case(ANNULUS)
            status = 42
            k(3) = 0.0
            k(4) = 1.0 - epsilon(1.0)
          end select
        case(INSIDE_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -21
          case(ANNULUS)
            k(3) = -epsilon(1.0)
            k(4) = 1.0 - epsilon(1.0)
          end select
        end select
      else ! k(3) >= 1.0
        select case(vertexLo%field)
        case(OUTSIDE_OUTER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            ! fine
          case(ANNULUS)
            ! fine
          case(ON_INNER)
            status = 43
            k(3) = 1.0
          case(INSIDE_INNER)
            status = 44
            k(3) = 1.0 - epsilon(1.0)
          end select
        case(ON_OUTER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -22
          case(ANNULUS)
            ! fine
          case(ON_INNER)
            status = 45
            k(3) = 1.0
          case(INSIDE_INNER)
            status = 46
            k(3) = 1.0 - epsilon(1.0)
          end select
        case(ANNULUS)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -23
          case(ANNULUS)
            ! fine
          case(ON_INNER)
            ! fine
          case(INSIDE_INNER)
            status = 47
            k(3) = 1.0 - epsilon(1.0)
          end select
        case(ON_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -24
          case(ANNULUS)
            status = -25
          end select
        case(INSIDE_INNER)
          select case(vertexHi%field)
          case(OUTSIDE_OUTER,ON_OUTER)
            status = -26
          case(ANNULUS)
            status = -27
          end select
        end select
      end if ! k(3) <= 0
    end if ! k(4) <= 0

  if (status < 0) return

    call quickSort(k)

    ! Calculate the details of the intersection points:

    if (k(1) > 0.0 .and. k(1) < 1.0) then
      numPts = numPts + 1

      intersectAngle = atan2(o%y + k(1) * v%y, o%x + k(1) * v%x)
      ptChain(numPts)%angle = intersectAngle
      ptChain(numPts)%radius = radiusHi
      ptChain(numPts)%pos%x = radiusHi * cos(intersectAngle)
      ptChain(numPts)%pos%y = radiusHi * sin(intersectAngle)
      ptChain(numPts)%field = ON_OUTER
      ptChain(numPts)%isStraight = .true.
    end if
    if (k(2) > 0.0 .and. k(2) < 1.0) then
      numPts = numPts + 1

      intersectAngle = atan2(o%y + k(2) * v%y, o%x + k(2) * v%x)
      ptChain(numPts)%angle = intersectAngle
      ptChain(numPts)%radius = radiusLo
      ptChain(numPts)%pos%x = radiusLo * cos(intersectAngle)
      ptChain(numPts)%pos%y = radiusLo * sin(intersectAngle)
      ptChain(numPts)%field = ON_INNER
      ptChain(numPts)%isStraight = .false.
      ptChain(numPts)%arcIsClockwise = .false.
    end if
    if (k(3) > 0.0 .and. k(3) < 1.0) then
      numPts = numPts + 1

      intersectAngle = atan2(o%y + k(3) * v%y, o%x + k(3) * v%x)
      ptChain(numPts)%angle = intersectAngle
      ptChain(numPts)%radius = radiusLo
      ptChain(numPts)%pos%x = radiusLo * cos(intersectAngle)
      ptChain(numPts)%pos%y = radiusLo * sin(intersectAngle)
      ptChain(numPts)%field = ON_INNER
      ptChain(numPts)%isStraight = .true.
    end if
    if (k(4) > 0.0 .and. k(4) < 1.0) then
      numPts = numPts + 1

      intersectAngle = atan2(o%y + k(4) * v%y, o%x + k(4) * v%x)
      ptChain(numPts)%angle = intersectAngle
      ptChain(numPts)%radius = radiusHi
      ptChain(numPts)%pos%x = radiusHi * cos(intersectAngle)
      ptChain(numPts)%pos%y = radiusHi * sin(intersectAngle)
      ptChain(numPts)%field = ON_OUTER
      ptChain(numPts)%isStraight = .false.
      ptChain(numPts)%arcIsClockwise = .true.
    end if

  end subroutine calcIntersectLineAnnulus 

  !---------------------------------------------------------------------
  subroutine calcPartOfPolygonInCircle(polygon, radius, polyVec, status)
  !---------------------------------------------------------------------
  ! Note! The calling routine should deallocate the pointer polyVec (and its component pointers!) after use.

    implicit none 

    type(CurvedSideT),      intent(inout) :: polygon(:)
    real(single),           intent(in)    :: radius
    type(PolycurveVectorT), pointer       :: polyVec(:)
    integer,                intent(out)   :: status

    type(CurvedSideT) :: ptChain(24) ! The absolute maximum conceivable number
    ! of points which can be in any single one of the output polycurves. This is deduced as follows: the maximum number of points in the input polygon is 12: if we assume that each of these sides crosses the circle twice, we get 12x2=24.

    type(PolycurveVectorT) :: tempPolyVec(24) ! This is the maximum number of pieces the intersection area between the input polygon and the annulus can fall into.

!*** actually, should make these allocatable really... otherwise the whole thing is tied to squares as input to calcPartOfPolygonInWedge.

    integer ::&
      numInPts, numMidPts, numTempSides(48), si, siNext, pi, numPolys, oldPi, niLast, polyNum
    logical, allocatable :: ptAlreadyInPoly(:), mask(:)

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    ! First work out which of the vertices of the polygon are outside, exactly on, or inside the circle:

    polygon%field = INSIDE_CIRCLE ! default
    where(polygon%radius > radius)&
      polygon%field = OUTSIDE_CIRCLE
    where(polygon%radius == radius)&
      polygon%field = ON_CIRCLE

    ! Now construct a chain of points comprising the vertices of the original polygon, plus any intersection points between the polygon and the circle, in order as they are encountered clockwise around the polygon. Store at each point the angle and whether the point is within, without or on the circle.

    ptChain%centre%x = 0.0
    ptChain%centre%y = 0.0

    numInPts = size(polygon)
    numMidPts = 1
    do si = 1, numInPts
      siNext = 1 + modulo(si, numInPts)

      ptChain(numMidPts) = polygon(si)

      call calcIntersectLineCircle(polygon(si), polygon(siNext)&
        , radius, ptChain, numMidPts, status)

      if (status < 0) then
        allocate(polyVec(0))
  return
      end if
    end do

    ! Now calculate the polyCurves:

    if (all(ptChain(1:numMidPts)%field == OUTSIDE_CIRCLE)&
    .or. (.not. any(ptChain(1:numMidPts)%field == INSIDE_CIRCLE)&
    .and. count(ptChain(1:numMidPts)%field == ON_CIRCLE) <= 1)) then

      if (pointIsInsidePoly(0.0, 0.0, ptChain(1:numMidPts)%pos, .true.)) then
      ! Result is just a simple circle.

        allocate(polyVec(1))

        allocate(polyVec(1)%poly(1))
        polyVec(1)%poly%pos%x = radius
        polyVec(1)%poly%pos%y = 0.0
        polyVec(1)%poly%centre%x = 0.0
        polyVec(1)%poly%centre%y = 0.0
        polyVec(1)%poly%isStraight = .false.
        polyVec(1)%poly%angle = 0.0
        polyVec(1)%poly%radius = radius
        polyVec(1)%poly%field = ON_CIRCLE
        polyVec(1)%poly%arcIsClockwise = .true.

      else ! there is no intersection.
        allocate(polyVec(0))
      end if
  return
    end if

    allocate(ptAlreadyInPoly(numMidPts), mask(numMidPts))
    ptAlreadyInPoly = .false.
    numPolys = 0
    do
      mask = ((ptChain(1:numMidPts)%field .ne. OUTSIDE_CIRCLE)&
        .and. .not. ptAlreadyInPoly)

    if (.not. any(mask)) exit

      numPolys = numPolys + 1
      allocate(tempPolyVec(numPolys)%poly(numMidPts))
      pi = smallestTrueIndex(mask)

      numTempSides(numPolys) = 1
      tempPolyVec(numPolys)%poly(numTempSides(numPolys)) = ptChain(pi)
      ptAlreadyInPoly(pi) = .true.
      oldPi = pi
      do
        pi = pi + 1
        if (pi > numMidPts) pi = 1

        if (ptChain(pi)%field == OUTSIDE_CIRCLE) then
          niLast = numTempSides(numPolys) - 1

          if (tempPolyVec(numPolys)%poly(niLast)%field == ON_CIRCLE) then
            ! Can assume tempPolyVec(numPolys)%poly(niLast)%arcIsClockwise

            mask = (ptChain(1:numMidPts)%field == ON_CIRCLE)

            ! Fetch the next (in a cyclic sense) true index of mask, ie the next point which lies ON_CIRCLE. Note that this only works if ptChain is convex. Note that 'cyclic' can mean that the 'next' point can actually be the same point again. No problem.

            if (.not. any(mask(oldPi:))) then ! cycle to the beginning:
              pi = smallestTrueIndex(mask)
            else
              mask(:oldPi) = .false.
              pi = smallestTrueIndex(mask)
            end if

          else
call error('regridders_bug5',"ssclib:calcIntersectLineAnnulus: Please contact the code developer.")
! can't go straight from INSIDE_CIRCLE to OUTSIDE_CIRCLE.
          end if
        end if

      if (ptAlreadyInPoly(pi)) exit

        numTempSides(numPolys) = numTempSides(numPolys) + 1
        tempPolyVec(numPolys)%poly(numTempSides(numPolys)) = ptChain(pi)
        ptAlreadyInPoly(pi) = .true.
        oldPi = pi
      end do
    end do

    ! Transfer to output:

    allocate(polyVec(numPolys))
    do polyNum = 1, numPolys
      if (numTempSides(polyNum) == 1&
      .and. tempPolyVec(polyNum)%poly(1)%isStraight) then
        allocate(polyVec(polyNum)%poly(0))

      else
        allocate(polyVec(polyNum)%poly(numTempSides(polyNum)))
        polyVec(polyNum)%poly&
          = tempPolyVec(polyNum)%poly(:numTempSides(polyNum))
      end if
    end do

  end subroutine calcPartOfPolygonInCircle

  !---------------------------------------------------------------------
  subroutine calcIntersectLineCircle(vertexLo, vertexHi, radius&
  , ptChain, numPts, status)
  !---------------------------------------------------------------------
  ! The aim is to fix all possible problems caused by numerical rounding within this routine.

    implicit none 

    type(CurvedSideT), intent(in)    :: vertexLo, vertexHi
    real(single),      intent(in)    :: radius
    type(CurvedSideT), intent(inout) :: ptChain(12)
    integer,           intent(inout) :: numPts
    integer,           intent(out)   :: status

    type(Point2DT) :: v, o
    real(single) :: a, b, c, rootContents, root
    real(single) :: k(2)

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    k = 0.0 ! just to give it some value - don't like
            ! floating variables, messy untidy things.

    if (  (vertexLo%field == INSIDE_CIRCLE .or. vertexLo%field == ON_CIRCLE)&
    .and. (vertexHi%field == INSIDE_CIRCLE .or. vertexHi%field == ON_CIRCLE))&
  return ! because can't be any intersections.

    ! Let us denote the vector between the origin and vertexLo%pos by _o_, the vector from vertexLo%pos to vertexHi%pos by _v_. If we extend the side _v_ in both directions then this line may intersect with a circle of radius r at zero, one or two points. Let us denote a vector from the origin to any of the intersection points by _c_. Clearly each _c_ equals _o_ + k*_v_ for some constant k. We can write down 3 equations:
    !
    !	o_x + k*v_x = c_x
    !	o_y + k*v_y = c_y
    !	c_x^2 +c_y^2 = r^2.
    !
    ! These reduce to the following quadratic equation:
    !
    !	(v_x^2+v_y^2)*k^2 + 2(o_x*v_x+o_y*v_y)*k + (o_x^2+o_y^2+r^2) = 0.
    !
    ! The first job is to solve this for k.

    v%x = vertexHi%pos%x - vertexLo%pos%x
    v%y = vertexHi%pos%y - vertexLo%pos%y
    o = vertexLo%pos

    a = v%x * v%x + v%y * v%y
call error('identicalPoints',"ssclib:calcIntersectLineCircle: The input &
&arguments vertexLo and vertexHi must be the same point because there is no &
&separation between them.")

    b = 2.0 * (o%x * v%x + o%y * v%y)
    c = o%x * o%x + o%y * o%y + radius * radius

    rootContents = b * b - 4.0 * a * c
    if (rootContents < 0.0) then
      if (.not. (vertexHi%field == OUTSIDE_CIRCLE&
           .and. vertexLo%field == OUTSIDE_CIRCLE))&
call error('regridders_bug6',"ssclib:calcIntersectLineAnnulus: Please contact the code developer.")

  return
    end if

    if (rootContents == 0.0) then
      if ( vertexHi%field == INSIDE_CIRCLE&
      .or. vertexLo%field == INSIDE_CIRCLE)&
call error('regridders_bug7',"ssclib:calcIntersectLineAnnulus: Please contact the code developer.")

  return ! otherwise, ignore the existence of this kissing point.
    end if

    ! Else rootContents > 0.0.

    ! No error testing on the number of roots here because any combination
    ! of vertex fields can in principle give rise to 2 points of intersection.

    root = sqrt(rootContents)
    k(1) = (-b + root) / 2.0 / a
    k(2) = (-b - root) / 2.0 / a

    ! Catch all the funny cases, repair if possible, error if not:
    call fixUpLineCircleKs(vertexLo%field, vertexHi%field&
      , OUTSIDE_CIRCLE, ON_CIRCLE, k, status)

  if (status < 0) return

    ! Now calculate the details of the intersection points:
    call addTwoCircleIntercepts(radius, o, v, ON_CIRCLE&
      , k, ptChain, numPts)

  end subroutine calcIntersectLineCircle

  !---------------------------------------------------------------------
  subroutine fixUpLineCircleKs(vertexLoField, vertexHiField&
  , LOCAL_OUTSIDE_CIRCLE, LOCAL_ON_CIRCLE, k, status)
  !---------------------------------------------------------------------

    implicit none

    integer(int8), intent(in)    :: vertexLoField,&
                                    vertexHiField,&
                                    LOCAL_OUTSIDE_CIRCLE,&
                                    LOCAL_ON_CIRCLE 
    real(single),  intent(inout) :: k(2)
    integer,       intent(out)   :: status

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    ! Sort the k values:
    if (k(1) > k(2)) call swap(k(1), k(2))

    if (k(2) <= 0.0) then ! k(1) must also be <= 0 because we just
    ! sorted them.

      if (vertexLoField == LOCAL_OUTSIDE_CIRCLE) then
        if (vertexHiField == LOCAL_OUTSIDE_CIRCLE) then
          ! fine
        else if (vertexHiField == LOCAL_ON_CIRCLE) then
          status = -1
        else ! vertexHiField == INSIDE_CIRCLE
          status = -2
        end if
      else if (vertexLoField == LOCAL_ON_CIRCLE) then
        ! fine
      else ! vertexLoField == INSIDE_CIRCLE
        status = 1
        k(2) = epsilon(1.0)
      end if
    else if (k(2) > 0.0 .and. k(2) < 1.0) then
      if (k(1) <= 0.0) then
        if (vertexLoField == LOCAL_OUTSIDE_CIRCLE) then
          if (vertexHiField == LOCAL_OUTSIDE_CIRCLE) then
            status = 2
            k(1) = epsilon(1.0)
          else if (vertexHiField == LOCAL_ON_CIRCLE) then
            status = 3
            k(1) = epsilon(1.0)
            k(2) = 1.0
          else ! vertexHiField == INSIDE_CIRCLE
            status = 4
            k(1) = epsilon(1.0)
            k(2) = 1.0 + epsilon(1.0)
          end if
        else if (vertexLoField == LOCAL_ON_CIRCLE) then
          ! fine
        else ! vertexLoField == INSIDE_CIRCLE
          ! fine
        end if
      else ! k(1) > 0 and < 1.
        if (vertexLoField == LOCAL_OUTSIDE_CIRCLE) then
          if (vertexHiField == LOCAL_OUTSIDE_CIRCLE) then
            ! fine
          else if (vertexHiField == LOCAL_ON_CIRCLE) then
            status = 5
            k(2) = 1.0
          else ! vertexHiField == INSIDE_CIRCLE
            status = 6
            k(2) = 1.0 + epsilon(1.0)
          end if
        else if (vertexLoField == LOCAL_ON_CIRCLE) then
          status = 7
          k(1) = 0.0
        else ! vertexLoField == INSIDE_CIRCLE
          status = 8
          k(1) = -epsilon(1.0)
        end if
      end if
    else ! k(2) >= 1.0
      if (k(1) <= 0.0) then
        if (vertexLoField == LOCAL_OUTSIDE_CIRCLE) then
          if (vertexHiField == LOCAL_OUTSIDE_CIRCLE) then
            status = 9
            k(1) = epsilon(1.0)
            k(2) = 1.0 - epsilon(1.0)
          else if (vertexHiField == LOCAL_ON_CIRCLE) then
            status = 10
            k(1) = epsilon(1.0)
          else ! vertexHiField == INSIDE_CIRCLE
            status = 11
            k(1) = epsilon(1.0)
          end if
        else if (vertexLoField == LOCAL_ON_CIRCLE) then
          status = 12
          k(2) = 1.0 - epsilon(1.0)
        else ! vertexLoField == INSIDE_CIRCLE
          status = 13
          k(2) = 1.0 - epsilon(1.0)
        end if
      else if (k(1) > 0.0 .and. k(1) <= 1.0) then
        if (vertexLoField == LOCAL_OUTSIDE_CIRCLE) then
          if (vertexHiField == LOCAL_OUTSIDE_CIRCLE) then
            status = 14
            k(2) = 1.0 - epsilon(1.0)
          else if (vertexHiField == LOCAL_ON_CIRCLE) then
            ! fine
          else ! vertexHiField == INSIDE_CIRCLE
            ! fine
          end if
        else if (vertexLoField == LOCAL_ON_CIRCLE) then
          status = 15
          k(1) = 0.0
          k(2) = 1.0 - epsilon(1.0)
        else ! vertexLoField == INSIDE_CIRCLE
          status = 16
          k(1) = -epsilon(1.0)
          k(2) = 1.0 - epsilon(1.0)
        end if
      else ! k(1) >= 1.0
        if (vertexLoField == LOCAL_OUTSIDE_CIRCLE) then
          if (vertexHiField == LOCAL_OUTSIDE_CIRCLE) then
            ! fine
          else if (vertexHiField == LOCAL_ON_CIRCLE) then
            ! fine
          else ! vertexHiField == INSIDE_CIRCLE
            status = 17
            k(1) = 1.0 - epsilon(1.0)
          end if
        else if (vertexLoField == LOCAL_ON_CIRCLE) then
          status = -3
        else ! vertexLoField == INSIDE_CIRCLE
          status = -4
        end if
      end if
    end if

  end subroutine fixUpLineCircleKs

  !---------------------------------------------------------------------
  subroutine addTwoCircleIntercepts(radius, o, v, LOCAL_ON_CIRCLE&
  , k, ptChain, numPts)
  !---------------------------------------------------------------------

    implicit none

    real(single),      intent(in)    :: radius
    type(Point2dT),    intent(in)    :: o, v
    integer(int8),     intent(in)    :: LOCAL_ON_CIRCLE
    real(single),      intent(inout) :: k(2)
    type(CurvedSideT), intent(inout) :: ptChain(:)
    integer,           intent(inout) :: numPts

    real(single) :: intersectAngle

    !-------------------------------------------------------------------
    ! Start:

    ! Calculate the details of the intersection points:

    if (k(1) > 0.0 .and. k(1) < 1.0) then
      numPts = numPts + 1

      intersectAngle = atan2(o%y + k(1) * v%y, o%x + k(1) * v%x)
      ptChain(numPts)%angle = intersectAngle
      ptChain(numPts)%radius = radius
      ptChain(numPts)%pos%x = radius * cos(intersectAngle)
      ptChain(numPts)%pos%y = radius * sin(intersectAngle)
      ptChain(numPts)%field = LOCAL_ON_CIRCLE
      ptChain(numPts)%isStraight = .true.
    end if
    if (k(2) > 0.0 .and. k(2) < 1.0) then
      numPts = numPts + 1

      intersectAngle = atan2(o%y + k(2) * v%y, o%x + k(2) * v%x)
      ptChain(numPts)%angle = intersectAngle
      ptChain(numPts)%radius = radius
      ptChain(numPts)%pos%x = radius * cos(intersectAngle)
      ptChain(numPts)%pos%y = radius * sin(intersectAngle)
      ptChain(numPts)%field = LOCAL_ON_CIRCLE
      ptChain(numPts)%isStraight = .false.
    end if

  end subroutine addTwoCircleIntercepts

  !---------------------------------------------------------------------
  logical function originLiesToRight(a, b)
  !---------------------------------------------------------------------
  ! given a vector defined by its starting and ending points a and b, determines if the origin lies to the right or left of the vector. (Origin lying directly in the line of the vector is arbitrarily counted as 'right'.)

    implicit none 

    type(Point2DT), intent(in) :: a, b

    real(single) :: aVec(3), v(3), c(3)

    !-------------------------------------------------------------------
    ! Start:

    aVec(1) = a%x
    aVec(2) = a%y
    aVec(3) = 0.0

    v(1) = b%x - a%x
    v(2) = b%y - a%y
    v(3) = 0.0

    c = vectorCross(aVec, v) ! in ssclib/array_utils.

    if (c(3) >= 0.0) then
      originLiesToRight = .true.
    else
      originLiesToRight = .false.
    end if

  end function originLiesToRight

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module regridders
