!
! Program: testexposure
!
! Authors:  Masaaki Sakano - SSC/LUX
!          
! Description: Test exposure module.
! 
! Comments:
!
!-
program testexposure
  use cal	! for convertPiToEeV (and CAL_setState)
  use dal
  use dss	! for RangeT and BoundT
  use errorhandling
  ! use test_utils	! for isNearlyEqual()
  use exposure

  implicit none

  integer(int32), parameter :: maxRep = 1	! for (memory) stress test.
  ! It takes 10sec ish with aral for maxRep = 100000 (in 2008).
  integer(int32) :: i

  do i=1,maxRep
     call test_convertPiToEeV
  end do

contains

  !---------------------------------------------------------------------
  subroutine test_convertPiToEeV
  !---------------------------------------------------------------------
    
    implicit none

    call message(AppMsg, Sparse, 'Try with EMOS2...')
    call CAL_setState(instrument=EMOS2)
    call test_convertPiToEeV_core	! 8 tests including 18 assertions

    call message(AppMsg, Sparse, 'Try with EPN...')
    call CAL_setState(instrument=EPN)
    call test_convertPiToEeV_core	! 8 tests including 18 assertions

  end subroutine test_convertPiToEeV


  !---------------------------------------------------------------------
  subroutine test_convertPiToEeV_core
  !---------------------------------------------------------------------
  ! Note! The cal state should be set for instrument before this routine
  ! is called.

    implicit none

    type(BoundT) :: piBoundsLower(2), piBoundsUpper(2), EeVBounds(2)
    type(RangeT) :: piRange, EeVrange
    type(RangeT) :: piRanges(2), EeVranges(2)
    real(double) :: piValues(2)  = (/ 30, 300 /), EeVValues(2)
    real(single) :: piValuesS(2) = (/ 30, 300 /), EeVValuesS(2)

    !-------------------------------------------------------------------
    ! Start

    piBoundsLower%type = INCLUSIVE	! defined in dsslib
    piBoundsUpper%type = INCLUSIVE	! defined in dsslib
    piBoundsLower(1)%value = 2
    piBoundsUpper(1)%value = 20
    piBoundsLower(2)%value = 200
    piBoundsUpper(2)%value = 2000

    !! Test of convertPiBoundToEeV(piBound)
    EeVBounds(1) = convertPiToEeV(piBoundsLower(1))	! convertPiBoundToEeV(piBound)
    if ((EeVBounds(1)%type  .ne. piBoundsLower(1)%type)  .or. &
        (EeVBounds(1)%value .ne. piBoundsLower(1)%value)) then
      write(errstr, *) "Inconsistent EeVBound = (", EeVBounds%type &
           , ",", EeVBounds%value, ")"
      call error("InconsistentEeVBound")
    end if

    !! Test of convertPiBoundsToEeV(piBounds)
    EeVBounds = convertPiToEeV(piBoundsLower)	! convertPiBoundsToEeV(piBounds)
    if ((EeVBounds(1)%type  .ne. piBoundsLower(1)%type)  .or. &
        (EeVBounds(1)%value .ne. piBoundsLower(1)%value) .or. &
        (EeVBounds(2)%type  .ne. piBoundsLower(2)%type)  .or. &
        (EeVBounds(2)%value .ne. piBoundsLower(2)%value)) then
      write(errstr, *) "Inconsistent EeVBounds..."
      call error("InconsistentEeVBounds")
    end if

    !! Test of convertPiRangeToEeV(piRange)
    piRange%lower = piBoundsLower(1)
    piRange%upper = piBoundsUpper(1)
    EeVrange = convertPiToEeV(piRange)	! = convertPiRangeToEeV(piRange)
    if ((EeVrange%lower%value .ne. piRange%lower%value) .or. &
        (EeVrange%upper%value .ne. piRange%upper%value)) then
      write(errstr, *) "Inconsistent EeVrange = (", EeVrange%lower%value &
           , ",", EeVrange%upper%value, ")"
      call error("InconsistentEeVrange")
    end if

    !! Test of convertPiRangesToEeV(piRange)
    piRanges%lower = piBoundsLower
    piRanges%upper = piBoundsUpper
    EeVranges = convertPiToEeV(piRanges)	! = convertPiRangesToEeV(piRanges)
    if ((EeVranges(1)%lower%value .ne. piRanges(1)%lower%value) .or. &
        (EeVranges(1)%upper%value .ne. piRanges(1)%upper%value) .or. &
        (EeVranges(2)%lower%value .ne. piRanges(2)%lower%value) .or. &
        (EeVranges(2)%upper%value .ne. piRanges(2)%upper%value)) then
      write(errstr, *) "Inconsistent EeVranges..."
      call error("InconsistentEeVranges")
    end if

    !! Test of convertPiToEeVScalarSingle(piValue)
    EeVValuesS(1) = convertPiToEeV(piValuesS(1))	! = convertPiToEeVScalarSingle(piValue)
    if (EeVValuesS(1) .ne. piValuesS(1)) then
      write(errstr, *) "Inconsistent EeVValueS = ", EeVValuesS(1), " (!= ", piValuesS(1), ")"
      call error("InconsistentEeVValueSingle")
    end if

    !! Test of convertPiToEeVScalarDouble(piValues)
    EeVValues(1) = convertPiToEeV(piValues(1))	! = convertPiToEeVScalarDouble(piValues)
    if (EeVValues(1) .ne. piValues(1)) then
      write(errstr, *) "Inconsistent EeVValue = ", EeVValues(1), " (!= ", piValues(1), ")"
      call error("InconsistentEeVValueDouble")
    end if

    !! Test of convertPiToEeVVectorSingle(piValuesS)
    EeVValuesS = convertPiToEeV(piValuesS)	! = convertPiToEeVVectorSingle(piValues)
    if ((EeVValuesS(1) .ne. piValuesS(1)) .or. &
        (EeVValuesS(2) .ne. piValuesS(2))) then
      write(errstr, *) "Inconsistent EeVValuesSingle..."
      call error("InconsistentEeVValuesSingle")
    end if

    !! Test of convertPiToEeVVectorDouble(piValues)
    EeVValues = convertPiToEeV(piValues)	! = convertPiToEeVVectorDouble(piValues)
    if ((EeVValues(1) .ne. piValues(1)) .or. &
        (EeVValues(2) .ne. piValues(2))) then
      write(errstr, *) "Inconsistent EeVValuesDouble..."
      call error("InconsistentEeVValuesDouble")
    end if

  end subroutine test_convertPiToEeV_core

end program testexposure
